#include "Go4kVSTiCore.h"

#include <math.h>

#include "Go4kVSTiGUI.h"
#include "..\..\win\resource.h"
#include <stdio.h>

#include <vector>
#include <map>
#include <list>
#include <string>

DWORD versiontag10 = 0x30316b34; // 4k10
DWORD versiontag11 = 0x31316b34; // 4k11
DWORD versiontag12 = 0x32316b34; // 4k12
DWORD versiontag13 = 0x33316b34; // 4k13
DWORD versiontagfk301   = 0xf000abba; // forked klang 3.01
DWORD versiontag   = 0xf001abba; // forked klang 3.02

static SynthObject SynthObj;

static GlobalSettingsS GlobalSettings;

extern "C" void __stdcall go4kENV_func();
extern "C" void __stdcall go4kVCO_func();
extern "C" void __stdcall go4kVCF_func();
extern "C" void __stdcall go4kDST_func();
extern "C" void __stdcall go4kDLL_func();
extern "C" void __stdcall go4kFOP_func();
extern "C" void __stdcall go4kFST_func();
extern "C" void __stdcall go4kPAN_func();
extern "C" void __stdcall go4kOUT_func();
extern "C" void __stdcall go4kACC_func();
extern "C" void __stdcall go4kFLD_func();
extern "C" void __stdcall go4kSLW_func();
extern "C" void __stdcall go4kFLDFROM_func();
extern "C" DWORD go4k_delay_buffer_ofs;
extern "C" float go4k_delay_buffer;
extern "C" DWORD go4kDLL_wrk_size;
extern "C" DWORD go4k_delay_buffer_size;
//extern "C" DWORD __stdcall go4k_get_delay_buffer_size();

extern "C" float go4k_transformed_values_buffer;
extern "C" float *go4k_transformed_values_cursor;
extern "C" DWORD go4k_transformed_values_update_counter;


extern "C" WORD go4k_delay_times;
extern "C" float LFO_NORMALIZE;

typedef void (__stdcall *go4kFunc)(void); 

void __stdcall NULL_func()
{
};

static go4kFunc SynthFuncs[NUM_MODULES] =
{
	NULL_func,
	go4kENV_func,
	go4kVCO_func,
	go4kVCF_func,
	go4kDST_func,
	go4kDLL_func,
	go4kFOP_func,
	go4kFST_func,
	go4kPAN_func,
	go4kOUT_func,
	go4kACC_func,
	go4kFLD_func,
	go4kSLW_func,
    go4kFLDFROM_func,
};


int num_dwords_to_skip_in_the_beginning_of_units_workspace_because_they_are_not_floats[NUM_MODULES] =
{
	0, // M_NONE = 0,
	1, // M_ENV (skip .state before .level)
	0, // M_VCO
	0, // M_VCF
	0, // M_DST
	1, // M_DLL, skip .index
	0, // M_FOP
	0, // M_FSTx, special case
	0, // M_PAN
	0, // M_OUT
	0, // M_ACC
	0, // M_FLD
    0, // M_SLW, special case
    0, // M_FLDFROM
};

int num_workspace_floats[NUM_MODULES] =
{
	0, // M_NONE = 0,
	1, // M_ENV
	1, // M_VCO
	4, // M_VCF
	2, // M_DST
	4, // M_DLL, and after these there's the whole delay buffer
	0, // M_FOP
	0, // M_FSTx, another special case
	0, // M_PAN
	0, // M_OUT
	0, // M_ACC
	0, // M_FLD
    0, // M_SLW, special case
    0, // M_FLDFROM
};


#define DENORMAL_CLEAN_THRESHOLD 0.00001
// "Cleaning" means, round small values to 0.0.
void clean_denormals(float *f, int num_floats)
{
    for (int i=0; i<num_floats; i++)
    {
        if (fabs(*f) < DENORMAL_CLEAN_THRESHOLD)
            *f = 0.0;
        f++;
    }
}

void clear_floats(float *f, int num_floats)
{
    for (int i=0; i<num_floats; i++)
        *f++ = 0.0;
}


// "Cleaning" means, round small values to 0.0.
void clean_fpu_denormals()
{
    float fpu_regs[8], *f;

    __asm {
        fstp dword ptr fpu_regs[0]
        fstp dword ptr fpu_regs[4]
        fstp dword ptr fpu_regs[8]
        fstp dword ptr fpu_regs[12]
        fstp dword ptr fpu_regs[16]
        fstp dword ptr fpu_regs[20]
        fstp dword ptr fpu_regs[24]
        fstp dword ptr fpu_regs[28]
    }

    f = fpu_regs;
    if (fabs(*f) <  DENORMAL_CLEAN_THRESHOLD) *f = 0.0; f++;    
    if (fabs(*f) <  DENORMAL_CLEAN_THRESHOLD) *f = 0.0; f++;    
    if (fabs(*f) <  DENORMAL_CLEAN_THRESHOLD) *f = 0.0; f++;    
    if (fabs(*f) <  DENORMAL_CLEAN_THRESHOLD) *f = 0.0; f++;    
    if (fabs(*f) <  DENORMAL_CLEAN_THRESHOLD) *f = 0.0; f++;    
    if (fabs(*f) <  DENORMAL_CLEAN_THRESHOLD) *f = 0.0; f++;    
    if (fabs(*f) <  DENORMAL_CLEAN_THRESHOLD) *f = 0.0; f++;    
    if (fabs(*f) <  DENORMAL_CLEAN_THRESHOLD) *f = 0.0; f++;    

    __asm {
        fld dword ptr fpu_regs[28]
        fld dword ptr fpu_regs[24]
        fld dword ptr fpu_regs[20]
        fld dword ptr fpu_regs[16]
        fld dword ptr fpu_regs[12]
        fld dword ptr fpu_regs[8]
        fld dword ptr fpu_regs[4]
        fld dword ptr fpu_regs[0]

        FCLEX ; clear fpu exceptions
    }

}




static float BeatsPerMinute = 120.0f;

// solo mode handling
static int SoloInstrumentIndex = 0;
static int Solo = 0;

// stream structures for recording sound
static DWORD samplesProcessed = 0;
static bool Recording = false;
static bool FirstRecordingEvent = false;
static DWORD PatternSize = 16;
static float SamplesPerTick;
static float TickScaler = 1.0f;
static DWORD MaxTicks;
static int InstrumentOn[MAX_INSTRUMENTS]; 
static std::vector<int> InstrumentRecord[MAX_INSTRUMENTS];
static std::vector<int> ReducedPatterns;
static std::vector<int> PatternIndices[MAX_INSTRUMENTS];
static int NumReducedPatterns = 0;


/* bool UnitTypeHasDelayBuffer(int type)
{
    return ((type == M_DLL) || (type == M_SLW));
} */

#define UnitTypeHasDelayBuffer(type) ((type == M_DLL) || (type == M_SLW))


/* bool UnitHasDelayBuffer(SOME_UNIT_valP v)
{
    return UnitTypeHasDelayBuffer(v->id);
} */

#define UnitHasDelayBuffer(v) (UnitTypeHasDelayBuffer(*(BYTE *)(v)))

bool UnitHasReverb(SOME_UNIT_valP v)
{
    if (v->id == M_DLL)
    {
        DLL_valP dll_v = (DLL_valP)v;
        return (dll_v->reverb);
    }
    else
        return false;
}

bool UnitHasNoteSync(SOME_UNIT_valP v)
{
    if (v->id == M_DLL)
    {
        DLL_valP dll_v = (DLL_valP)v;
        return (dll_v->synctype == 2);
    }
    else
        return false;
}


int GetModulationLinkInstrument(SOME_UNIT_valP v)
{
    if (v->id == M_FSTx)
        return ((FSTx_valP)(v))->dest_stack;
    else if (v->id == M_FLDFROM)
        return ((FLDFROM_valP)(v))->source_stack;
    else 
    {
        MessageBox(0, "GetModulationLinkInstrument(): unknown unit id.", "Error", MB_OK | MB_SETFOREGROUND);
        return -1;
    }
}

char GetModulationLinkUnit(SOME_UNIT_valP v)
{
    if (v->id == M_FSTx)
        return ((FSTx_valP)(v))->dest_unit;
    else if (v->id == M_FLDFROM)
        return ((FLDFROM_valP)(v))->source_unit;
    else 
    {
        MessageBox(0, "GetModulationLinkUnit(): unknown unit id.", "Error", MB_OK | MB_SETFOREGROUND);
        return -1;
    }
}

char GetModulationLinkSlot(SOME_UNIT_valP v)
{
    if (v->id == M_FSTx)
        return ((FSTx_valP)(v))->dest_slot;
    else if (v->id == M_FLDFROM)
        return ((FLDFROM_valP)(v))->source_slot;
    else 
    {
        MessageBox(0, "GetModulationLinkSlot(): unknown unit id.", "Error", MB_OK | MB_SETFOREGROUND);
        return -1;
    }
}

char GetModulationLinkID(SOME_UNIT_valP v)
{
    if (v->id == M_FSTx)
        return ((FSTx_valP)(v))->dest_id;
    else if (v->id == M_FLDFROM)
        return ((FLDFROM_valP)(v))->source_id;
    else 
    {
        MessageBox(0, "GetModulationLinkID(): unknown unit id.", "Error", MB_OK | MB_SETFOREGROUND);
        return -1;
    }
}


void SetModulationLinkUnit(SOME_UNIT_valP v, char value)
{
    if (v->id == M_FSTx)
        ((FSTx_valP)(v))->dest_unit = value;
    else if (v->id == M_FLDFROM)
        ((FLDFROM_valP)(v))->source_unit = value;
    else
        MessageBox(0, "SetModulationLinkUnit(): unknown unit id.", "Error", MB_OK | MB_SETFOREGROUND);
}

void SetModulationLinkSlot(SOME_UNIT_valP v, char value)
{
    if (v->id == M_FSTx)
        ((FSTx_valP)(v))->dest_slot = value;
    else if (v->id == M_FLDFROM)
        ((FLDFROM_valP)(v))->source_slot = value;
    else
        MessageBox(0, "SetModulationLinkSlot(): unknown unit id.", "Error", MB_OK | MB_SETFOREGROUND);
}

void SetModulationLinkID(SOME_UNIT_valP v, char value)
{
    if (v->id == M_FSTx)
        ((FSTx_valP)(v))->dest_id = value;
    else if (v->id == M_FLDFROM)
        ((FLDFROM_valP)(v))->source_id = value;
    else 
    {
        MessageBox(0, "SetModulationLinkID(): unknown unit id.", "Error", MB_OK | MB_SETFOREGROUND);
    }
}



void MaskDenormalControlBit()
{
short int x;
__asm FSTCW     x
x |= 0x02;          // to mask the Denormal control bit
__asm FLDCW     x
}

void UnmaskDenormalControlBit()
{
short int x;
__asm FSTCW     x
x &=  0xFFFD;       // to unmask the Denormal control bit
__asm FLDCW     x
}


// init synth
void Go4kVSTi_Init()
{
	static bool initialized = false;
	// do one time initialisation here (e.g. wavtable generation ...)
	if (!initialized)
	{                
        // UnmaskDenormalControlBit();

		memset(&SynthObj, 0, sizeof(SynthObj));
        memset(&GlobalSettings, 0, sizeof(GlobalSettings));

        SynthObj.synth_mutex = CreateMutex(NULL, FALSE, NULL);        

        // Initialize the synth 

		BeatsPerMinute = 120.0f;
		SoloInstrumentIndex = 0;
		Solo = 0;
		Go4kVSTi_ResetPatch();
		initialized = true;
	}
}


void Go4kVSTi_Destroy()
{
    CloseHandle(SynthObj.synth_mutex);
}


// reset synth

void Go4kVSTi_ClearInstrumentSlot(int instrument_index, int slot)
{
	memset(SynthObj.InstrumentValues[instrument_index][slot], 0, MAX_SLOT_VALUES);
	for (int i = 0; i < MAX_POLYPHONY; i++)
	{
		float* w = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+i].unit_workspace[slot*MAX_SLOT_VALUES]);
		memset(w, 0, MAX_SLOT_VALUES*4);
	}
}

void Go4kVSTi_ClearInstrumentWorkspace(int instrument_index)
{
	// clear workspace
	InstrumentWorkspace* w = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY]);
	memset(w, 0, sizeof(InstrumentWorkspace)*MAX_POLYPHONY);
}

void ResetInstrumentInputSettings(int instrument_index)
{
    // Set default instrument input settings
    InstrumentInputSettingsP i = &(SynthObj.InstrumentInputSettings[instrument_index]);
    // Channels 16+ obviously won't be able to receive MIDI input before they are configured to channels 0-15
    if (instrument_index > 15) 
        i->MIDIChannel = 15;
    else
        i->MIDIChannel = instrument_index;
    i->NoteLo = 0; // C-0
    i->NoteHi = 127; // G-10
    i->VelocityLo = 1; // velocity 0 means "note off"... we could do a "release trigger" in the VSTi by hooking velocity 0, but it wouldn't be straight-forward to implement in the asm player
    i->VelocityHi = 127;
    memset(i->DummyReservedSpace, 0, sizeof(i->DummyReservedSpace));
}

// set reset_all_delay_buffers to false when mass-resettings things, because otherwise
// you'll be wiping the same memory dozens of times
void Go4kVSTi_ResetInstrument(int instrument_index, bool reset_all_delay_buffers)
{
	char name[128];
	sprintf(name, "Instr %d", instrument_index+1);
    int namelengthincludingterminator = strlen(name) + 1;
    if (namelengthincludingterminator > 63) namelengthincludingterminator = 63;
    if (namelengthincludingterminator > 0)
	    memcpy(SynthObj.InstrumentNames[instrument_index], name, namelengthincludingterminator);
    SynthObj.InstrumentNames[instrument_index][namelengthincludingterminator] = 0;

    // BYTE* v = SynthObj.InstrumentMIDISettings[instrument_index];

	// clear values
	BYTE* v = SynthObj.InstrumentValues[instrument_index][0];
	memset(v, 0, MAX_UNITS*MAX_SLOT_VALUES);

	// clear workspace
	InstrumentWorkspace* w = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY]);
	memset(w, 0, sizeof(InstrumentWorkspace)*MAX_POLYPHONY);

	// set default units
	Go4kVSTi_InitSlot(SynthObj.InstrumentValues[instrument_index][0], instrument_index, M_ENV);
	Go4kVSTi_InitSlot(SynthObj.InstrumentValues[instrument_index][1], instrument_index, M_VCO);
	Go4kVSTi_InitSlot(SynthObj.InstrumentValues[instrument_index][2], instrument_index, M_FOP); ((FOP_valP)(SynthObj.InstrumentValues[instrument_index][2]))->flags = FOP_MULP;
	//Go4kVSTi_InitSlot(SynthObj.InstrumentValues[instrument_index][22], instrument_index, M_DLL);
	Go4kVSTi_InitSlot(SynthObj.InstrumentValues[instrument_index][23], instrument_index, M_PAN);
	Go4kVSTi_InitSlot(SynthObj.InstrumentValues[instrument_index][24], instrument_index, M_OUT);

	SynthObj.InstrumentSignalValid[instrument_index] = 1;
	SynthObj.SignalTrace[instrument_index] = 0.0f;
	SynthObj.ControlInstrument[instrument_index] = 0;
	SynthObj.VoiceIndex[instrument_index] = 0;

    ResetInstrumentInputSettings(instrument_index);

    SynthObj.HeldNote[instrument_index] = -1;

    if (reset_all_delay_buffers)
	    Go4kVSTi_ClearDelayLinesAndSlews();
} // Go4kVSTi_ResetInstrument()

void Go4kVSTi_ClearGlobalSlot(int slot)
{
	memset(SynthObj.GlobalValues[slot], 0, MAX_SLOT_VALUES);
	float* w = &(SynthObj.GlobalWork.unit_workspace[slot*MAX_SLOT_VALUES]);
	memset(w, 0, MAX_SLOT_VALUES*4);
}

void Go4kVSTi_ClearGlobalWorkspace()
{
	// clear workspace
	memset(&(SynthObj.GlobalWork), 0, sizeof(InstrumentWorkspace));
}

void Go4kVSTi_ResetGlobalUnits(bool reset_all_delay_buffers)
{
	// clear values
	memset(SynthObj.GlobalValues, 0, sizeof(SynthObj.GlobalValues));

	// clear workspace
	memset(&(SynthObj.GlobalWork), 0, sizeof(InstrumentWorkspace));

	// set default units
	Go4kVSTi_InitSlot(SynthObj.GlobalValues[0], MAX_INSTRUMENTS, M_ACC); ((ACC_valP)(SynthObj.GlobalValues[0]))->flags = ACC_AUX;
	Go4kVSTi_InitSlot(SynthObj.GlobalValues[1], MAX_INSTRUMENTS, M_DLL); 
		((DLL_valP)(SynthObj.GlobalValues[1]))->reverb = 1; 
		((DLL_valP)(SynthObj.GlobalValues[1]))->leftreverb = 1; 
		((DLL_valP)(SynthObj.GlobalValues[1]))->feedback = 125; 
		((DLL_valP)(SynthObj.GlobalValues[1]))->pregain = 40; 
	Go4kVSTi_InitSlot(SynthObj.GlobalValues[2], MAX_INSTRUMENTS, M_FOP); ((FOP_valP)(SynthObj.GlobalValues[2]))->flags = FOP_XCH;
	Go4kVSTi_InitSlot(SynthObj.GlobalValues[3], MAX_INSTRUMENTS, M_DLL);
		((DLL_valP)(SynthObj.GlobalValues[3]))->reverb = 1; 
		((DLL_valP)(SynthObj.GlobalValues[3]))->leftreverb = 0; 
		((DLL_valP)(SynthObj.GlobalValues[3]))->feedback = 125; 
		((DLL_valP)(SynthObj.GlobalValues[3]))->pregain = 40; 
	Go4kVSTi_InitSlot(SynthObj.GlobalValues[4], MAX_INSTRUMENTS, M_FOP); ((FOP_valP)(SynthObj.GlobalValues[4]))->flags = FOP_XCH;
	Go4kVSTi_InitSlot(SynthObj.GlobalValues[5], MAX_INSTRUMENTS, M_ACC); 
	Go4kVSTi_InitSlot(SynthObj.GlobalValues[6], MAX_INSTRUMENTS, M_FOP); ((FOP_valP)(SynthObj.GlobalValues[6]))->flags = FOP_ADDP2;
	Go4kVSTi_InitSlot(SynthObj.GlobalValues[24], MAX_INSTRUMENTS, M_OUT); 
	
	SynthObj.GlobalSignalValid = 1;

    if (reset_all_delay_buffers)
	    Go4kVSTi_ClearDelayLinesAndSlews();
} // Go4kVSTi_ResetGlobalUnits()

void Go4kVSTi_ResetGlobalSettings()
{
	PatternSize = 16;
	SynthObj.Polyphony = 1;
    GlobalSettings.ClipOutput = 1;
    GlobalSettings.NoteQuantization = 2;
    GlobalSettings.Output16Bit = 1;
    GlobalSettings.RecordEnvs = 0;
    GlobalSettings.RecordNotes = 0;
    GlobalSettings.SwingStrength = 50;
    GlobalSettings.Undenormalize = 0;
    GlobalSettings.PatternSize = 16;
    GlobalSettings.ObjectFormat = 0;
    GlobalSettings.AutoPanicMask = AUTO_PANIC_ON_INVALID_OPERATION;
    GlobalSettings.AutoPanicDoPanic = 1;
} // Go4kVSTi_ResetGlobalSettings()

// reset synth
void Go4kVSTi_ResetPatch()
{
	for (int i = 0; i < MAX_INSTRUMENTS; i ++)
	{
		Go4kVSTi_ResetInstrument(i, false);
	}
	// reset global thingies
	Go4kVSTi_ResetGlobalUnits(true);
    Go4kVSTi_ResetGlobalSettings();
}

void Go4kVSTi_FlipInstrumentSlots(int instrument_index, int a, int b)
{
	DWORD temp[MAX_SLOT_VALUES];
	BYTE* v1 = SynthObj.InstrumentValues[instrument_index][a];
	BYTE* v2 = SynthObj.InstrumentValues[instrument_index][b];
	memcpy(temp, v2, MAX_SLOT_VALUES);
	memcpy(v2, v1, MAX_SLOT_VALUES);
	memcpy(v1, temp, MAX_SLOT_VALUES);
	for (int i = 0; i < MAX_POLYPHONY; i++)
	{
		float* w1 = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+i].unit_workspace[a*MAX_SLOT_VALUES]);
		float* w2 = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+i].unit_workspace[b*MAX_SLOT_VALUES]);
		memcpy(temp, w2, MAX_SLOT_VALUES*4);
		memcpy(w2, w1, MAX_SLOT_VALUES*4);
		memcpy(w1, temp, MAX_SLOT_VALUES*4);
	}
	// reset dll workspaces, they are invalid now
	if (UnitHasDelayBuffer((SOME_UNIT_valP)(v1)) || UnitHasDelayBuffer((SOME_UNIT_valP)(v2))) // tss oli alun perin &&, oli luultavasti bugi
	{
		Go4kVSTi_ClearDelayLinesAndSlews();
        Go4kVSTi_UpdateDelayAndSlewTimes();
	}
} // Go4kVSTi_FlipInstrumentSlots()

void Go4kVSTi_FlipGlobalSlots(int a, int b)
{
	DWORD temp[MAX_SLOT_VALUES];
	BYTE* v1 = SynthObj.GlobalValues[a];
	BYTE* v2 = SynthObj.GlobalValues[b];
	memcpy(temp, v2, MAX_SLOT_VALUES);
	memcpy(v2, v1, MAX_SLOT_VALUES);
	memcpy(v1, temp, MAX_SLOT_VALUES);
	for (int i = 0; i < MAX_POLYPHONY; i++)
	{
		float* w1 = &(SynthObj.GlobalWork.unit_workspace[a*MAX_SLOT_VALUES]);
		float* w2 = &(SynthObj.GlobalWork.unit_workspace[b*MAX_SLOT_VALUES]);
		memcpy(temp, w2, MAX_SLOT_VALUES*4);
		memcpy(w2, w1, MAX_SLOT_VALUES*4);
		memcpy(w1, temp, MAX_SLOT_VALUES*4);
	}
	// reset dll workspaces, they are invalid now
	if (UnitHasDelayBuffer((SOME_UNIT_valP)(v1)) || UnitHasDelayBuffer((SOME_UNIT_valP)(v2))) // tss oli alun perin &&, lie bugi?
	{
		Go4kVSTi_ClearDelayLinesAndSlews();
		Go4kVSTi_UpdateDelayAndSlewTimes();
	}
}

// init a unit slot
void Go4kVSTi_InitSlot(BYTE* slot, int instrument_index, int type)
{
		// set default values
	slot[0] = type;
	if (type == M_NONE)
	{
    }
	else if (type == M_ENV)
	{
		ENV_valP v = (ENV_valP)slot;
		v->attac		=  64;
		v->decay		=  64;
		v->sustain		=  64;
		v->release		=  64;
		v->gain			= 128;
	}
	else if (type == M_VCO)
	{
		VCO_valP v = (VCO_valP)slot;
		v->transpose	=  64;
		v->detune		=  64;
		v->phaseofs		=   0;
		v->gain			=0x55;
		v->color		=  64;
		v->shape		=  64;
		v->gain			= 128;
		v->flags		= VCO_SINE;
	}
	else if (type == M_VCF)
	{
		VCF_valP v = (VCF_valP)slot;
		v->freq			=  64;
		v->res			=  64;
		v->type			= VCF_LOWPASS;
	}
	else if (type == M_DST)
	{
		DST_valP v = (DST_valP)slot;
		v->drive		=  64;
		v->snhfreq		= 128;
	}
	else if (type == M_DLL)
	{
		DLL_valP v = (DLL_valP)slot;
		v->reverb		=   0;
		v->delayidx		=   0;
		v->count		=   1;
		v->pregain		=  64;
		v->dry			=  128;
		v->feedback		=  64;
		v->damp			=  64;
		v->guidelay		=  40;
		v->synctype		=   1;
		v->leftreverb	=   0;
		v->depth		=	0;
		v->freq			=	0;
	}
	else if (type == M_FOP)
	{
		FOP_valP v = (FOP_valP)slot;
		v->flags		= FOP_MULP;
	}
	else if (type == M_FSTx)
	{
		FSTx_valP v = (FSTx_valP)slot;
		v->amount		=  64;
        v->operation         = FST_OPERATION_SET;
		v->dest_stack	= -1;
		v->dest_unit	= -1;
		v->dest_slot	= -1;
		v->dest_id		= -1;
	}
	else if (type == M_PAN)
	{
		PAN_valP v = (PAN_valP)slot;
		v->panning		=  64;
	}
	else if (type == M_OUT)
	{
		OUT_valP v = (OUT_valP)slot;
		v->gain			=  64;
		v->auxsend		=   0;
	}
	else if (type == M_ACC)
	{
		ACC_valP v = (ACC_valP)slot;
		v->flags		= ACC_OUT;
	}
	else if (type == M_FLD)
	{
		FLD_valP v = (FLD_valP)slot;
		v->value		=  64;
	}
	else if (type == M_SLW)
    {		
		SLW_valP v = (SLW_valP)slot;
		v->delayidx		=   0;
        v->calcmode		=   SLW_CALCMODE_LINEAR;
		v->guidelay		=  40;
        v->synctype		=   SLW_SYNCTYPE_BPM;
    }
	else if (type == M_FLDFROM)
    {		
		FLDFROM_valP v = (FLDFROM_valP)slot;
        v->amount		=  64;
        v->source_stack = -1;
		v->source_unit	= -1;
		v->source_slot	= -1;
		v->source_id	= -1;
    }

} // Go4kVSTi_InitSlot()


// init a instrument slot i.e. unit
void Go4kVSTi_InitInstrumentSlot(int instrument_index, int s, int type)
{
	// clear values and workspace
	Go4kVSTi_ClearInstrumentSlot(instrument_index, s);
	// init with default values
	Go4kVSTi_InitSlot(SynthObj.InstrumentValues[instrument_index][s], instrument_index, type);
	if (UnitTypeHasDelayBuffer(type))
	{
		Go4kVSTi_ClearDelayLinesAndSlews();
		Go4kVSTi_UpdateDelayAndSlewTimes();
	}
}

// init a global slot
void Go4kVSTi_InitGlobalSlot(int s, int type)
{
	// clear values and workspace
	Go4kVSTi_ClearGlobalSlot(s);
	// init with default values
	Go4kVSTi_InitSlot(SynthObj.GlobalValues[s], MAX_INSTRUMENTS, type);
	if (UnitTypeHasDelayBuffer(type))
	{
		Go4kVSTi_ClearDelayLinesAndSlews();
		Go4kVSTi_UpdateDelayAndSlewTimes();
	}
}

// panic
void Go4kVSTi_Panic()
{
	for (int i = 0; i < MAX_INSTRUMENTS; i++)
	{
		// clear workspace
		InstrumentWorkspace* w = &(SynthObj.InstrumentWork[i*MAX_POLYPHONY]);
		memset(w, 0, sizeof(InstrumentWorkspace)*MAX_POLYPHONY);
		SynthObj.SignalTrace[i] = 0.0f;
        SynthObj.HeldNote[i] = -1;
	}
	// clear workspace
	memset(&(SynthObj.GlobalWork), 0, sizeof(InstrumentWorkspace));
	Go4kVSTi_ClearDelayLinesAndSlews();
    SynthObj.PanicCount++;
}

static float delayTimeFraction[33] = 
{	
	4.0f * (1.0f/32.0f) * (2.0f/3.0f),
	4.0f * (1.0f/32.0f),
	4.0f * (1.0f/32.0f) * (3.0f/2.0f),
	4.0f * (1.0f/16.0f) * (2.0f/3.0f),
	4.0f * (1.0f/16.0f),
	4.0f * (1.0f/16.0f) * (3.0f/2.0f),
	4.0f * (1.0f/8.0f) * (2.0f/3.0f),
	4.0f * (1.0f/8.0f),
	4.0f * (1.0f/8.0f) * (3.0f/2.0f),
	4.0f * (1.0f/4.0f) * (2.0f/3.0f),
	4.0f * (1.0f/4.0f),
	4.0f * (1.0f/4.0f) * (3.0f/2.0f),
	4.0f * (1.0f/2.0f) * (2.0f/3.0f),
	4.0f * (1.0f/2.0f),
	4.0f * (1.0f/2.0f) * (3.0f/2.0f),
	4.0f * (1.0f) * (2.0f/3.0f),
	4.0f * (1.0f),
	4.0f * (1.0f) * (3.0f/2.0f),
	4.0f * (2.0f) * (2.0f/3.0f),
	4.0f * (2.0f),
	4.0f * (2.0f) * (3.0f/2.0f),
	4.0f * (3.0f/8.0f),
	4.0f * (5.0f/8.0f),
	4.0f * (7.0f/8.0f),
	4.0f * (9.0f/8.0f),
	4.0f * (11.0f/8.0f),
	4.0f * (13.0f/8.0f),
	4.0f * (15.0f/8.0f),
	4.0f * (3.0f/4.0f),
	4.0f * (5.0f/4.0f),
	4.0f * (7.0f/4.0f),
	4.0f * (3.0f/2.0f),
	4.0f * (3.0f/2.0f),
};


int GetMusicalBPMDelayTime(int guidelay, float bpm)
{
	float ftime;
	float quarterlength = 60.0f/bpm;

	ftime = quarterlength*delayTimeFraction[guidelay>>2];
	int delaytime = 44100.0f*ftime;
	if (delaytime >= 65536)
		delaytime = 65535;

    return delaytime;
}

void Go4kVSTi_UpdateDelayAndSlewTimes() // aaa todo k
{
	int delayindex = 17;
	for (int i = 0; i <= MAX_INSTRUMENTS; i++)
	{
		for (int u = 0; u < MAX_UNITS; u++)
		{
			DLL_valP v;
			if (i < MAX_INSTRUMENTS)
				v = (DLL_valP)(SynthObj.InstrumentValues[i][u]);
			else
				v = (DLL_valP)(SynthObj.GlobalValues[u]);

			if (v->id == M_DLL)
			{
//				DLL_valP v = (DLL_valP)(SynthObj.InstrumentValues[i][u]); // <-- this has to be a bug.
				if (v->reverb)
				{
					if (v->leftreverb)
					{
						v->delayidx = 1;
						v->count = 8;
					}
					else
					{
						v->delayidx = 9;
						v->count = 8;
					}
				}
				else
				{
					int delaytime;
					if (v->synctype == DLL_SYNCTYPE_NOTE_PITCH) // note pitch sync
					{
						v->delayidx = 0;
						v->count = 1;
					}
					else
					{
						if (v->synctype == DLL_SYNCTYPE_BPM) // musical bpm sync
						{
                            delaytime = GetMusicalBPMDelayTime(v->guidelay, Go4kVSTi_GetBPM());
						}
						else 
						{ // no sync, absolute delay time value
							delaytime = v->guidelay*16;
						}
						(&go4k_delay_times)[delayindex] = delaytime;
						v->delayidx = delayindex;
						v->count = 1;
					}
					delayindex++;
				}
			} // M_DLL
            else if (v->id == M_SLW)
            {
				int delaytime;

                SLW_valP slw = (SLW_valP)(v);
                if (slw->synctype == SLW_SYNCTYPE_BPM)
                {
                    delaytime = GetMusicalBPMDelayTime(slw->guidelay, Go4kVSTi_GetBPM());
                }
                else
                    delaytime = slw->guidelay*16;
                
				(&go4k_delay_times)[delayindex] = delaytime;
				slw->delayidx = delayindex;				
                slw->delaytimesmp = delaytime;

				delayindex++;
            }
		}
	} // for MAX_INSTRUMENTS
} // void Go4kVSTi_UpdateDelayAndSlewTimes()

// clear delay lines
void Go4kVSTi_ClearDelayLinesAndSlews()
{       
    memset((&go4k_delay_buffer), 0, go4k_delay_buffer_size); // baaa todo what is this "80"???
	//memset((&go4k_delay_buffer), 0, 80*((65536+4)*4)); // baaa todo what is this "80"???
}


// clear delay lines
void Go4kVSTi_ClearDelaySingleLineOrSlewBuffer()
{
//	memset((&go4k_delay_buffer), 0, 80*((65536+4)*4)); // baaa todo what is this "80"???
}


// set global bpm
void Go4kVSTi_SetBPM(float bpm)
{
	BeatsPerMinute = bpm;
	LFO_NORMALIZE = bpm/(44100.0*60.0);
	Go4kVSTi_UpdateDelayAndSlewTimes();
}

// get bpm
float Go4kVSTi_GetBPM()
{
	return BeatsPerMinute;
}

// enable solo mode for a single instrument only
void Go4kVSTi_Solo(int instrument_index, int solo)
{
	if (solo)
	{
		SoloInstrumentIndex = instrument_index;
		Solo = true;
	}
	else
	{
		Solo = false;
	}
}

// sample times tick the whole synth pipeline. results are left and right output sample

bool GetSynthMutex()
{
    DWORD r = WaitForSingleObject(SynthObj.synth_mutex, 2000);
    return (r == WAIT_OBJECT_0);
}

bool ReleaseSynthMutex()
{
    return (ReleaseMutex(SynthObj.synth_mutex));
}

#define SIGNAL_INACTIVE_THRESHOLD (0.00001f)
// #define SIGNAL_INACTIVE_THRESHOLD (0.8f)

// the "tick" here is not the same length in samples as the "tick" in the song
void Go4kVSTi_Tick(float *oleft, float *oright, int samples)
{
    unsigned int num_delay_buffer_units_for_instrument[MAX_INSTRUMENTS];
    bool instrument_signal_active[MAX_INSTRUMENTS];

    if (!GetSynthMutex())
        return;
    
    //SOME_UNIT_valP delay_buffer_units[MAX_UNITS*MAX_INSTRUMENTS];

	// Pre-count how many delay buffers each instrument has, so we don't have to run this loop for every sample.
    // It saves around 1500 for loop iterations per sample. This could be sped up further by keeping a separate list
    // of delay units (and all other types of units).
    for (int i = 0; i < MAX_INSTRUMENTS; i++) 
    {
        num_delay_buffer_units_for_instrument[i] = 0;
	    for (int u = 0; u < MAX_UNITS; u++)
	    {            
		    BYTE* val = SynthObj.InstrumentValues[i][u];
		    if (UnitHasDelayBuffer((SOME_UNIT_valP)(val))) 
                num_delay_buffer_units_for_instrument[i]++;
	    }
        instrument_signal_active[i] = false;
    }

    bool fpu_problem_occurred = false;

	// do as many samples as requested
	int sample_counter = 0; // I have to use long_variable_names, because I can't make sense of what is what if there are multiple nested blocks with variables "s", "u", "i" etc 
	while (sample_counter < samples)
	{
		float left=0.0f;
		float right=0.0f;
		
        // __asm fninit;

        // reset transformed values cursor to start of buffer
        go4k_transformed_values_cursor = (float *)(&go4k_transformed_values_buffer);

        go4k_transformed_values_update_counter = sample_counter;

        // reset delay buffer cursor to start of buffer
		go4k_delay_buffer_ofs = (DWORD)(&go4k_delay_buffer);

        // bool this_instrument_panic = false;

		// loop all instruments
		for (int instrument_index = 0; instrument_index < MAX_INSTRUMENTS; instrument_index++)
		{
			// solo mode and not the instrument we want soloed?
			if (Solo && instrument_index != SoloInstrumentIndex)
			{
				// loop all voices and clear outputs
				for (int voice_counter = 0; voice_counter < SynthObj.Polyphony; voice_counter++)
				{
					InstrumentWorkspaceP iwork = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+voice_counter]);
					iwork->dlloutl = 0.0f;
					iwork->dlloutr = 0.0f;
					iwork->outl = 0.0f;
					iwork->outr = 0.0f;
				}
				// adjust delay index
				for (int unit_index = 0; unit_index < MAX_UNITS; unit_index++)
				{		
					BYTE* val = SynthObj.InstrumentValues[instrument_index][unit_index];
                    if (UnitHasDelayBuffer((SOME_UNIT_valP)(val)))
						go4k_delay_buffer_ofs += (5+65536)*4*SynthObj.Polyphony;
				}
				// go to next instrument
				continue;
			}
			// if the instrument signal stack is valid and we still got a signal from that instrument
			//if (SynthObj.InstrumentSignalValid[instrument_index]  && (fabs(SynthObj.SignalTrace[instrument_index]) > SIGNAL_INACTIVE_THRESHOLD) )
            if (SynthObj.InstrumentSignalValid[instrument_index]  && (SynthObj.SignalTrace[instrument_index] != 0.0) ) // signal trace is cut to 0.0, so we can compare with !=
			{
				float sumSignals_of_this_instrument = 0.0f;
                if ((SynthObj.SignalTrace[instrument_index] > 0.0) && (SynthObj.SignalTrace[instrument_index] < 0.00001))
                    sumSignals_of_this_instrument = 0.0f;
				// loop all voices
				for (int voice_index = 0; voice_index < SynthObj.Polyphony; voice_index++)
				{
					InstrumentWorkspaceP iwork = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+voice_index]);
					float *i_w_unit_workspace = iwork->unit_workspace;
					DWORD inote = iwork->note;
					// loop each slot
					for (int unit_index = 0; unit_index < MAX_UNITS; unit_index++)
					{		
						BYTE* val = SynthObj.InstrumentValues[instrument_index][unit_index];
						float *wrk = &(iwork->unit_workspace[unit_index*MAX_SLOT_VALUES]);
						if (val[0] == M_FSTx) // && iwork->note
						{
							FSTx_valP v = (FSTx_valP)val;
							// if a target slot is set
							if (v->dest_slot != -1)
							{
								InstrumentWorkspaceP mwork;
								int polyphonicStore = SynthObj.Polyphony;
								int stack = v->dest_stack;
								// local storage?
								if (stack == -1 || stack == instrument_index)
								{
									// only store the sample in the current workspace
									polyphonicStore = 1;
									mwork = iwork;
								}
								else if (stack == MAX_INSTRUMENTS)
									mwork = &(SynthObj.GlobalWork);
								else
									mwork = &(SynthObj.InstrumentWork[stack*MAX_POLYPHONY]);
								
								float* mdest = &(mwork->unit_workspace[v->dest_unit*MAX_SLOT_VALUES + v->dest_slot]);
								float amount = (2.0f*v->amount - 128.0f)*0.0078125f;
                                int operation = v->operation;
                                int condition = v->condition;

                                // Pointer to this unit's workspace for accessing and updating the previn value
                                FSTx_wrkP fst_work_previn = (FSTx_wrkP)(&iwork->unit_workspace[unit_index*MAX_SLOT_VALUES+0]);

								for (int stc = 0; stc < polyphonicStore; stc++)
								{									
									__asm
									{
										push	eax
										push	ebx
							                                        
										mov		ebx, operation

                                        mov     eax, fst_work_previn

                                        test    [condition], 0xFFFFFFFF

                                        jz  fstx_condition_is_met

	                                    fld     dword ptr [eax]          ; previn   input                             
	                                    fucomi  st(0), st(1) // carry flag is set if st0 < st1                                        
	                                    fstp	st(0) // pop prevout value from fpu stack, we don't need it                                        
                                        fst     dword ptr [eax] // store back new "prevout"... it's actually "previous input", TODO fix naming
                                        jc      fstx_condition_not_met

                        fstx_condition_is_met:
										fld		amount
										fmul	st(0), st(1)

										test	ebx, FST_OPERATION_ADD

                                        mov		eax, mdest  // this is a direct pointer to the actual destination float

										jz		fstx_condstore_func_set
										fadd	dword ptr [eax] 
									fstx_condstore_func_set:
										fstp 	dword ptr [eax]
									fstx_condstore_func_done:

                                fstx_condition_not_met:

										pop		ebx
										pop		eax
									} // asm


									mdest += sizeof(InstrumentWorkspace)/4; // go to next voice
								} // for polyphonic store

								// remove signal on pop flag
								if (operation & FST_OPERATION_POP)
								{
									_asm  fstp	st(0);
								}

							}
						} // if (val[0] == M_FSTx)
						else if (val[0] == M_FLDFROM) // && iwork->note
						{
							FLDFROM_valP v = (FLDFROM_valP)val;
							// if a target slot is set
							if (v->source_slot != -1)
							{
								InstrumentWorkspaceP mwork;
								// int polyphonicModulationLink = SynthObj.Polyphony;
								int stack = v->source_stack;
								// local storage?
								if (stack == -1 || stack == instrument_index)
								{
									// only load the sample from the current workspace (/voice/stack)
                                    // any idea what should be done with polyphonic cross-voice "load from"?
									// polyphonicModulationLink = 1;
									mwork = iwork;
								}
								else if (stack == MAX_INSTRUMENTS)
									mwork = &(SynthObj.GlobalWork);
								else
									mwork = &(SynthObj.InstrumentWork[stack*MAX_POLYPHONY]);
								
								float* msource = &(mwork->unit_workspace[v->source_unit*MAX_SLOT_VALUES + v->source_slot]);
								// float amount = (2.0f*v->amount - 128.0f)*0.0078125f;                                
								//for (int stc = 0; stc < polyphonicModulationLink; stc++)
								//{									
									__asm
									{
										push	eax
										mov		eax, msource
										fld	    dword ptr [eax]
										// fld		amount
										// fmul	st(0), st(1)							
										pop		eax
									}

									// msource += sizeof(InstrumentWorkspace)/4;
								// }
							}
						} // if (val[0] == M_FLDFROM)
                        else
						{ // "ordinary" unit handled by one of 4klang.asm functions

                            // IN		:		WRK	(ebp) = this unit's workspace
                            // IN		:		VAL	(esi) = this unit's values
                            // IN		:		ecx	= this instrument's workspace (was called "global workspace" in 4klang's comments)

							if (val[0]) // if non-empty slot
							{
                                int unit_fpu_problem_alert=0;

								// set up and call synth core func
								__asm
								{
									pushad
									xor		eax, eax
									mov		esi, val
									lodsb
									mov		eax, dword ptr [SynthFuncs+eax*4]
									mov		ebx, inote
									mov		ecx, i_w_unit_workspace
									mov		ebp, wrk

                                    FCLEX ; clear fpu exceptions

									call	eax						

									popad
                                } // __asm

                                if (GlobalSettings.AutoPanicMask)
                                    __asm
                                    {
                                    // check if the unit caused an FPU problem
                                        push eax
                                        fstsw   ax
                                        test    al, byte ptr [GlobalSettings.AutoPanicMask]
                                        jz      no_fpu_problem_yeah
                                        // mov     eax, eax // place breakpoint on this line to catch an exception (or denormal or whatever you test for)
                                        mov     dword ptr [unit_fpu_problem_alert], 1
    no_fpu_problem_yeah:              
                                        pop eax
								    } // __asm

                                if (unit_fpu_problem_alert)
                                { 
                                  fpu_problem_occurred = true;
                                  clear_floats(wrk + num_dwords_to_skip_in_the_beginning_of_units_workspace_because_they_are_not_floats[val[0]], num_workspace_floats[val[0]]);
                                  __asm fninit;
                                  SynthObj.InstrumentFPUProblemsCount[instrument_index] += FPU_PROBLEM_INCREMENT;
                                  SynthObj.UnitFPUProblemsCount[instrument_index*MAX_UNITS+unit_index] += FPU_PROBLEM_INCREMENT;
                                }

							} // if (val[0]) i.e. if non-empty slot

						} // if ordinary 4klang.asm unit function
					} // for units in voice

					// check for end of note
					DWORD envstate = *((BYTE*)(i_w_unit_workspace)); // WARNING TODO <--- this assumes that the first unit of every instrument is an ENVelope :(
					if (envstate == ENV_STATE_OFF)
					{
						iwork->note = 0;
					}
                    // if (!fpu_exception_occurred)
					sumSignals_of_this_instrument += fabsf(iwork->outl) + fabsf(iwork->outr) + fabsf(iwork->dlloutl) + fabsf(iwork->dlloutr);
				} // for voice_index, loop all (1 or 2) voices of instrument [instrument_index]

				// update envelope follower only for non control instruments. (1s attack rate) for total instrument signal
				if (SynthObj.ControlInstrument[instrument_index])
					SynthObj.SignalTrace[instrument_index] = 1.0f;
				else
                { // not a control instrument
					SynthObj.SignalTrace[instrument_index] = sumSignals_of_this_instrument + 0.999977324f * ( SynthObj.SignalTrace[instrument_index] - sumSignals_of_this_instrument );
                    if ((fabs(SynthObj.SignalTrace[instrument_index]) < SIGNAL_INACTIVE_THRESHOLD))
                    {
                        SynthObj.SignalTrace[instrument_index] = 0.0;
				        // loop all voices
				        for (int p = 0; p < SynthObj.Polyphony; p++)
				        {
					        InstrumentWorkspaceP iwork = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+p]);
					        iwork->dlloutl = 0.0f;
					        iwork->dlloutr = 0.0f;
					        iwork->outl = 0.0f;
					        iwork->outr = 0.0f;
				        }
                    }
                    else
                        // only active, valid, non-control instruments are summed to the main output, so that's the only case where we raise this flag
                        instrument_signal_active[instrument_index] = true;

                } // if not a control instrument
			} // if (SynthObj.InstrumentSignalValid[instrument_index]  && (fabs(SynthObj.SignalTrace[instrument_index]) > SIGNAL_TRACE_INACTIVE_THRESHOLD) )
			// instrument stack invalid, or no no signal from instrument            
			else
			{
                go4k_delay_buffer_ofs += (5+65536)*4*SynthObj.Polyphony * num_delay_buffer_units_for_instrument[instrument_index];

				// loop all voices
				for (int p = 0; p < SynthObj.Polyphony; p++)
				{
					InstrumentWorkspaceP iwork = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+p]);
					iwork->dlloutl = 0.0f;
					iwork->dlloutr = 0.0f;
					iwork->outl = 0.0f;
					iwork->outr = 0.0f;
				}
                // instrument_signal_active[instrument_index] = false;
			}
		} // for (instrument_index)


		// if the global stack is valid
		if (SynthObj.GlobalSignalValid)
		{
			InstrumentWorkspaceP gwork = &(SynthObj.GlobalWork);
			float *lwrk = gwork->unit_workspace;
			DWORD gnote = 1;
			gwork->note = 1;
			// loop all global slots
			for (int global_unit_idx = 0; global_unit_idx < MAX_UNITS; global_unit_idx++)
			{		
				BYTE* g_val = SynthObj.GlobalValues[global_unit_idx];
				float *wrk = &(lwrk[global_unit_idx*MAX_SLOT_VALUES]);
				// manually accumulate signals
				if (g_val[0] == M_ACC) // do the same as what go4kACC_func does
				{
				    float ACCL = 0.0f;
				    float ACCR = 0.0f;
					ACC_valP av = (ACC_valP)g_val;
					if (av->flags == ACC_OUT)
					{
						for (int instr_idx = 0; instr_idx < MAX_INSTRUMENTS; instr_idx++)
						{
                            if (instrument_signal_active[instr_idx])
                            {
							    for (int p = 0; p < SynthObj.Polyphony; p++)
							    {
                                    ACCL += SynthObj.InstrumentWork[instr_idx*MAX_POLYPHONY+p].outl;
                                    ACCR += SynthObj.InstrumentWork[instr_idx*MAX_POLYPHONY+p].outr;
							    }
                            }
						}
					} // if ACC_OUT accumulator of global unit stack
					else
					{ // (av->flags == ACC_AUX)
						for (int instr_idx = 0; instr_idx < MAX_INSTRUMENTS; instr_idx++)
						{
							for (int p = 0; p < SynthObj.Polyphony; p++)
							{
                                if (instrument_signal_active[instr_idx])
                                {
								    ACCL += SynthObj.InstrumentWork[instr_idx*MAX_POLYPHONY+p].dlloutl;
								    ACCR += SynthObj.InstrumentWork[instr_idx*MAX_POLYPHONY+p].dlloutr;
							    }
                            }
						}
					}
					// push the accumulated signals on the fp stack
					__asm
					{ // "ACCR = 0; fld ACCR" corresponds to "fldz + fadd" in 4klang.asm / go4kACC_func
						fld		ACCR
						fld		ACCL
					}
				}
				// not an ACC unit, check if it's a store then
				else if (g_val[0] == M_FSTx) // && gwork->note
				{
					FSTx_valP g_v = (FSTx_valP)g_val;
					// if a target slot is set
					if (g_v->dest_slot != -1)
					{
						InstrumentWorkspaceP mwork;
						int polyphonicStore = SynthObj.Polyphony;
						int stack = g_v->dest_stack;
						// local storage?
						if (stack == -1 || stack == MAX_INSTRUMENTS)
						{
							// only store the sample in the current workspace
							polyphonicStore = 1;
							mwork = &(SynthObj.GlobalWork);
						}
						else
							mwork = &(SynthObj.InstrumentWork[stack*MAX_POLYPHONY]);
						
						float* mdest = &(mwork->unit_workspace[g_v->dest_unit*MAX_SLOT_VALUES + g_v->dest_slot]);
						float amount = (2.0f*g_v->amount - 128.0f)*0.0078125f;
                        int operation = g_v->operation;
                        int condition = g_v->condition;

					//InstrumentWorkspaceP iwork = &(SynthObj.InstrumentWork[i*MAX_POLYPHONY+p]);
					//	float *wrk = &(iwork->workspace[s*MAX_SLOT_VALUES]);
                                // Pointer to this unit's workspace for accessing and updating the previn value
                    //            FSTx_wrkP fst_work_previn = (FSTx_wrkP)(&iwork->workspace[s*MAX_SLOT_VALUES+0]);

                        // Pointer to this unit's workspace for accessing and updating the previn value
                        FSTx_wrkP fst_work_previn = (FSTx_wrkP)(&SynthObj.GlobalWork.unit_workspace[global_unit_idx*MAX_SLOT_VALUES+0]);

						for (int stc = 0; stc < polyphonicStore; stc++)
						{
								__asm
								{
									push	eax
									push	ebx
						                                        
									mov		ebx, operation

                                    mov     eax, fst_work_previn

                                    test    [condition], 0xFFFFFFFF

                                    jz  fstx_global_condition_is_met

                                    fld     dword ptr [eax]                                       
                                    fucomi  st(0), st(1) // carry flag is set if st0 < st1                                        
                                    fstp	st(0) // pop prevout value from fpu stack, we don't need it                                        
                                    fst     dword ptr [eax] // store back new "prevout"... it's actually "previous input", TODO fix naming
                                    jc      fstx_global_condition_not_met

                    fstx_global_condition_is_met:
									fld		amount
									fmul	st(0), st(1)

									test	ebx, FST_OPERATION_ADD

                                    mov		eax, mdest  // this is a direct pointer to the actual destination float

									jz		fstx_global_func_set
									fadd	dword ptr [eax] 
								fstx_global_func_set:
									fstp 	dword ptr [eax]
								fstx_global_func_done:

                            fstx_global_condition_not_met:

									pop		ebx
									pop		eax
								} // asm

							mdest += sizeof(InstrumentWorkspace)/4;
						} // for polyphonicStore


						// remove signal on pop flag
						if (operation & FST_OPERATION_POP)
						{
							_asm  fstp	st(0);
						}

					}
				}
				// the global stack's unit [global_unit_idx] is not one of the special units, which
                // we handle differently in this VSTi plugin, so call the 4klang.asm function
				else
				{
					if (g_val[0])
					{
                        int global_unit_fpu_problem_alert=0;

						__asm
						{
							pushad
							xor		eax, eax
							mov		esi, g_val
							lodsb
							mov		eax, dword ptr [SynthFuncs+eax*4]
							mov		ebx, gnote
							mov		ecx, lwrk
							mov		ebp, wrk

                            FCLEX ; clear fpu exceptions

							call	eax						
							popad
						}

                        if (GlobalSettings.AutoPanicMask)
                            __asm
                            {
                            // check if the unit caused an FPU problem
                                push eax
                                fstsw   ax
                                test    al, byte ptr [GlobalSettings.AutoPanicMask]
                                jz      no_fpu_problem_in_global_unit_yeah
                                // mov     eax, eax // place breakpoint on this line to catch an exception (or denormal or whatever you test for)
                                mov     dword ptr [global_unit_fpu_problem_alert], 1
no_fpu_problem_in_global_unit_yeah:              
                                pop eax
						    } // __asm

                        if (global_unit_fpu_problem_alert)
                        { 
                          fpu_problem_occurred = true;
                          clear_floats(wrk + num_dwords_to_skip_in_the_beginning_of_units_workspace_because_they_are_not_floats[g_val[0]], num_workspace_floats[g_val[0]]);
                          __asm fninit;
                          SynthObj.InstrumentFPUProblemsCount[MAX_INSTRUMENTS] += FPU_PROBLEM_INCREMENT;
                          SynthObj.UnitFPUProblemsCount[MAX_INSTRUMENTS*MAX_UNITS+global_unit_idx] += FPU_PROBLEM_INCREMENT;
                        }
					} // if g_val[0], non-empty global unit
				} // non-special global unit
			} // for (global_unit_idx)
			left = gwork->outl;
			right = gwork->outr;
		} // if (SynthObj.GlobalSignalValid)

		// clip 
		if (left < -1.0f)
			left = -1.0f;
		if (left > 1.0f)
			left = 1.0f;
		if (right < -1.0f)
			right = -1.0f;
		if (right > 1.0f)
			right = 1.0f;

		*(oleft++) = left;
		*(oright++) = right;

		sample_counter++;

        //go4k_transformed_values_update_counter = s;
        //go4k_transformed_values_update_counter = (go4k_transformed_values_update_counter + 1) & 255;
        //go4k_transformed_values_update_counter = 0;

	} // end sample loop

	if (Recording)
	{
		samplesProcessed += samples;
	}


    // "auto-panic"
    if (fpu_problem_occurred)
    {
        if (GlobalSettings.AutoPanicDoPanic)
        {
            SynthObj.AutoPanicNotificationCounter = 10; // let the GUI show the user that auto-panic was activated        
            Go4kVSTi_Panic();
        }
        // if auto panic is not turned on, then the FPU problems are just displayed
    }

    ReleaseSynthMutex();
} // Go4kVSTi_Tick()

////////////////////////////////////////////////////////////////////////////
// 
//	Synth input processing 
//
////////////////////////////////////////////////////////////////////////////

// prepare for recording the midi stream
void Go4kVSTi_Record(bool record, int patternsize, float patternquant)
{
	Recording = record;
	// if we started recording, clear all record streams
	if (Recording)
	{
		TickScaler = patternquant;
		PatternSize = (int)(patternsize*TickScaler);
		// set first recording event variable to true to enable start time reset on the first occuring event
		FirstRecordingEvent = true;
		// clear all buffers
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			InstrumentOn[i] = -1;
			InstrumentRecord[i].clear();
			for (int j = 0; j < 256*256; j++)
				InstrumentRecord[i].push_back(0);
		}
	}
	else
	{
		// nothing recorded?
		if (FirstRecordingEvent == true)
			return;

		Recording = true;

    	Go4kVSTi_StopVoicesFromAllInstruments();

		Recording = false;
		// how many ticks did we record?
		MaxTicks = (int)(0.5f+((float)samplesProcessed/SamplesPerTick));
		int remainder = MaxTicks % 16;
		MaxTicks += remainder;
		// fold patterns if TickScaler < 1, this is needed because directly recording at half samplerate leads to information loss
		if (TickScaler < 1.0f)
		{	
			MaxTicks = (int)((float)MaxTicks*TickScaler);
			for (int i = 0; i < MAX_INSTRUMENTS; i++)
			{
				for (int j = 0; j < (int)(256*256*TickScaler); j++)
					InstrumentRecord[i][j] = InstrumentRecord[i][(int)(j/TickScaler)];
			}			
		}
		// init reduced patterns
		ReducedPatterns.clear();
		NumReducedPatterns = 0;
		// add NULL pattern
		for (int i = 0; i < PatternSize; i++)
		{
			ReducedPatterns.push_back(0);
		}
		// now add minimal reduced instrument patterns		
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			PatternIndices[i].clear();
			// loop all patterns
			for (int j = 0; j < MaxTicks/PatternSize; j++)
			{
				int pindex = j;
				bool found = false;
				// compare each tick of the current pattern to all already reduced patterns
				for (int k = 0; k < ReducedPatterns.size()/PatternSize; k++)
				{
					bool patternMatch = true;
					for (int l = 0; l < PatternSize; l++)
					{						
						if (InstrumentRecord[i][j*PatternSize+l] != ReducedPatterns[k*PatternSize+l])
						{
							patternMatch = false;
							break;
						}
					}
					if (patternMatch)
					{
						NumReducedPatterns++;
						pindex = k;
						found = true;
						break;
					}
				}
				// add new pattern if necessary
				if (!found)
				{
					for (int k = 0; k < PatternSize; k++)
					{
						ReducedPatterns.push_back(InstrumentRecord[i][j*PatternSize+k]);
					}			
					pindex = (ReducedPatterns.size()/PatternSize)-1;
				}
				// add new pattern index
				PatternIndices[i].push_back(pindex);
			}				
		}

		// open file dialog and save if desired ...
		// the save function called from this one is GoSynth_SaveByteStream
		GetStreamFileName_And_Save();
	}
}

void Go4kVSTi_AddVoiceToInstrument(int instrument_index, int note, int velocity)
{
	InstrumentWorkspaceP work,work2;
	work = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+0]);
	work->release = 1;
	work2 = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+1]);
	work2->release = 1;
	// filp worspace
	if (SynthObj.Polyphony > 1)
	{
		work = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+SynthObj.VoiceIndex[instrument_index]]);
		SynthObj.VoiceIndex[instrument_index] = SynthObj.VoiceIndex[instrument_index] ^ 0x1;
	}
	// add new note	
	memset(work, 0, (2+MAX_UNITS*MAX_SLOT_VALUES)*4); // The "2" are the release and note fields of struct InstrumentWorkspace
    SynthObj.HeldNote[instrument_index] = note; // this is the MIDI key/note held-down from the VST host's point of view, before "velocity to note" translation

    // the work->note is the note pitch value given to the synth, and it stays in the set value
    // for playing the release part of the envelope even after the MIDI key/note is released from the VST host side
    if (SynthObj.InstrumentInputSettings[instrument_index].MapVelocityToNote) 
        work->note = velocity;
    else
	    work->note = note; 
	SynthObj.SignalTrace[instrument_index] = 1.0f;
	// check if its a controll instrument which is played
	SynthObj.ControlInstrument[instrument_index] = 1;
	for (int i = 0; i < MAX_UNITS; i++)
	{
		if (SynthObj.InstrumentValues[instrument_index][i][0] == M_OUT) // if it doesn't have an OUT unit, it's a "control instrument"
		{
			SynthObj.ControlInstrument[instrument_index] = 0;
			break;
		}
	}

	// record song
	if (Recording)
	{
		// check for the first recording event
		if (FirstRecordingEvent)
		{
			FirstRecordingEvent = false;
			// reset time stamp for very first event
			samplesProcessed = 0;
			// Set samples per tick for this recording
			if (TickScaler >= 1.0f)
				SamplesPerTick = 44100.0f*4.0f*60.0f/(BeatsPerMinute*16*TickScaler);
			else
				SamplesPerTick = 44100.0f*4.0f*60.0f/(BeatsPerMinute*16);
		}
		int CurrentTick = (int)(0.5f+((float)samplesProcessed/SamplesPerTick));
		if (InstrumentOn[instrument_index] >= 0)
		{
			for (int i = CurrentTick-1; i > InstrumentOn[instrument_index]; i--)
			{
				if (!InstrumentRecord[instrument_index][i])
					InstrumentRecord[instrument_index][i] = -1;
			}
		}
		if (!InstrumentRecord[instrument_index][CurrentTick])
		{
            if (SynthObj.InstrumentInputSettings[instrument_index].MapVelocityToNote)
			    InstrumentRecord[instrument_index][CurrentTick] = velocity;
            else
                InstrumentRecord[instrument_index][CurrentTick] = note;
			InstrumentOn[instrument_index] = CurrentTick;
		}
	}
} // Go4kVSTi_AddVoiceToInstrument

// add a voice with given parameters to synth
void Go4kVSTi_AddVoice(int channel, int note, int velocity)
{
    for (int i=0; i<MAX_INSTRUMENTS; i++)
    {
        InstrumentInputSettingsP iis = &(SynthObj.InstrumentInputSettings[i]);
        if ((channel == iis->MIDIChannel)
            && (note >= iis->NoteLo)
            && (note <= iis->NoteHi)
            && (velocity >= iis->VelocityLo)
            && (velocity <= iis->VelocityHi)
            )
        {            
            Go4kVSTi_AddVoiceToInstrument(i, note, velocity); // velocity is only used for "map velocity to note"
            return;
        }
    }
}

void Go4kVSTi_StopVoiceFromInstrument(int instrument_index)
{
	InstrumentWorkspaceP work,work2;

    // onko bugi tss vai envelopessa vai miss?

	// release notes
	work = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+0]);
	work->release = 1;
	work2 = &(SynthObj.InstrumentWork[instrument_index*MAX_POLYPHONY+1]);
	work2->release = 1;
	// record song
	if (Recording)
	{
		int CurrentTick = (int)(0.5f+((float)samplesProcessed/SamplesPerTick));
		if (InstrumentOn[instrument_index] >= 0)
		{
			for (int i = CurrentTick-1; i > InstrumentOn[instrument_index]; i--)
			{
				if (!InstrumentRecord[instrument_index][i])
					InstrumentRecord[instrument_index][i] = -1;
			}
		}
//		if (!InstrumentRecord[channel][CurrentTick])
			InstrumentOn[instrument_index] = -1;
	}
    SynthObj.HeldNote[instrument_index] = -1;
} // Go4kVSTi_StopVoiceFromInstrument


void Go4kVSTi_StopVoicesFromAllInstruments()
{
	for (int i = 0; i < MAX_INSTRUMENTS; i++)
		Go4kVSTi_StopVoiceFromInstrument(i);
}

// stop a voice with given parameters in synth
void Go4kVSTi_StopVoice(int channel, int note)
{
    // loop through instruments and find the first one that has a voice on (channel, note) to be stopped
    for (int i=0; i<MAX_INSTRUMENTS; i++)
    {
        InstrumentInputSettingsP iis = &(SynthObj.InstrumentInputSettings[i]);
        if ((channel == iis->MIDIChannel)
            && (SynthObj.HeldNote[i] == note) 
            // We cannot do note-offs based on note-off velocity, because it might differ from note-on velocity (for a number of reasons).
            // So we'll just find the first matching instrument that's listening to the given MIDI channel and has the given note being held.


            //&& (note >= iis->NoteLo)
            //&& (note <= iis->NoteHi)
            //&& ((velocity == 0) || ((velocity >= iis->VelocityLo) && (velocity <= iis->VelocityHi)))
            // && (InstrumentOn[i] >= 0)
            )
        {
            Go4kVSTi_StopVoiceFromInstrument(i);
            return;
        }
    }
}

void Go4kVSTi_StopAllVoicesOnChannel(int channel)
{
    for (int i=0; i<MAX_INSTRUMENTS; i++)
    {
        InstrumentInputSettingsP iis = &(SynthObj.InstrumentInputSettings[i]);
        if ((channel == iis->MIDIChannel)
            // && (InstrumentOn[i] >= 0)
           )
        {
            Go4kVSTi_StopVoiceFromInstrument(i);
        }
    }
}

//----------------------------------------------------------------------------------------------------
// convenience functions for loading and storing patches and instruments
//----------------------------------------------------------------------------------------------------

void FlipSlotModulations(int stack, int unit1, int unit2)
{
	// look in all instruments if a store unit had its target on one of the changed units
	for (int i = 0; i <= MAX_INSTRUMENTS; i++)
	{
		BYTE* values;
		if (i < MAX_INSTRUMENTS)
			values = SynthObj.InstrumentValues[i][0];
		else
			values = SynthObj.GlobalValues[0];
		for (int u = 0; u < MAX_UNITS; u++)
		{
			if (values[u*MAX_UNITS+0] == M_FSTx)
			{
				FSTx_valP v = (FSTx_valP)(&values[u*MAX_UNITS+0]);
								
				int target_inst;
				if (v->dest_stack == -1)
					target_inst = i;
				else
					target_inst = v->dest_stack;

				// the store points to another stack, so continue
				if (target_inst != stack)
					continue;

				// a up/down process 
				if (unit2 != -1)
				{
					// if the target unit was unit1 or unit2 realign target unit
					if (v->dest_unit == unit1)
					{
						v->dest_unit = unit2;
					}
					else if (v->dest_unit == unit2)
					{
						v->dest_unit = unit1;
					}
				}
			}
		}
	}
}

// autoconvert 1.0 instrument stacks
bool Autoconvert10(int stack)
{
	// get desired stack
	BYTE* values;
	if (stack < MAX_INSTRUMENTS)
		values = SynthObj.InstrumentValues[stack][0];
	else
		values = SynthObj.GlobalValues[0];

	// replace the delay with the new one
	for (int u = 0; u < MAX_UNITS; u++)
	{
		if (values[u*MAX_UNITS+0] == M_DLL)
		{
			DLL10_val ov;
			memcpy(&ov, &values[u*MAX_UNITS+0], sizeof(DLL10_val));
			DLL_valP nv = (DLL_valP)(&values[u*MAX_UNITS+0]);
			nv->id = ov.id;
			nv->pregain = ov.pregain;
			nv->dry = ov.dry;
			nv->feedback = ov.feedback;
			nv->damp = ov.damp;
			nv->freq = 0;
			nv->depth = 0;
			nv->guidelay = ov.guidelay;
			nv->synctype = ov.synctype;
			nv->leftreverb = ov.leftreverb;
			nv->reverb = ov.reverb;
		}
	}
	return true;
}

// autoconvert 1.1 instrument stacks
bool Autoconvert11(int stack)
{
	// get desired stack
	BYTE* values;
	if (stack < MAX_INSTRUMENTS)
		values = SynthObj.InstrumentValues[stack][0];
	else
		values = SynthObj.GlobalValues[0];

	// replace the osc with the new one
	for (int u = 0; u < MAX_UNITS; u++)
	{
		if (values[u*MAX_UNITS+0] == M_VCO)
		{
			VCO11_val ov;
			memcpy(&ov, &values[u*MAX_UNITS+0], sizeof(VCO11_val));
			VCO_valP nv = (VCO_valP)(&values[u*MAX_UNITS+0]);
			nv->id = ov.id;
			nv->transpose = ov.transpose;
			nv->detune = ov.detune;
			nv->phaseofs = ov.phaseofs;
			nv->gate = 0x55;
			nv->color = ov.color;
			nv->shape = ov.shape;
			nv->gain = ov.gain;
			nv->flags = ov.flags;
		}
	}
	return true;
}


// ---------------------- "chunk" thingy, in-memory file for reading and writing ---------------------

Chunk temp_chunk;


void chunk_init(ChunkP c)
{
    c->size = 0;
    c->readpos = 0;
}

size_t chunk_write(const void *str, size_t element_size, size_t element_count, ChunkP chunk)
{
    BYTE *s = (BYTE *)str;

    size_t byteswritten = 0;
    for (size_t i=0; (i<element_size*element_count) && (chunk->size < MAX_CHUNK_SIZE); i++)
    {
        chunk->data[chunk->size++] = *s++;
        byteswritten++;
    }
    return byteswritten / element_size;
}

size_t chunk_read(void *str, size_t element_size, size_t element_count, ChunkP chunk)
{
    BYTE *s = (BYTE *)str;

    size_t bytesread = 0;
    for (size_t i=0; (i<element_size*element_count) && (chunk->readpos < chunk->size); i++)
    {
        *s++ = chunk->data[chunk->readpos++];
        bytesread++;
    }
    return bytesread / element_size;
}

size_t chunk_write_FILE(FILE *stream, ChunkP chunk)
{
    size_t max = (MAX_CHUNK_SIZE - chunk->size);
    size_t bytesread = fread(&(chunk->data[chunk->size]), 1, max, stream);
    chunk->size += bytesread;
    return bytesread;
}


bool read_instrument_values_from_chunk(
        ChunkP chunk,
        BYTE *dest,
        DWORD num_units_per_instrument_in_file,        
        int num_values_per_unit_in_file,
        int num_values_per_unit_actually_used)
{
    int num_values_to_skip = num_values_per_unit_in_file - num_values_per_unit_actually_used;
    if (num_values_to_skip > 100)
        return false;

    for (int j=0; j<num_units_per_instrument_in_file; j++)
    {
        chunk_read(dest, 1, num_values_per_unit_actually_used, chunk);
        if (num_values_per_unit_actually_used < num_values_per_unit_in_file)
        { // some of the values have to be skipped
            char dummy_buffer[100];
            chunk_read(dummy_buffer, 1, num_values_to_skip, chunk);
        }
        dest += MAX_SLOT_VALUES;
    }
    return true;
} // read_instrument_values_from_chunk()


// load patch from data chunk
// the given chunk has to be initialized and preloaded with the data, of course
// returns false, if loading was unsuccesful
// returns true, if loading was succesful
bool Go4kVSTi_LoadPatchFromDataChunk(ChunkP chunk)
{
	Go4kVSTi_ResetPatch(); // reset ALL PATCH DATA

	DWORD version;
    DWORD num_instruments_in_file = 0;
    DWORD num_units_per_instrument_in_file = 0;
	bool version10 = false;
	bool version11 = false;
	bool version12 = false;
    bool version13 = false;
    DWORD num_instrument_settings_in_file = 0;
    DWORD num_values_per_unit_in_file = 0;
    DWORD num_values_per_unit_actually_used = 0; // how many of the values (or "unit slots" as they're called in 4klang 3.1.1) might actually contain stuff
    DWORD num_global_settings_in_file = 0;

	chunk_read(&version, 1, 4, chunk);

	if (versiontag == version)
    {         
        // this version Forked Klang file
        chunk_read(&num_instruments_in_file, 1, 4, chunk);
        chunk_read(&num_units_per_instrument_in_file, 1, 4, chunk);
        num_values_per_unit_in_file = MAX_SLOT_VALUES;
        num_values_per_unit_actually_used = MAX_SLOT_VALUES;
        num_instrument_settings_in_file = num_instruments_in_file;
        chunk_read(&num_global_settings_in_file, 1, 4, chunk);
    }
    else if (version == versiontagfk301)
	{        
        // Older Forked Klang file
        chunk_read(&num_instruments_in_file, 1, 4, chunk);
        chunk_read(&num_units_per_instrument_in_file, 1, 4, chunk);
        num_values_per_unit_in_file = MAX_SLOT_VALUES;
        num_values_per_unit_actually_used = MAX_SLOT_VALUES;
        num_instrument_settings_in_file = num_instruments_in_file;
        num_global_settings_in_file = 0;
    }
	// version 1.3 file
	else if (version == versiontag13)
	{
		// only mulp2 unit added and layout for instruments changed, no need for message
		//MessageBox(0,"Autoconvert. Please save file again", "1.3 File Format", MB_OK | MB_SETFOREGROUND);
		version13 = true;
        num_instruments_in_file = 16;
        num_units_per_instrument_in_file = 32;
        num_values_per_unit_in_file = 32;
        num_values_per_unit_actually_used = 16;
        num_instrument_settings_in_file = 0; // 4klang doesn't have instrument input settings
	}
	// version 1.2 file
	else if (version == versiontag12)
	{
		// only fld unit added, no need for message
		//MessageBox(0,"Autoconvert. Please save file again", "1.2 File Format", MB_OK | MB_SETFOREGROUND);
		version12 = true;
        num_instruments_in_file = 16;
        num_units_per_instrument_in_file = 32;
        num_values_per_unit_in_file = 32;
        num_values_per_unit_actually_used = 16;
        num_instrument_settings_in_file = 0; // 4klang doesn't have instrument input settings
	}
	// version 1.1 file
	else if (version == versiontag11)
	{
		MessageBox(0,"Autoconvert. Please save file again", "1.1 File Format", MB_OK | MB_SETFOREGROUND);
		version11 = true;
		version12 = true;
        num_instruments_in_file = 16;
        num_units_per_instrument_in_file = 32;
        num_values_per_unit_in_file = 32;
        num_values_per_unit_actually_used = 16;
        num_instrument_settings_in_file = 0; // 4klang doesn't have instrument input settings
	}
	// version 1.0 file
	else if (version == versiontag10)
	{
		MessageBox(0,"Autoconvert. Please save file again", "1.0 File Format", MB_OK | MB_SETFOREGROUND);
		version10 = true;
		version11 = true;
		version12 = true;
        num_instruments_in_file = 16;
        num_units_per_instrument_in_file = 32;
        num_values_per_unit_in_file = 32;
        num_values_per_unit_actually_used = 16;
        num_instrument_settings_in_file = 0; // 4klang doesn't have instrument input settings
	}
	// unknown, unsupported format
	else
	{
		MessageBox(0,"The file was created with an unknown version of 4klang or something", "File Format Error", MB_ICONERROR | MB_SETFOREGROUND);
		return false;
	}

    // Check that we can load this stuff

    if (num_instruments_in_file > MAX_INSTRUMENTS)
    {
			MessageBox(0,"The file contains too many instruments", "File Format Error", MB_ICONERROR | MB_SETFOREGROUND);
			return false;
    }

    if (num_units_per_instrument_in_file > MAX_UNITS)
    {
			MessageBox(0,"The file has too many units per instrument", "File Format Error", MB_ICONERROR | MB_SETFOREGROUND);
			return false;
    }

    if (num_values_per_unit_actually_used > MAX_SLOT_VALUES) // it doesn't matter if the file has extra unused values
    {
            MessageBox(0,"The file has too many values (\"unit slots\") per unit", "File Format Error", MB_ICONERROR | MB_SETFOREGROUND);				
			return false;
    }
	
	// read data
	chunk_read(&(SynthObj.Polyphony), 1, 4, chunk);
    chunk_read(SynthObj.InstrumentNames, 1, num_instruments_in_file*64, chunk);

	for (int i=0; i<num_instruments_in_file; i++)
    {            
        // ResetInstrumentInputSettings(i);
        if (!read_instrument_values_from_chunk(chunk, &SynthObj.InstrumentValues[i][0][0], num_units_per_instrument_in_file, num_values_per_unit_in_file, num_values_per_unit_actually_used))
        {
            MessageBox(0,"The file stinks!!! Will not load.", "File Format Error", MB_ICONERROR | MB_SETFOREGROUND);
            return false;
        }
        for (int j=num_units_per_instrument_in_file; j<MAX_UNITS; j++)
        { // clear any remaining units
            Go4kVSTi_ClearInstrumentSlot(i, j);
        }
	}

    
    // if there were fewer instruments in the file than our MAX_INSTRUMENTS, reset the remaining ones
    for (int i=num_instruments_in_file; i<MAX_INSTRUMENTS; i++)
    {
        // Go4kVSTi_ResetInstrument(i);
        // ResetInstrumentInputSettings(i);
    }

    // The same thing for global units
    if (!read_instrument_values_from_chunk(chunk, (BYTE *)((void *)(SynthObj.GlobalValues)), num_units_per_instrument_in_file, num_values_per_unit_in_file, num_values_per_unit_actually_used))
    {
        MessageBox(0,"The file smells! Will not load.", "File Format Error", MB_ICONERROR | MB_SETFOREGROUND);
        return false;
    }
    for (int j=num_units_per_instrument_in_file; j<MAX_UNITS; j++)
    { // clear any remaining global units
        Go4kVSTi_ClearGlobalSlot(j);
    }

    if (num_instrument_settings_in_file > 0)
        chunk_read(SynthObj.InstrumentInputSettings, 1, sizeof(InstrumentInputSettingsS)*num_instrument_settings_in_file, chunk);

    if (num_global_settings_in_file > 0)
        chunk_read(&GlobalSettings, 1, num_global_settings_in_file * sizeof(DWORD), chunk);

//	fclose(file);

	// convert 1.0 file format
	if (version10)
	{
		// convert all instruments
		for (int i = 0; i < num_instruments_in_file; i++)
		{
			if (!Autoconvert10(i))
			{
				char errmsg[64];
				sprintf(errmsg, "Instrument %d could not be converted", i+1);
				MessageBox(0, errmsg, "Error", MB_OK | MB_SETFOREGROUND);
			}
		}
		// convert global
		if (!Autoconvert10(GLOBAL_INSTRUMENT_IDX))
		{
			char errmsg[64];
			sprintf(errmsg, "Global could not be converted");
			MessageBox(0, errmsg, "Error", MB_OK | MB_SETFOREGROUND);
		}
	}
	// convert 1.1 file format
	if (version11)
	{
		// convert all instruments
		for (int i = 0; i < num_instruments_in_file; i++)
		{
			if (!Autoconvert11(i))
			{
				char errmsg[64];
				sprintf(errmsg, "Instrument %d could not be converted", i+1);
				MessageBox(0, errmsg, "Error", MB_OK | MB_SETFOREGROUND);
			}
		}
		// convert global
		if (!Autoconvert11(MAX_INSTRUMENTS))
		{
			char errmsg[64];
			sprintf(errmsg, "Global could not be converted");
			MessageBox(0, errmsg, "Error", MB_OK | MB_SETFOREGROUND);
		}
	}
	// convert 1.2 file format
	if (version12)
	{
		// nothing to do, only fld unit added at the end
	}

	Go4kVSTi_UpdateDelayAndSlewTimes();

    return true;
} // Go4kVSTi_LoadPatchFromDataChunk()


// load patch data
void Go4kVSTi_LoadPatch(char *filename)
{
	FILE *file = fopen(filename, "rb");
	if (file)
	{
        chunk_init(&temp_chunk);
        chunk_write_FILE(file, &temp_chunk);

		fclose(file);

        Go4kVSTi_LoadPatchFromDataChunk(&temp_chunk);
	}
}

int Go4kVSTi_SavePatchToDataChunk(ChunkP chunk)
{
    DWORD num_instruments = MAX_INSTRUMENTS;
    DWORD num_units_per_instrument = MAX_UNITS;
    
    chunk_init(chunk);

	chunk_write(&versiontag, 1, 4, chunk);
    chunk_write(&num_instruments, 1, 4, chunk);
    chunk_write(&num_units_per_instrument, 1, 4, chunk);

    DWORD numglobalsettings = sizeof(GlobalSettings) / sizeof(DWORD);
    chunk_write(&numglobalsettings, 1, 4, chunk);

	chunk_write(&(SynthObj.Polyphony), 1, 4, chunk);
	chunk_write(SynthObj.InstrumentNames, 1, MAX_INSTRUMENTS*64, chunk);
	chunk_write(SynthObj.InstrumentValues, 1, MAX_INSTRUMENTS*MAX_UNITS*MAX_SLOT_VALUES, chunk);
	chunk_write(SynthObj.GlobalValues, 1, MAX_UNITS*MAX_SLOT_VALUES, chunk);
    chunk_write(SynthObj.InstrumentInputSettings, 1, sizeof(InstrumentInputSettingsS)*MAX_INSTRUMENTS, chunk);
    
    chunk_write(&GlobalSettings, 1, sizeof(GlobalSettings), chunk);

    return (int)chunk->size;
}

// save patch data
void Go4kVSTi_SavePatch(const char *filename)
{
	FILE *file = fopen(filename, "wb");
	if (file)
	{
        if (Go4kVSTi_SavePatchToDataChunk(&temp_chunk))
            fwrite(&temp_chunk.data, 1, temp_chunk.size, file);

		fclose(file);

	}
}

// load instrument dada to specified instrument
bool Go4kVSTi_LoadInstrumentFromDataChunk(ChunkP chunk, int instrument_index)
{
	DWORD version;
	bool version10 = false;
	bool version11 = false;
	bool version12 = false;
	bool version13 = false;

    DWORD num_units_per_instrument_in_file = 0;
    DWORD num_values_per_unit_in_file = 0;
    DWORD num_values_per_unit_actually_used = 0; // how many of the values (or "unit slots" as they're called in 4klang 3.1.1) might actually contain stuff

	chunk_read(&version, 1, 4, chunk);

    if (versiontag == version)
    {
        chunk_read(&num_units_per_instrument_in_file, 1, 4, chunk);
        num_values_per_unit_in_file = MAX_SLOT_VALUES;
        num_values_per_unit_actually_used = MAX_SLOT_VALUES;
    }
	else if (version == versiontag13)
	{
		version13 = true;
        num_units_per_instrument_in_file = 32;
        num_values_per_unit_in_file = 32;
        num_values_per_unit_actually_used = 16;
    }
    // version 1.2 file
	else if (version == versiontag12)
	{
		// only fld unit added, no need for message
		//MessageBox(0,"Autoconvert. Please save file again", "1.2 File Format", MB_OK | MB_SETFOREGROUND);
		version12 = true;
		version13 = true;
        num_units_per_instrument_in_file = 32;
        num_values_per_unit_in_file = 32;
        num_values_per_unit_actually_used = 16;

	}
	// version 1.1 file
	else if (version == versiontag11)
	{
		MessageBox(0,"Autoconvert. Please save file again", "1.1 File Format", MB_OK | MB_SETFOREGROUND);
		version11 = true;
		version12 = true;
		version13 = true;
        num_units_per_instrument_in_file = 32;
        num_values_per_unit_in_file = 32;
        num_values_per_unit_actually_used = 16;
	}
	// version 1.0 file
	else if (version == versiontag10)
	{
		MessageBox(0,"Autoconvert. Please save file again", "1.0 File Format", MB_OK | MB_SETFOREGROUND);
		version10 = true;
		version11 = true;
		version12 = true;
		version13 = true;
        num_units_per_instrument_in_file = 32;
        num_values_per_unit_in_file = 32;
        num_values_per_unit_actually_used = 16;
	}
	// newer format than supported
	else
	{
		MessageBox(0,"The file was created with a newer version of 4klang or maybe with Microsoft Excel.", "File Format Error", MB_ICONERROR | MB_SETFOREGROUND);
		// fclose(file);
		return false;
	}
	
	// int slots=MAX_UNITS;

	if (instrument_index < MAX_INSTRUMENTS)
	{
		Go4kVSTi_ResetInstrument(instrument_index, true);
		chunk_read(SynthObj.InstrumentNames[instrument_index], 1, 64, chunk);

//			fread(SynthObj.InstrumentValues[instrument_index], 1, slots*MAX_SLOT_VALUES, file);

        if (!read_instrument_values_from_chunk(chunk, &SynthObj.InstrumentValues[instrument_index][0][0], num_units_per_instrument_in_file, num_values_per_unit_in_file, num_values_per_unit_actually_used))
        {
            MessageBox(0,"The file stinks!!! Will not load.", "File Format Error", MB_ICONERROR | MB_SETFOREGROUND);
            return false;
        }
        for (int j=num_units_per_instrument_in_file; j<MAX_UNITS; j++)
        { // clear any remaining units
            Go4kVSTi_ClearInstrumentSlot(instrument_index, j);
        }

	}
	else
	{
		Go4kVSTi_ResetGlobalUnits(true);
		chunk_read(SynthObj.InstrumentNames[instrument_index], 1, 64, chunk);

//			fread(SynthObj.GlobalValues, 1, slots*MAX_SLOT_VALUES, file);

        // The same thing for global units
        if (!read_instrument_values_from_chunk(chunk, (BYTE *)((void *)(SynthObj.GlobalValues)), num_units_per_instrument_in_file, num_values_per_unit_in_file, num_values_per_unit_actually_used))
        {
            MessageBox(0,"The file smells! Will not load.", "File Format Error", MB_ICONERROR | MB_SETFOREGROUND);
            return false;
        }
        for (int j=num_units_per_instrument_in_file; j<MAX_UNITS; j++)
        { // clear any remaining global units
            Go4kVSTi_ClearGlobalSlot(j);
        }
	}	

	// convert 1.0 file format
	if (version10)
	{
		// convert instruments
		if (instrument_index < MAX_INSTRUMENTS)
		{
			if (!Autoconvert10(instrument_index))
			{
				char errmsg[64];
				sprintf(errmsg, "Instrument %d could not be converted", instrument_index+1);
				MessageBox(0, errmsg, "Error", MB_OK | MB_SETFOREGROUND);
                return false;
			}
		}
		// convert global
		else
		{
			if (!Autoconvert10(MAX_INSTRUMENTS))
			{
				char errmsg[64];
				sprintf(errmsg, "Global could not be converted");
				MessageBox(0, errmsg, "Error", MB_OK | MB_SETFOREGROUND);
                return false;
			}
		}
	}
	// convert 1.1 file format
	if (version11)
	{
		// convert instruments
		if (instrument_index < MAX_INSTRUMENTS)
		{
			if (!Autoconvert11(instrument_index))
			{
				char errmsg[64];
				sprintf(errmsg, "Instrument %d could not be converted", instrument_index+1);
				MessageBox(0, errmsg, "Error", MB_OK | MB_SETFOREGROUND);
                return false;
			}
		}
		// convert global
		else
		{
			if (!Autoconvert11(MAX_INSTRUMENTS))
			{
				char errmsg[64];
				sprintf(errmsg, "Global could not be converted");
				MessageBox(0, errmsg, "Error", MB_OK | MB_SETFOREGROUND);
                return false;
			}
		}
	}
	// convert 1.2 file format
	if (version12)
	{
		// nothing to do, only fld unit added at the end
	}

	if (version13)
	{
		// 
	}

	Go4kVSTi_UpdateDelayAndSlewTimes();

    return true;
} // Go4kVSTi_LoadInstrumentFromDataChunk()



void Go4kVSTi_LoadInstrument(char* filename, int instrument_index)
{
	FILE *file = fopen(filename, "rb");
	if (file)
	{
        chunk_init(&temp_chunk);
        chunk_write_FILE(file, &temp_chunk);

		fclose(file);

        Go4kVSTi_LoadInstrumentFromDataChunk(&temp_chunk, instrument_index);
	}
}

// save instrument data from current channel
void Go4kVSTi_SaveInstrument(char* filename, int instrument_index)
{
	FILE *file = fopen(filename, "wb");
	if (file)
	{
		fwrite(&versiontag, 1, 4, file);

        DWORD num_units_per_instrument = MAX_UNITS;
        fwrite(&num_units_per_instrument, 1, 4, file);

		if (instrument_index < MAX_INSTRUMENTS)
		{
			fwrite(SynthObj.InstrumentNames[instrument_index], 1, 64, file);
			fwrite(SynthObj.InstrumentValues[instrument_index], 1, MAX_UNITS*MAX_SLOT_VALUES, file);
		}
		else if (instrument_index == MAX_INSTRUMENTS)
		{ 
            
			fwrite(SynthObj.InstrumentNames[instrument_index], 1, 64, file);
			fwrite(SynthObj.GlobalValues, 1, MAX_UNITS*MAX_SLOT_VALUES, file);
		}
        else
            MessageBox(0, "apuva", "apuuva", MB_OK | MB_SETFOREGROUND);
		fclose(file);
	}
}

// load unit data into specified slot
void Go4kVSTi_LoadUnit(char* filename, BYTE* slot)
{
	FILE *file = fopen(filename, "rb");
	if (file)
	{
		fread(slot, 1, MAX_SLOT_VALUES, file);
		fclose(file);
	//if (slot[0] == M_DLL)
        if (UnitHasDelayBuffer((SOME_UNIT_valP)(slot))) // TONTTOROO ONKS TSS VIKAA
		{
			Go4kVSTi_ClearDelayLinesAndSlews();
			Go4kVSTi_UpdateDelayAndSlewTimes();
		}
	}
}

// save unit date from specified slot
void Go4kVSTi_SaveUnit(char* filename, BYTE* slot)
{
	FILE *file = fopen(filename, "wb");
	if (file)
	{
		fwrite(slot, 1, MAX_SLOT_VALUES, file);
		fclose(file);
	}
}

int GetShift2(int in)
{
	int ret = 0;
	while (in = (in >> 1))
	{
		ret++;
	}
	return ret;
}

struct SynthUses
{
	bool env_gm;
	bool env_adrm;

	bool vco_phaseofs;
	bool vco_shape;
	bool vco_fm;
	bool vco_pm;
	bool vco_tm;
	bool vco_dm;
	bool vco_cm;
	bool vco_gm;
	bool vco_sm;
	bool vco_gate;

	bool vcf_fm;
	bool vcf_rm;

	bool dst_use;
	bool dst_snh;
	bool dst_dm;
	bool dst_sm;

	bool dll_use;
	bool dll_notesync;
	bool dll_damp;
	bool dll_chorus;
	bool dll_use_mod;
	bool dll_pm;
	bool dll_fm;
	bool dll_im;
	bool dll_dm;
	bool dll_sm;
	bool dll_am;

	bool fop_loadnote;

	bool pan_use;
	bool pan_pm;

    bool fst_use; // this didn't exist in 4klang, why?

	bool fstg_use;
	
	bool out_am;
	bool out_gm;

	bool fld_use;
	bool fld_vm;

    bool slw_use;
    bool slw_calcmode_linear;
    bool slw_calcmode_rms;

    bool fldfrom_use;
}; // struct SynthUses

void GetUses(SynthUses *uses, bool InstrumentUsed[])
{
	for (int i = 0; i <= MAX_INSTRUMENTS; i++)
	{
		// skip usage checks for non used instruments
		if (i < MAX_INSTRUMENTS)
			if (!InstrumentUsed[i])
				continue;

		for (int u = 0; u < MAX_UNITS; u++)
		{
			BYTE *v;
			if (i < MAX_INSTRUMENTS)
				v = SynthObj.InstrumentValues[i][u];
			else
				v = SynthObj.GlobalValues[u];

			if (v[0] == M_VCO)
			{
				if (((VCO_valP)v)->phaseofs != 0)
					uses->vco_phaseofs = true;
				if (((VCO_valP)v)->shape != 64)
					uses->vco_shape = true;
				if (((VCO_valP)v)->flags & VCO_GATE)
					uses->vco_gate = true;
			}
			if (v[0] == M_DST)
			{
				uses->dst_use = true;
				if (((DST_valP)v)->snhfreq != 128)
					uses->dst_snh = true;
			}
			if (v[0] == M_DLL)
			{
				uses->dll_use = true;
				if (((DLL_valP)v)->synctype == 2)
					uses->dll_notesync = true;
				if (((DLL_valP)v)->damp > 0)
					uses->dll_damp = true;
				if ((((DLL_valP)v)->freq > 0) && (((DLL_valP)v)->depth > 0))
					uses->dll_chorus = true;
			}
			if (v[0] == M_PAN)
			{
				if (((PAN_valP)v)->panning != 64)
					uses->pan_use = true;
			}
			if (v[0] == M_FOP)
			{
				if (((FOP_valP)v)->flags == FOP_LOADNOTE)
					uses->fop_loadnote = true;
			}
			if (v[0] == M_FLD)
			{				
				uses->fld_use = true;
			}
			if (v[0] == M_SLW)
			{				
				uses->slw_use = true;
                SLW_valP slw = (SLW_valP)v;
                if (slw->calcmode == SLW_CALCMODE_LINEAR) 
                    uses->slw_calcmode_linear = true;
                else if (slw->calcmode == SLW_CALCMODE_RMS)
                    uses->slw_calcmode_rms = true;
                else
                {
                    MessageBox(0, "Unknown calcmode in SLW unit. Who knows what will happen now.", "Error", MB_ICONERROR | MB_SETFOREGROUND | MB_OK);
                }
			}

            if (IsModulationLinkUnit((SOME_UNIT_valP)(v))) 
              // This unit writes to or reads from something in another unit that might be an optional feature 
              // of that other unit, and wouldn't be needed if no other unit refers to it.
			{ 
				InstrumentWorkspaceP mwork;
				int stack = GetModulationLinkInstrument((SOME_UNIT_valP)(v));
				int unit = GetModulationLinkUnit((SOME_UNIT_valP)(v));
				int slot = GetModulationLinkSlot((SOME_UNIT_valP)(v));
				// local storage?
				if (stack == -1 || stack == i)
				{
					stack = i;

                    if (v[0] == M_FSTx)
                        uses->fst_use = true;
                    else if (v[0] == M_FLDFROM)
                        uses->fldfrom_use = true;
				}
				else
				{
                    if (v[0] == M_FSTx)
					    uses->fstg_use = true;
                    else if (v[0] == M_FLDFROM)
                        MessageBox(0,"AAAAARGH. Who gave you an M_FLDFROM with an inter-instrument reference.", "Please wait while coding part II.", MB_OK | MB_SETFOREGROUND);
				}

                // find out the target unit type
				BYTE *v2;
				if (stack < MAX_INSTRUMENTS)
					v2 = SynthObj.InstrumentValues[stack][unit];
				else
					v2 = SynthObj.GlobalValues[unit];

				if (v2[0] == M_ENV)
				{
					if (slot == 2)
						uses->env_gm = true;
					if (slot == 3)
						uses->env_adrm = true;
					if (slot == 4)
						uses->env_adrm = true;
					if (slot == 6)
						uses->env_adrm = true;
				}
				if (v2[0] == M_VCO)
				{
					if (slot == 1)
						uses->vco_tm = true;
					if (slot == 2)
						uses->vco_dm = true;
					if (slot == 3)
						uses->vco_fm = true;
					if (slot == 4)
						uses->vco_pm = true;
					if (slot == 5)
						uses->vco_cm = true;
					if (slot == 6)
						uses->vco_sm = true;
					if (slot == 7)
						uses->vco_gm = true;
				}
				if (v2[0] == M_VCF)
				{
					if (slot == 4)
						uses->vcf_fm = true;
					if (slot == 5)
						uses->vcf_rm = true;
				}
				if (v2[0] == M_DST)
				{
					if (slot == 2)
						uses->dst_dm = true;
					if (slot == 3)
						uses->dst_sm = true;
				}
				if (v2[0] == M_DLL)
				{
					if (slot == 0)
					{
						uses->dll_pm = true;
						uses->dll_use_mod = true;
					}
					if (slot == 1)
					{
						uses->dll_fm = true;
						uses->dll_use_mod = true;
					}
					if (slot == 2)
					{
						uses->dll_im = true;
						uses->dll_use_mod = true;
					}
					if (slot == 3)
					{
						uses->dll_dm = true;
						uses->dll_damp = true;
						uses->dll_use_mod = true;
					}
					if (slot == 4)
					{
						uses->dll_sm = true;
						uses->dll_use_mod = true;
					}					
					if (slot == 5)
					{
						uses->dll_am = true;
						uses->dll_use_mod = true;
					}
				}
				if (v2[0] == M_PAN)
				{
					if (slot == 0)
						uses->pan_pm = true;
				}
				if (v2[0] == M_OUT)
				{
					if (slot == 0)
						uses->out_am = true;
					if (slot == 1)
						uses->out_gm = true;
				}
				if (v2[0] == M_FLD)
				{
					if (slot == 0)
						uses->fld_vm = true;
				}
			}
		}
	}
} // GetUses()

#ifdef EXPORT_OBJECT_FILE
int randiSeed = 23;
int randi()
{	
	return( ((randiSeed = randiSeed * 214013L + 2531011L) >> 16)  & 0x7fff );
}
#endif

void Go4kVSTi_SaveByteStream(HINSTANCE hInst, char* filename, int useenvlevels, int useenotevalues, int clipoutput, int undenormalize, int objformat, int output16, int swing_strength)
{
	std::string fpath = filename;
#ifdef EXPORT_OBJECT_FILE
	unsigned char* CryptTable = (unsigned char*)malloc(1024*1024);
	randiSeed = 23;
	for (int i = 0; i < 1024*1024; i++)
		CryptTable[i] = randi() & 0xff;
	char temppath[1024];
	GetTempPath(1024, temppath);
	std::string path = temppath;
	// construct temp strings
	std::string incfile = path + "4klang.inc";
#else
	std::string incfile = filename;
#endif
    //MessageBox(0, "TOTTOROOOO KOODAUS KESKEN", "HUMPPATIRALLAA", MB_OK | MB_SETFOREGROUND);
    //return;

	int maxinst = 0;
	FILE *file = fopen(incfile.c_str(), "w");
	if (file)
	{
		bool InstrumentUsed[MAX_INSTRUMENTS];
		// determine max instrument count
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			InstrumentUsed[i] = false;
			for (int j = 0; j < PatternIndices[i].size(); j++)
			{
				if (PatternIndices[i][j])
				{
					InstrumentUsed[i] = true;
					break;
				}
			}
		}
		// determine max instrument count and create a mapping instrument index -> used instrument index
		int InstrumentIndex[MAX_INSTRUMENTS];
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			// dummy init
			InstrumentIndex[i] = -1;
			if (InstrumentUsed[i])
			{
				InstrumentIndex[i] = maxinst;
				maxinst++;
			}

		}
		std::vector<WORD> delay_times;
		std::vector<int> delay_indices;
		bool hasReverb = false;
		bool hasNoteSync = false;
		// add array items for notesync (1) and reverb (16) times, 1 + 16 = 17
		for (int i = 0; i < 17; i++)
		{
			DWORD times = (&go4k_delay_times)[i];
			delay_times.push_back(times);
			delay_indices.push_back(i);
		}
		delay_times[0] = 0;
		for (int i = 0; i <= MAX_INSTRUMENTS; i++)
		{
			for (int u = 0; u < MAX_UNITS; u++)
			{
				//	// used instrument or global?
				if (InstrumentUsed[i] || i == MAX_INSTRUMENTS)
				{
					SOME_UNIT_valP v;
					if (i < MAX_INSTRUMENTS)
						v = (SOME_UNIT_valP)(SynthObj.InstrumentValues[i][u]);
					else
						v = (SOME_UNIT_valP)(SynthObj.GlobalValues[u]);

                    // if (v->id == M_DLL)
					if (UnitHasDelayBuffer(v))
					{
						if (!UnitHasReverb(v))
						{
							// if not notesync
							if (!UnitHasNoteSync(v))
							{
								DWORD times = (&go4k_delay_times)[delay_indices.size()];
								// check if indexed value already existed
								int found = -1;
								for (int j = 17; j < delay_times.size(); j++)
								{
									if (delay_times[j] == times)
									{
										found = j;
										break;
									}
								}
								if (found != -1)
								{
									// already in list, so let index point to that one
									delay_indices.push_back(found);
								}
								else
								{
									// new value, so push it
									delay_times.push_back(times);
									delay_indices.push_back(delay_times.size()-1);
								}
							}
							else
							{
								hasNoteSync = true;
								delay_indices.push_back(0);
							}
						}
						else
						{
							hasReverb = true;
						}
					}
				}
				// no used instrument
				else
				{
					DLL_valP v;
					v = (DLL_valP)(SynthObj.InstrumentValues[i][u]);
					if (v->id == M_DLL)
					{
						if (!v->reverb)
						{
							// if not notesync
							if (v->synctype != 2)
							{
								// just push a dummy index
								delay_indices.push_back(-1);
							}
						}
					}
				}
			}
		} // for MAX_INSTRUMENTS

		// if we dont have reverb, remove values and adjust indices
		if (!hasReverb)
		{
			// move values
			for (int i = 17; i < delay_times.size(); i++) // "17" = 1 + 8 + 8, 1 for notesync, 8 for left reverb taps, 8 for right reverb taps
			{
				delay_times[i-16] = delay_times[i];
			}
			// remove obsolete values
			for (int i = 0; i < 16; i++)
			{
				delay_times.pop_back();
			}
			// adjust indices
			for (int i = 0; i < delay_indices.size(); i++)
			{
				if (delay_indices[i] != 0)
					delay_indices[i] -= 16;
			}
		}
		// if we dont have notesync, remove values and adjust indices
		if (!hasNoteSync)
		{
			// move values
			for (int i = 1; i < delay_times.size(); i++)
			{
				delay_times[i-1] = delay_times[i];
			}
			// remove obsolete values
			delay_times.pop_back();
			// adjust indices
			for (int i = 0; i < delay_indices.size(); i++)
			{
				delay_indices[i] -= 1;
			}
		}

		SynthUses uses;
		memset(&uses, 0, sizeof(SynthUses));
		GetUses(&uses, InstrumentUsed);

		// write inc file
		fprintf(file, "%%macro export_func 1\n");
//		fprintf(file, "   %%define %%1 _%%1\n");
		fprintf(file, "   global _%%1\n");
		fprintf(file, "   _%%1:\n");
		fprintf(file, "%%endmacro\n");

//		fprintf(file, "%%macro export_dword_array 2\n");
//		fprintf(file, "   %%define %%1 _%%1\n");
//		fprintf(file, "   global %%1\n");
//		fprintf(file, "   %%1 resd %%2\n");
//		fprintf(file, "%%endmacro\n");

/*		if (objformat == 0)
			MessageBox(0,"WINDOWS","",MB_OK);
		if (objformat == 1)
			MessageBox(0,"LINUX","",MB_OK);
		if (objformat == 2)
			MessageBox(0,"MACOSX","",MB_OK);
*/

	if (objformat == 0) // 0:windows, 1:linux, 2:osx
	{
		// use sections only for windows, because that's where they seem to work.
		// linux doesnt have any crinkler like packer to take advantage of multiple sections
		// and osx export has a bug in current nasm and is not able to export custom sections
		
		fprintf(file, "%%ifndef LINUX\n");
		fprintf(file, "%%define USE_SECTIONS\n");
		fprintf(file, "%%endif\n");
	}

		fprintf(file, "%%define SAMPLE_RATE	%d\n", 44100);
		fprintf(file, "%%define MAX_INSTRUMENTS	%d\n", maxinst);
		fprintf(file, "%%define MAX_VOICES %d\n", SynthObj.Polyphony);
		fprintf(file, "%%define HLD 1\n");
		fprintf(file, "%%define BPM %f\n", BeatsPerMinute);
		fprintf(file, "%%define MAX_PATTERNS %d\n", PatternIndices[0].size());
		fprintf(file, "%%define PATTERN_SIZE_SHIFT %d\n", GetShift2(PatternSize));
		fprintf(file, "%%define PATTERN_SIZE (1 << PATTERN_SIZE_SHIFT)\n");
		fprintf(file, "%%define	MAX_TICKS (MAX_PATTERNS*PATTERN_SIZE)\n");
		fprintf(file, "%%define	SAMPLES_PER_TICK %d\n", (int)(44100.0f*4.0f*60.0f/(BeatsPerMinute*16*TickScaler)));
		fprintf(file, "%%define DEF_LFO_NORMALIZE %.10f\n", LFO_NORMALIZE);
		fprintf(file, "%%define	MAX_SAMPLES	(SAMPLES_PER_TICK*MAX_TICKS)\n");
		fprintf(file, "%s%%define 	GO4K_USE_16BIT_OUTPUT\n",output16?"":";");

        if (swing_strength > 50)
        {
            fprintf(file, "%%define GO4K_USE_SWING\n");
            fprintf(file, "%%define SWING_STRENGTH_PERCENT %d\n", swing_strength);
            fprintf(file, "%%define SWING_TICK_DIFFERENCE ((SAMPLES_PER_TICK * 2) * (2 * SWING_STRENGTH_PERCENT - 100) / 100)\n");
            fprintf(file, "%%define SAMPLES_PER_TICK_SWING_DOWNTICK (SAMPLES_PER_TICK + (SWING_TICK_DIFFERENCE / 2))\n");
            fprintf(file, "%%define SAMPLES_PER_TICK_SWING_UPTICK (SAMPLES_PER_TICK - (SWING_TICK_DIFFERENCE / 2))\n");
        }

		fprintf(file, "%s%%define 	GO4K_USE_ENVELOPE_RECORDINGS\n",useenvlevels?"":";");
		fprintf(file, "%s%%define 	GO4K_USE_NOTE_RECORDINGS\n",useenotevalues?"":";");
	if (undenormalize)
		fprintf(file, "%%define 	GO4K_USE_UNDENORMALIZE\n");
	if (clipoutput)
		fprintf(file, "%%define 	GO4K_CLIP_OUTPUT\n");
	if (uses.dst_use)
		fprintf(file, "%%define 	GO4K_USE_DST\n");
	if (uses.dll_use)
		fprintf(file, "%%define 	GO4K_USE_DLL\n");
	if (uses.pan_use)
		fprintf(file, "%%define 	GO4K_USE_PAN\n");
		fprintf(file, "%%define 	GO4K_USE_GLOBAL_DLL\n"); // "global dll" aka aux
	if (uses.fstg_use)
		fprintf(file, "%%define 	GO4K_USE_FSTG\n");
	if (uses.fld_use)
		fprintf(file, "%%define 	GO4K_USE_FLD\n");
	// todo: add "opt out" for env check
		fprintf(file, "%%define 	GO4K_USE_ENV_CHECK\n");
		
    if (uses.slw_use)
    {
        fprintf(file, "%%define     GO4K_USE_SLW\n");
        if (uses.slw_calcmode_rms)
            fprintf(file, "%%define     GO4K_USE_SLW_CALCMODE\n");
    }

    if (uses.fldfrom_use)
    {
        fprintf(file, "%%define		GO4K_USE_FLDFROM\n");
    }

	if (uses.env_gm)
		fprintf(file, "%%define 	GO4K_USE_ENV_MOD_GM\n");
	if (uses.env_adrm)
		fprintf(file, "%%define		GO4K_USE_ENV_MOD_ADR\n");
	// todo: add "opt out" for vco check
		fprintf(file, "%%define 	GO4K_USE_VCO_CHECK\n");
	if (uses.vco_phaseofs)
		fprintf(file, "%%define 	GO4K_USE_VCO_PHASE_OFFSET\n");
	if (uses.vco_shape)
		fprintf(file, "%%define 	GO4K_USE_VCO_SHAPE\n");
	if (uses.vco_gate)
		fprintf(file, "%%define		GO4K_USE_VCO_GATE\n");
	if (uses.vco_fm)
		fprintf(file, "%%define 	GO4K_USE_VCO_MOD_FM\n");
	if (uses.vco_pm)
		fprintf(file, "%%define 	GO4K_USE_VCO_MOD_PM\n");
	if (uses.vco_tm)
		fprintf(file, "%%define 	GO4K_USE_VCO_MOD_TM\n");
	if (uses.vco_dm)
		fprintf(file, "%%define 	GO4K_USE_VCO_MOD_DM\n");
	if (uses.vco_cm)
		fprintf(file, "%%define 	GO4K_USE_VCO_MOD_CM\n");
	if (uses.vco_gm)
		fprintf(file, "%%define 	GO4K_USE_VCO_MOD_GM\n");
	if (uses.vco_sm)
		fprintf(file, "%%define 	GO4K_USE_VCO_MOD_SM\n");
		fprintf(file, "%%define 	GO4K_USE_VCF_CHECK\n");
	if (uses.vcf_fm)
		fprintf(file, "%%define 	GO4K_USE_VCF_MOD_FM\n");
	if (uses.vcf_rm)
		fprintf(file, "%%define 	GO4K_USE_VCF_MOD_RM\n");
		fprintf(file, "%%define 	GO4K_USE_VCF_HIGH\n");
		fprintf(file, "%%define 	GO4K_USE_VCF_BAND\n");
		fprintf(file, "%%define 	GO4K_USE_VCF_PEAK\n");
		fprintf(file, "%%define 	GO4K_USE_DST_CHECK\n");
	if (uses.dst_snh)
		fprintf(file, "%%define 	GO4K_USE_DST_SH\n");
	if (uses.dst_dm)
		fprintf(file, "%%define 	GO4K_USE_DST_MOD_DM\n");
	if (uses.dst_sm)
		fprintf(file, "%%define 	GO4K_USE_DST_MOD_SH\n");
	if (uses.dll_notesync)
		fprintf(file, "%%define		GO4K_USE_DLL_NOTE_SYNC\n");
	if (uses.dll_chorus)
		fprintf(file, "%%define		GO4K_USE_DLL_CHORUS\n");
		fprintf(file, "%%define		GO4K_USE_DLL_CHORUS_CLAMP\n");
	if (uses.dll_damp)
		fprintf(file, "%%define 	GO4K_USE_DLL_DAMP\n");
		fprintf(file, "%%define 	GO4K_USE_DLL_DC_FILTER\n");
	if (uses.fop_loadnote)
		fprintf(file, "%%define		GO4K_USE_FOP_LOADNOTE\n");
		fprintf(file, "%%define 	GO4K_USE_FSTG_CHECK\n");
	if (uses.pan_pm)
		fprintf(file, "%%define 	GO4K_USE_PAN_MOD\n");
	if (uses.out_am)
		fprintf(file, "%%define 	GO4K_USE_OUT_MOD_AM\n");
	if (uses.out_gm)
		fprintf(file, "%%define 	GO4K_USE_OUT_MOD_GM\n");
		fprintf(file, "%%define		GO4K_USE_WAVESHAPER_CLIP\n");
	if (uses.fld_vm)
		fprintf(file, "%%define 	GO4K_USE_FLD_MOD_VM\n");
	if (uses.dll_use_mod)
		fprintf(file, "%%define 	GO4K_USE_DLL_MOD\n");
	if (uses.dll_pm)
		fprintf(file, "%%define 	GO4K_USE_DLL_MOD_PM\n");
	if (uses.dll_fm)
		fprintf(file, "%%define 	GO4K_USE_DLL_MOD_FM\n");
	if (uses.dll_im)
		fprintf(file, "%%define 	GO4K_USE_DLL_MOD_IM\n");	
	if (uses.dll_dm)
		fprintf(file, "%%define 	GO4K_USE_DLL_MOD_DM\n");
	if (uses.dll_sm)
		fprintf(file, "%%define 	GO4K_USE_DLL_MOD_SM\n");
	if (uses.dll_am)
		fprintf(file, "%%define 	GO4K_USE_DLL_MOD_AM\n");

; //----------------------------------------------------------------------------------------
; // synth defines
; //----------------------------------------------------------------------------------------

		fprintf(file, "%%define	MAX_DELAY			65536\n");
        fprintf(file, "%%define	MAX_UNITS 32\n");
        fprintf(file, "; \"MAX_WORKSPACE_SLOTS\" is called \"MAX_UNIT_SLOTS\" in 4klang 3.11\n");
fprintf(file, ";// in 4klang 3.0.1 and 3.0.0, MAX_WORKSPACE_SLOTS was 8 in the 4k export asm file, but MAX_SLOT_VALUES was 32 on the VSTi/CPP side\n");
fprintf(file, ";//      and the number 256 was hard-coded as the size of the workspace. (32 units * 8 slots per unit)\n");
fprintf(file, ";// in 4klang 3.1.1, \"MAX_UNIT_SLOTS\" is 16 both in the VST plugin and in the 4k export\n");
fprintf(file, ";// in Forked Klang 3.0.1, MAX_WORKSPACE_SLOTS is 16 in the VST plugin, and 8 in the 4k export\n");
        fprintf(file, "%%define MAX_WORKSPACE_SLOTS	8\n");
		fprintf(file, "%%define GO4K_BEGIN_CMDDEF(def_name)\n");
		fprintf(file, "%%define GO4K_END_CMDDEF db 0\n");
		fprintf(file, "%%define GO4K_BEGIN_PARAMDEF(def_name)\n");
		fprintf(file, "%%define GO4K_END_PARAMDEF\n");

		// fprintf(file, "GO4K_ENV_ID		equ		1\n");
        fprintf(file, "%%define GO4K_ENV_ID ((go4k_synth_cmd_ENV_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_ENV 5\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "	db	%%2\n");
		fprintf(file, "	db	%%3\n");
		fprintf(file, "	db	%%4\n");
		fprintf(file, "	db	%%5\n");
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define	ATTAC(val)		val	\n");
		fprintf(file, "%%define	DECAY(val)		val	\n");
		fprintf(file, "%%define	SUSTAIN(val)	val	\n");
		fprintf(file, "%%define	RELEASE(val)	val	\n");
		fprintf(file, "%%define	GAIN(val)		val	\n");
		fprintf(file, "struc	go4kENV_val\n");
		fprintf(file, "	.attac		resd	1\n");
		fprintf(file, "	.decay		resd	1\n");
		fprintf(file, "	.sustain	resd	1\n");
		fprintf(file, "	.release	resd	1\n");
		fprintf(file, "	.gain		resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kENV_wrk\n");
		fprintf(file, "	.state		resd	1\n");
		fprintf(file, "	.level		resd	1\n");
		fprintf(file, "	.gm			resd	1\n");
		fprintf(file, "	.am			resd	1\n");
		fprintf(file, "	.dm			resd	1\n");
		fprintf(file, "	.sm			resd	1\n");
		fprintf(file, "	.rm			resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "%%define ENV_STATE_ATTAC		0\n");
		fprintf(file, "%%define ENV_STATE_DECAY		1\n");
		fprintf(file, "%%define ENV_STATE_SUSTAIN	2\n");
		fprintf(file, "%%define ENV_STATE_RELEASE	3\n");
		fprintf(file, "%%define ENV_STATE_OFF		4\n");

//		fprintf(file, "GO4K_VCO_ID		equ		2\n");
        fprintf(file, "%%define GO4K_VCO_ID ((go4k_synth_cmd_VCO_func_ptr - go4k_synth_commands) / 4)\n");

		fprintf(file, "%%macro	GO4K_VCO 8\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "	db	%%2\n");
		fprintf(file, "%%ifdef GO4K_USE_VCO_PHASE_OFFSET	\n");
		fprintf(file, "	db	%%3\n");
		fprintf(file, "%%endif	\n");
		fprintf(file, "%%ifdef GO4K_USE_VCO_GATE	\n");
		fprintf(file, "	db	%%4\n");
		fprintf(file, "%%endif	\n");
		fprintf(file, "	db	%%5\n");
		fprintf(file, "%%ifdef GO4K_USE_VCO_SHAPE	\n");
		fprintf(file, "	db	%%6\n");
		fprintf(file, "%%endif	\n");
		fprintf(file, "	db	%%7\n");
		fprintf(file, "	db	%%8\n");
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define	TRANSPOSE(val)	val	\n");
		fprintf(file, "%%define	DETUNE(val)		val	\n");
		fprintf(file, "%%define	PHASE(val)		val	\n");
		fprintf(file, "%%define	GATES(val)		val	\n");
		fprintf(file, "%%define	COLOR(val)		val	\n");
		fprintf(file, "%%define	SHAPE(val)		val \n");
		fprintf(file, "%%define	FLAGS(val)		val	\n");
		fprintf(file, "%%define SINE		0x01\n");
		fprintf(file, "%%define TRISAW		0x02\n");
		fprintf(file, "%%define PULSE		0x04\n");
		fprintf(file, "%%define NOISE		0x08\n");
		fprintf(file, "%%define LFO			0x10\n");
		fprintf(file, "%%define GATE		0x20\n");
		fprintf(file, "struc	go4kVCO_val\n");
		fprintf(file, "	.transpose	resd	1\n");
		fprintf(file, "	.detune		resd	1\n");
		fprintf(file, "%%ifdef GO4K_USE_VCO_PHASE_OFFSET	\n");
		fprintf(file, "	.phaseofs	resd	1\n");
		fprintf(file, "%%endif	\n");	
		fprintf(file, "%%ifdef GO4K_USE_VCO_GATE	\n");
		fprintf(file, "	.gate		resd	1\n");
		fprintf(file, "%%endif	\n");
		fprintf(file, "	.color		resd	1\n");
		fprintf(file, "%%ifdef GO4K_USE_VCO_SHAPE	\n");
		fprintf(file, "	.shape		resd	1\n");
		fprintf(file, "%%endif	\n");
		fprintf(file, "	.gain		resd	1\n");
		fprintf(file, "	.flags		resd	1	\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kVCO_wrk\n");
		fprintf(file, "	.phase		resd	1\n");
		fprintf(file, "	.tm			resd	1\n");
		fprintf(file, "	.dm			resd	1\n");
		fprintf(file, "	.fm			resd	1\n");
		fprintf(file, "	.pm			resd	1\n");
		fprintf(file, "	.cm			resd	1\n");
		fprintf(file, "	.sm			resd	1\n");
		fprintf(file, "	.gm			resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");

//		fprintf(file, "GO4K_VCF_ID		equ		3\n");
        fprintf(file, "%%define GO4K_VCF_ID ((go4k_synth_cmd_VCF_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_VCF 3\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "	db	%%2\n");
		fprintf(file, "	db	%%3\n");
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define LOWPASS		0x1\n");
		fprintf(file, "%%define HIGHPASS	0x2\n");
		fprintf(file, "%%define BANDPASS	0x4\n");
		fprintf(file, "%%define	BANDSTOP	0x3\n");
		fprintf(file, "%%define ALLPASS		0x7\n");
		fprintf(file, "%%define	PEAK		0x8\n");
		fprintf(file, "%%define	FREQUENCY(val)	val\n");
		fprintf(file, "%%define	RESONANCE(val)	val\n");
		fprintf(file, "%%define	VCFTYPE(val)	val\n");
		fprintf(file, "struc	go4kVCF_val\n");
		fprintf(file, "	.freq		resd	1\n");
		fprintf(file, "	.res		resd	1\n");
		fprintf(file, "	.type		resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kVCF_wrk\n");
		fprintf(file, "	.low		resd	1\n");
		fprintf(file, "	.high		resd	1\n");
		fprintf(file, "	.band		resd	1\n");
		fprintf(file, "	.freq		resd	1\n");
		fprintf(file, "	.fm			resd	1\n");
		fprintf(file, "	.rm			resd	1	\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");

//		fprintf(file, "GO4K_DST_ID		equ		4\n");
        fprintf(file, "%%define GO4K_DST_ID ((go4k_synth_cmd_DST_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_DST 2\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "%%ifdef GO4K_USE_DST_SH	\n");
		fprintf(file, "	db	%%2\n");
		fprintf(file, "%%endif	\n");
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define	DRIVE(val)		val\n");
		fprintf(file, "%%define	SNHFREQ(val)	val\n");
		fprintf(file, "struc	go4kDST_val\n");
		fprintf(file, "	.drive		resd	1\n");
		fprintf(file, "%%ifdef GO4K_USE_DST_SH	\n");
		fprintf(file, "	.snhfreq	resd	1\n");
		fprintf(file, "%%endif	\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kDST_wrk\n");
		fprintf(file, "%%ifdef GO4K_USE_DST_SH	\n");
		fprintf(file, "	.out		resd	1\n");
		fprintf(file, "	.snhphase	resd	1\n");
		fprintf(file, "%%endif	\n");
		fprintf(file, "	.dm			resd	1\n");
		fprintf(file, "	.sm			resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");

//		fprintf(file, "GO4K_DLL_ID		equ		5\n");
        fprintf(file, "%%define GO4K_DLL_ID ((go4k_synth_cmd_DLL_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_DLL 8\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "	db	%%2\n");
		fprintf(file, "	db	%%3\n");
		fprintf(file, "%%ifdef GO4K_USE_DLL_DAMP	\n");
		fprintf(file, "	db	%%4\n");
		fprintf(file, "%%endif		\n");
		fprintf(file, "%%ifdef GO4K_USE_DLL_CHORUS	\n");
		fprintf(file, "	db	%%5\n");
		fprintf(file, "	db	%%6\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "	db	%%7\n");
		fprintf(file, "	db	%%8\n");
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define PREGAIN(val)	val\n");
		fprintf(file, "%%define	DRY(val)		val\n");
		fprintf(file, "%%define	FEEDBACK(val)	val\n");
		fprintf(file, "%%define	DEPTH(val)		val\n");
		fprintf(file, "%%define DAMP(val)		val\n");
		fprintf(file, "%%define	DELAY(val)		val\n");
		fprintf(file, "%%define	COUNT(val)		val\n");
		fprintf(file, "struc	go4kDLL_val\n");
		fprintf(file, "	.pregain	resd	1\n");
		fprintf(file, "	.dry		resd	1\n");
		fprintf(file, "	.feedback	resd	1\n");
		fprintf(file, "%%ifdef GO4K_USE_DLL_DAMP	\n");
		fprintf(file, "	.damp		resd	1	\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "%%ifdef GO4K_USE_DLL_CHORUS\n");
		fprintf(file, "	.freq		resd	1\n");
		fprintf(file, "	.depth\n");
		fprintf(file, "%%endif\n");	
		fprintf(file, "	.delay		resd	1\n");
		fprintf(file, "	.count		resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kDLL_wrk\n");
		fprintf(file, "	.index		resd	1\n");
		fprintf(file, "	.store		resd	1\n");
		fprintf(file, "	.dcin		resd	1\n");
		fprintf(file, "	.dcout		resd	1\n");
		fprintf(file, "%%ifdef GO4K_USE_DLL_CHORUS\n");
		fprintf(file, "	.phase		resd	1\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "	.buffer		resd	MAX_DELAY\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kDLL_wrk2\n");
		fprintf(file, " .pm			resd	1\n");
		fprintf(file, " .fm			resd	1\n");
		fprintf(file, " .im			resd	1\n");	
		fprintf(file, " .dm			resd	1\n");
		fprintf(file, " .sm			resd	1\n");
		fprintf(file, " .am			resd	1\n");
		fprintf(file, " .size\n");
		fprintf(file, "endstruc\n");

//		fprintf(file, "GO4K_FOP_ID	equ			6\n");
        fprintf(file, "%%define GO4K_FOP_ID ((go4k_synth_cmd_FOP_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_FOP 1\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define	OP(val)			val\n");
		fprintf(file, "%%define FOP_POP			0x1\n");
		fprintf(file, "%%define FOP_ADDP		0x2\n");
		fprintf(file, "%%define FOP_MULP		0x3\n");
		fprintf(file, "%%define FOP_PUSH		0x4\n");
		fprintf(file, "%%define FOP_XCH			0x5\n");
		fprintf(file, "%%define FOP_ADD			0x6\n");
		fprintf(file, "%%define FOP_MUL			0x7\n");
		fprintf(file, "%%define FOP_ADDP2		0x8\n");
		fprintf(file, "%%define FOP_LOADNOTE	0x9\n");
		fprintf(file, "struc	go4kFOP_val\n");
		fprintf(file, "	.flags		resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kFOP_wrk\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");


        fprintf(file, "%%define GO4K_FST_ID ((go4k_synth_cmd_FST_func_ptr - go4k_synth_commands) / 4)\n");
        fprintf(file, "%%macro	GO4K_FST 4\n");
        fprintf(file, "	db	%%1\n");
        fprintf(file, "	db	%%2\n");
        fprintf(file, "	db  %%3\n");
        fprintf(file, "	dw	%%4\n");
        fprintf(file, "%%endmacro\n");

        //; In 4klang 3.11, the DEST parameter of the GO4K_FST macro was changed from a byte (db)
        //; to a word (dw), because the number of units per instrument was increased from 32 to 64,
        //; and the number of in-unit workspace slots from 8 to 16, and 64*16 doesn't fit in a byte.

        fprintf(file, "%%define	AMOUNT(val)		val\n");
        fprintf(file, "%%define	CONDITION(val)	val\n");
        fprintf(file, "%%define	OPERATION(val)	val\n");
        fprintf(file, "%%define	DEST(val)		val\n");

        fprintf(file, "%%define	FST_CONDITION_ALWAYS    0x00\n");
        fprintf(file, "%%define FST_CONDITION_SMALLER_THAN_PREVIOUS			0x01\n");

        fprintf(file, "%%define	FST_OPERATION_SET			0x00\n");
        fprintf(file, "%%define	FST_OPERATION_ADD			0x40\n");
        fprintf(file, "%%define	FST_OPERATION_POP			0x80\n");

        fprintf(file, "%%define	FST_CONDITION_ALWAYS			0x00\n");
        fprintf(file, "%%define	FST_CONDITION_IF_SMALLER		0x01\n");
        fprintf(file, "%%define	FST_CONDITION_IF_LARGER			0x02\n");
        fprintf(file, "%%define	FST_CONDITION_IF_SIGN_CHANGED   0x04\n");

        fprintf(file, "; The _val strucs map the TRANSFORMED, i.e. float values expanded from the byte data specified using the GO4K_UNIT macros.\n");
        fprintf(file, "struc	go4kFST_val\n");
        fprintf(file, "	.amount		resd	1\n");
        fprintf(file, "	.size\n");
        fprintf(file, "endstruc\n");
        fprintf(file, "; go4kFST_val.size = 1*4\n");

        fprintf(file, "struc	go4kFST_wrk\n");
        fprintf(file, "    .previn     resd    1\n");
        fprintf(file, "	.size\n");
        fprintf(file, "endstruc\n");
        fprintf(file, "; go4kFST_wrk.size = 1*4\n");




//		fprintf(file, "GO4K_PAN_ID		equ		8\n");
        fprintf(file, "%%define GO4K_PAN_ID ((go4k_synth_cmd_PAN_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_PAN 1\n");
		fprintf(file, "%%ifdef GO4K_USE_PAN\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define	PANNING(val)	val\n");
		fprintf(file, "struc	go4kPAN_val\n");
		fprintf(file, "%%ifdef GO4K_USE_PAN\n");
		fprintf(file, "	.panning	resd	1\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kPAN_wrk\n");
		fprintf(file, "	.pm			resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");

//		fprintf(file, "GO4K_OUT_ID		equ		9\n");
        fprintf(file, "%%define GO4K_OUT_ID ((go4k_synth_cmd_OUT_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_OUT 2\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "%%ifdef GO4K_USE_GLOBAL_DLL	\n");
		fprintf(file, "	db	%%2\n");
		fprintf(file, "%%endif	\n");
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define	AUXSEND(val)	val\n");
		fprintf(file, "struc	go4kOUT_val\n");
		fprintf(file, "	.gain		resd	1\n");
		fprintf(file, "%%ifdef GO4K_USE_GLOBAL_DLL	\n");
		fprintf(file, "	.auxsend	resd	1\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kOUT_wrk\n");
		fprintf(file, "	.am			resd	1\n");
		fprintf(file, "	.gm			resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");

//		fprintf(file, "GO4K_ACC_ID		equ		10\n");
        fprintf(file, "%%define GO4K_ACC_ID ((go4k_synth_cmd_ACC_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_ACC 1\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define OUTPUT			0\n");
		fprintf(file, "%%define	AUX				8\n");
		fprintf(file, "%%define ACCTYPE(val)	val\n");
		fprintf(file, "struc	go4kACC_val\n");
		fprintf(file, "	.acctype	resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kACC_wrk\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");

		fprintf(file, "%%ifdef GO4K_USE_FLD\n");
//		fprintf(file, "GO4K_FLD_ID	equ		11\n");
        fprintf(file, "%%define GO4K_FLD_ID ((go4k_synth_cmd_FLD_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_FLD 1\n");
		fprintf(file, "	db	%%1\n"); // bug fix from 3.01
		fprintf(file, "%%endmacro\n");
		fprintf(file, "%%define	VALUE(val)	val\n");
		fprintf(file, "struc	go4kFLD_val\n");
		fprintf(file, "	.value		resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kFLD_wrk\n");
		fprintf(file, "	.vm			resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "%%endif\n");

		fprintf(file, "%%ifdef GO4K_USE_FSTG\n");
//		fprintf(file, "GO4K_FSTG_ID	equ		12\n");
        fprintf(file, "%%define GO4K_FSTG_ID ((go4k_synth_cmd_FSTG_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "%%macro	GO4K_FSTG 4\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "	db	%%2\n");
		fprintf(file, "	db	%%3\n");
		fprintf(file, "	dw	%%4\n"); // %2 used to be dd in older 4klangs
		fprintf(file, "%%endmacro\n");
		fprintf(file, "struc	go4kFSTG_val\n");
		fprintf(file, "	.amount		resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kFSTG_wrk\n");
		fprintf(file, "	.previn\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "%%endif\n");

		fprintf(file, "%%ifdef GO4K_USE_SLW\n");
		fprintf(file, "; //----------------------------------------------------------------------------------------\n");
		fprintf(file, "; // SLW structs\n");
		fprintf(file, "; //----------------------------------------------------------------------------------------\n");
		fprintf(file, "%%define GO4K_SLW_ID ((go4k_synth_cmd_SLW_func_ptr - go4k_synth_commands) / 4)\n");
		fprintf(file, "\n");
		fprintf(file, "%%macro	GO4K_SLW 2\n");
		fprintf(file, "	db	%%1\n");
		fprintf(file, "	db	%%2\n");
		fprintf(file, "%%endmacro\n");

		fprintf(file, "%%define	DELAYIDX(val)		val\n");
		fprintf(file, "%%define	CALCMODE(val)		val\n");

        fprintf(file, "%%define SLW_CALCMODE_LINEAR 0\n");
        fprintf(file, "%%define SLW_CALCMODE_RMS    1\n");

		fprintf(file, "struc	go4kSLW_val\n");
		fprintf(file, "	.delayidx	resd	1\n");
		fprintf(file, "	.calcmode	resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kSLW_wrk\n");
		fprintf(file, "	.index		resd	1\n");
		fprintf(file, "	.store		resd	1\n");
		fprintf(file, "	.dcin		resd	1\n");
		fprintf(file, "	.dcout		resd	1\n");
		fprintf(file, "	.bufsum		resd	1\n");
		fprintf(file, "	.buffer		resd	MAX_DELAY\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "struc	go4kSLW_wrk2\n"); 
		fprintf(file, " .mitamitamitadm			resd	1\n");
 		fprintf(file, " .size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "%%endif\n"); // 



		fprintf(file, "; //----------------------------------------------------------------------------------------\n");
		fprintf(file, "; // FLDFROM structs\n");
		fprintf(file, "; //----------------------------------------------------------------------------------------\n");
		fprintf(file, "%%ifdef GO4K_USE_FLDFROM\n");
		fprintf(file, "%%define GO4K_FLDFROM_ID ((go4k_synth_cmd_FLDFROM_func_ptr - go4k_synth_commands) / 4)\n");

		fprintf(file, "%%macro	GO4K_FLDFROM 1\n");
		fprintf(file, " dw	%%1\n");
		fprintf(file, "%%endmacro\n");

		fprintf(file, "%%define	SOURCE(val)		val\n");

		fprintf(file, "struc	go4kFLDFROM_val\n");
        fprintf(file, ".dummy1	resd	1\n");
		fprintf(file, ".size\n");
        fprintf(file, "endstruc\n");

		fprintf(file, "struc	go4kFLDFROM_wrk\n");
		fprintf(file, ".size\n");
		fprintf(file, "endstruc\n");
		fprintf(file, "%%endif\n");


		fprintf(file, "struc	go4k_instrument\n");
		fprintf(file, "	.release	resd	1\n");
		fprintf(file, "	.note		resd	1\n");
		fprintf(file, "	.workspace	resd	MAX_UNITS*MAX_WORKSPACE_SLOTS\n");
		fprintf(file, "	.dlloutl	resd	1\n");
		fprintf(file, "	.dlloutr	resd	1\n");
		fprintf(file, "	.outl		resd	1\n");
		fprintf(file, "	.outr		resd	1\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");

		fprintf(file, "struc	go4k_synth\n");
		fprintf(file, "	.instruments	resb	go4k_instrument.size * MAX_INSTRUMENTS * MAX_VOICES\n");
		fprintf(file, "	.global			resb	go4k_instrument.size * MAX_VOICES\n");
		fprintf(file, "	.size\n");
		fprintf(file, "endstruc\n");
#define INDEXED_PATTERN_VALUES__
#ifdef INDEXED_PATTERN_VALUES
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// the pattern values
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		fprintf(file, "; //----------------------------------------------------------------------------------------\n");
		fprintf(file, "; // Pattern Values, sorted by quantity\n");
		fprintf(file, "; //----------------------------------------------------------------------------------------\n");
		fprintf(file, "%%ifdef USE_SECTIONS\n");
		fprintf(file, "section .g4kmuc0 data align=1\n");
		fprintf(file, "%%else\n");
		fprintf(file, "section .data\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "go4k_pattern_values\n");
		std::map<int, int> noteCountMap;
		for (int i = 0; i < ReducedPatterns.size()/PatternSize; i++)
		{
			for (int j = 0; j < PatternSize; j++)
			{
				int note = ReducedPatterns[i*PatternSize+j];
				if (note == -1)
					note = 1;
				if (noteCountMap.find(note) != noteCountMap.end())
				{
					noteCountMap[note] = noteCountMap[note] + 1;
				}
				else
				{
					noteCountMap[note] = 1;
				}
			}
		}
		std::list<std::pair<int, int> > noteIndexList;
		std::map<int, int>::iterator it;
		for (it = noteCountMap.begin(); it != noteCountMap.end(); it++)
		{
			std::pair<int, int> sval(it->second, it->first);
			noteIndexList.push_back(sval);
		}
		noteIndexList.sort();
		std::list<std::pair<int, int> >::reverse_iterator rit;
		fprintf(file, "\tdb\t");
		for (rit = noteIndexList.rbegin(); rit != noteIndexList.rend(); rit++)
		{
			fprintf(file, "%d, ", (*rit).second);
		}
#endif
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// the patterns
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		fprintf(file, "%%ifdef USE_SECTIONS\n");
		fprintf(file, "section .g4kmuc1 data align=1\n");
		fprintf(file, "%%else\n");
		fprintf(file, "section .data align=1\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "go4k_patterns\n");
		for (int i = 0; i < ReducedPatterns.size()/PatternSize; i++)
		{
			fprintf(file, "\tdb\t");
			for (int j = 0; j < PatternSize; j++)
			{
#ifdef INDEXED_PATTERN_VALUES
				int noteindex = 0;
				for (rit = noteIndexList.rbegin(); rit != noteIndexList.rend(); rit++)
				{
					// TODO: check for hold
					if (ReducedPatterns[i*PatternSize+j] == (*rit).second)
						fprintf(file, "%d, ", noteindex);
					noteindex++;
				}
#else
				if (ReducedPatterns[i*PatternSize+j] >= 0)
					fprintf(file, "%d, ", ReducedPatterns[i*PatternSize+j]);
				else
					fprintf(file, "HLD, ");
#endif
			}		
			fprintf(file, "\n");			
		}
		fprintf(file, "go4k_patterns_end\n");
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// the pattern indices
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		fprintf(file, "%%ifdef USE_SECTIONS\n");
		fprintf(file, "section .g4kmuc2 data align=1\n");
		fprintf(file, "%%else\n");
		fprintf(file, "section .data\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "go4k_pattern_lists\n");
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			if (!InstrumentUsed[i]) continue;
			fprintf(file, "Instrument%dList\t\tdb\t", i);
			for (int j = 0; j < PatternIndices[i].size(); j++)
			{
				fprintf(file, "%d, ", PatternIndices[i][j]);
			}
			fprintf(file, "\n");
		}
		fprintf(file, "go4k_pattern_lists_end\n");
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// the instrument commands
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		fprintf(file, "%%ifdef USE_SECTIONS\n");
		fprintf(file, "section .g4kmuc3 data align=1\n");
		fprintf(file, "%%else\n");
		fprintf(file, "section .data\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "go4k_synth_instructions\n");
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			if (!InstrumentUsed[i]) continue;
			fprintf(file, "GO4K_BEGIN_CMDDEF(Instrument%d)\n", i);
			for (int u = 0; u < MAX_UNITS; u++)
			{
                switch(SynthObj.InstrumentValues[i][u][0])
                {
                    case M_ENV:
					    fprintf(file, "\tdb GO4K_ENV_ID\n");
                        break;
                    case M_VCO:
    					fprintf(file, "\tdb GO4K_VCO_ID\n");
                        break;
                    case M_VCF:
					    fprintf(file, "\tdb GO4K_VCF_ID\n");
                        break;
                    case M_DST:
					    fprintf(file, "\tdb GO4K_DST_ID\n");
                        break;
                    case M_DLL:
					    fprintf(file, "\tdb GO4K_DLL_ID\n");
                        break;
                    case M_FOP:
					    fprintf(file, "\tdb GO4K_FOP_ID\n");
                        break;
                    case M_FSTx:
                        {
					        FSTx_valP v = (FSTx_valP)(SynthObj.InstrumentValues[i][u]);
					        // local storage
					        if (v->dest_stack == -1 || v->dest_stack == i)
						        fprintf(file, "\tdb GO4K_FST_ID\n");
					        // global storage
					        else
						        fprintf(file, "\tdb GO4K_FSTG_ID\n");
                        }
				        break;
                    case M_PAN:
					    fprintf(file, "\tdb GO4K_PAN_ID\n");
                        break;
                    case M_OUT:
					    fprintf(file, "\tdb GO4K_OUT_ID\n");
                        break;
                    case M_ACC:
					    fprintf(file, "\tdb GO4K_ACC_ID\n");
                        break;
                    case M_FLD:
					    fprintf(file, "\tdb GO4K_FLD_ID\n");
                        break;
                    case M_SLW:
					    fprintf(file, "\tdb GO4K_SLW_ID\n");
                        break;
                    case M_FLDFROM:
                        fprintf(file, "\tdb GO4K_FLDFROM_ID\n");
                        break;
                    case M_NONE:
                        break;
                    default:
                        fprintf(file, "\tERROR! UNKNOWN UNIT ID %d IN INSTRUMENT %d SLOT %d\n", SynthObj.InstrumentValues[i][u][0], i, u);
                }
			}
			fprintf(file, "GO4K_END_CMDDEF\n");
		};
		fprintf(file, "GO4K_BEGIN_CMDDEF(Global)\n");
		for (int u = 0; u < MAX_UNITS; u++)
		{
			if (SynthObj.GlobalValues[u][0] == M_ENV)
				fprintf(file, "\tdb GO4K_ENV_ID\n");
			else if (SynthObj.GlobalValues[u][0] == M_VCO)
				fprintf(file, "\tdb GO4K_VCO_ID\n");
			else if (SynthObj.GlobalValues[u][0] == M_VCF)
				fprintf(file, "\tdb GO4K_VCF_ID\n");
			else if (SynthObj.GlobalValues[u][0] == M_DST)
				fprintf(file, "\tdb GO4K_DST_ID\n");
			else if (SynthObj.GlobalValues[u][0] == M_DLL)
				fprintf(file, "\tdb GO4K_DLL_ID\n");
			else if (SynthObj.GlobalValues[u][0] == M_FOP)
				fprintf(file, "\tdb GO4K_FOP_ID\n");
			else if (SynthObj.GlobalValues[u][0] == M_FSTx)
			{
				FSTx_valP v = (FSTx_valP)(SynthObj.GlobalValues[u]);
				// local storage
				if (v->dest_stack == -1 || v->dest_stack == MAX_INSTRUMENTS)
					fprintf(file, "\tdb GO4K_FST_ID\n");
				// global storage
				else
					fprintf(file, "\tdb GO4K_FSTG_ID\n");
			}
			else if (SynthObj.GlobalValues[u][0] == M_PAN)
				fprintf(file, "\tdb GO4K_PAN_ID\n");
			else if (SynthObj.GlobalValues[u][0] == M_OUT)
				fprintf(file, "\tdb GO4K_OUT_ID\n");
			else if (SynthObj.GlobalValues[u][0] == M_ACC)
				fprintf(file, "\tdb GO4K_ACC_ID\n");
			else if (SynthObj.GlobalValues[u][0] == M_FLD)
				fprintf(file, "\tdb GO4K_FLD_ID\n");
            else if (SynthObj.GlobalValues[u][0] == M_FLDFROM)
                fprintf(file, "\tdb GO4K_FLDFROM_ID\n");
            else if (SynthObj.GlobalValues[u][0] != M_NONE)
                fprintf(file, "\tERROR! UNKNOWN UNIT ID %d IN GLOBAL SLOT %d\n", SynthObj.GlobalValues[u][0], u);

		}
		fprintf(file, "GO4K_END_CMDDEF\n");
		fprintf(file, "go4k_synth_instructions_end\n");
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// the instrument data
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		fprintf(file, "%%ifdef USE_SECTIONS\n");
		fprintf(file, "section .g4kmuc4 data align=1\n");
		fprintf(file, "%%else\n");
		fprintf(file, "section .data\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "go4k_synth_parameter_values\n");
		int delayindex = 0;
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			if (!InstrumentUsed[i]) continue;
			fprintf(file, "GO4K_BEGIN_PARAMDEF(Instrument%d)\n", i);
			for (int u = 0; u < MAX_UNITS; u++)
			{
                SOME_UNIT_valP unitval = (SOME_UNIT_valP)(SynthObj.InstrumentValues[i][u]);
                if (unitval->id == M_ENV)
				{
					ENV_valP v = (ENV_valP)(unitval);
					fprintf(file, "\tGO4K_ENV\tATTAC(%d),DECAY(%d),SUSTAIN(%d),RELEASE(%d),GAIN(%d)\n", v->attac, v->decay, v->sustain, v->release, v->gain);
				}
				else if (unitval->id == M_VCO)
				{
					VCO_valP v = (VCO_valP)(unitval);
					char type[10]; type[0] = 0;
					char lfo[10]; lfo[0] = 0;
					if (v->flags & VCO_SINE)
						sprintf(type, "SINE");
					if (v->flags & VCO_TRISAW)
						sprintf(type, "TRISAW");
					if (v->flags & VCO_PULSE)
						sprintf(type, "PULSE");
					if (v->flags & VCO_NOISE)
						sprintf(type, "NOISE");
					if (v->flags & VCO_GATE)
						sprintf(type, "GATE");
					if (v->flags & VCO_LFO)
						sprintf(lfo, "|LFO");
					fprintf(file, "\tGO4K_VCO\tTRANSPOSE(%d),DETUNE(%d),PHASE(%d),GATES(%d),COLOR(%d),SHAPE(%d),GAIN(%d),FLAGS(%s%s)\n", v->transpose, v->detune, v->phaseofs, v->gate, v->color, v->shape, v->gain, type, lfo);
				}
				else if (unitval->id == M_VCF)
				{
					VCF_valP v = (VCF_valP)(unitval);
					char type[10]; type[0] = 0;
					if (v->type == VCF_LOWPASS)
						sprintf(type, "LOWPASS");
					if (v->type == VCF_HIGHPASS)
						sprintf(type, "HIGHPASS");
					if (v->type == VCF_BANDPASS)
						sprintf(type, "BANDPASS");
					if (v->type == VCF_BANDSTOP)
						sprintf(type, "BANDSTOP");
					if (v->type == VCF_ALLPASS)
						sprintf(type, "ALLPASS");
					if (v->type == VCF_PEAK)
						sprintf(type, "PEAK");
					fprintf(file, "\tGO4K_VCF\tFREQUENCY(%d),RESONANCE(%d),VCFTYPE(%s)\n", v->freq, v->res, type);
				}
                else if (unitval->id == M_DST)
				{
					DST_valP v = (DST_valP)(unitval);
					fprintf(file, "\tGO4K_DST\tDRIVE(%d), SNHFREQ(%d)\n", v->drive, v->snhfreq);
				}
				else if (unitval->id == M_DLL)
				{
					DLL_valP v = (DLL_valP)(unitval);
					if (v->delayidx < delay_indices.size())
					{
						fprintf(file, "\tGO4K_DLL\tPREGAIN(%d),DRY(%d),FEEDBACK(%d),DAMP(%d),FREQUENCY(%d),DEPTH(%d),DELAY(%d),COUNT(%d)\n", 
							v->pregain, v->dry, v->feedback, v->damp, v->freq, v->depth, delay_indices[v->delayidx], v->count);	
					}
					// error handling in case indices are fucked up
					else
					{
						fprintf(file, "\tGO4K_DLL\tPREGAIN(%d),DRY(%d),FEEDBACK(%d),DAMP(%d),FREQUENCY(%d),DEPTH(%d),DELAY(%d),COUNT(%d) ; ERROR\n", 
							v->pregain, v->dry, v->feedback, v->damp, v->freq, v->depth, v->delayidx, v->count);
					}
				}
                else if (unitval->id == M_FOP)
				{
					FOP_valP v = (FOP_valP)(unitval);
					char type[16]; type[0] = 0;
					if (v->flags == FOP_POP)
						sprintf(type, "FOP_POP");
					if (v->flags == FOP_PUSH)
						sprintf(type, "FOP_PUSH");
					if (v->flags == FOP_XCH)
						sprintf(type, "FOP_XCH");
					if (v->flags == FOP_ADD)
						sprintf(type, "FOP_ADD");
					if (v->flags == FOP_ADDP)
						sprintf(type, "FOP_ADDP");
					if (v->flags == FOP_MUL)
						sprintf(type, "FOP_MUL");
					if (v->flags == FOP_MULP)
						sprintf(type, "FOP_MULP");
					if (v->flags == FOP_ADDP2)
						sprintf(type, "FOP_ADDP2");
					if (v->flags == FOP_LOADNOTE)
						sprintf(type, "FOP_LOADNOTE");
					fprintf(file, "\tGO4K_FOP\tOP(%s)\n", type);
				}
				else if (unitval->id == M_FSTx)
				{
					FSTx_valP v = (FSTx_valP)(unitval);
					// local storage
					if (v->dest_stack == -1 || v->dest_stack == i)
					{
						// skip empty units on the way to target
						int emptySkip = 0;
						for (int e = 0; e < v->dest_unit; e++)
						{
							if (SynthObj.InstrumentValues[i][e][0] == M_NONE)
								emptySkip++;
						}
						std::string operations;
						operations = "FST_OPERATION_SET";
						if (v->operation & FST_OPERATION_ADD)
							operations = "FST_OPERATION_ADD";
						//if (v->operation & FST_OPERATION_MUL)
						//	operations = "FST_MUL";
						if (v->operation & FST_OPERATION_POP)
							operations += "+FST_OPERATION_POP";
                        std::string condition;
                        condition = "FST_CONDITION_ALWAYS";
                        if (v->condition & FST_CONDITION_IF_SMALLER)
                            condition += "+FST_CONDITION_IF_SMALLER";

						fprintf(file, "\tGO4K_FST\tAMOUNT(%d),CONDITION(%s),OPERATION(%s),DEST(%d*MAX_WORKSPACE_SLOTS+%d)\n", 
                            v->amount, condition.c_str(), operations.c_str(), v->dest_unit-emptySkip, v->dest_slot);
					}
					// global storage
					else
					{
						int storestack;
						if (v->dest_stack == MAX_INSTRUMENTS)
							storestack = maxinst;
						else
							storestack = InstrumentIndex[v->dest_stack];
						// skip empty units on the way to target
						int emptySkip = 0;
						for (int e = 0; e < v->dest_unit; e++)
						{
							if (v->dest_stack == MAX_INSTRUMENTS)
							{
								if (SynthObj.GlobalValues[e][0] == M_NONE)
									emptySkip++;
							}
							else
							{
								if (SynthObj.InstrumentValues[v->dest_stack][e][0] == M_NONE)
									emptySkip++;
							}
						}
						// invalid store target, possibly due non usage of the target instrument
						if (storestack == -1)
						{
							fprintf(file, "\tGO4K_FSTG\tAMOUNT(0),CONDITION(0),OPERATION(0),DEST(7*4+go4k_instrument.workspace)\n");
						}
						else
						{
							// fprintf(file, "\tGO4K_FSTG\tAMOUNT(%d),DEST(%d*go4k_instrument.size*MAX_VOICES+%d*MAX_WORKSPACE_SLOTS*4+%d*4+go4k_instrument.workspace)\n", v->amount, storestack, v->dest_unit-emptySkip, v->dest_slot);

						    std::string operations;
						    operations = "FST_OPERATION_SET";
						    if (v->operation & FST_OPERATION_ADD)
							    operations = "FST_OPERATION_ADD";
						    //if (v->operation & FST_OPERATION_MUL)
						    //	operations = "FST_MUL";
						    if (v->operation & FST_OPERATION_POP)
							    operations += "+FST_OPERATION_POP";
                            std::string condition;
                            condition = "FST_CONDITION_ALWAYS";
                            if (v->condition & FST_CONDITION_IF_SMALLER)
                                condition += "+FST_CONDITION_IF_SMALLER";

							fprintf(file, "\tGO4K_FSTG\tAMOUNT(%d),CONDITION(%s),OPERATION(%s),DEST((%d*go4k_instrument.size*MAX_VOICES/4)+(%d*MAX_WORKSPACE_SLOTS+%d)+(go4k_instrument.workspace/4))\n", 
                                    v->amount, condition.c_str(), operations.c_str(), storestack, v->dest_unit-emptySkip, v->dest_slot);
						}
					}
				} // M_FSTx
                else if (unitval->id == M_PAN)
				{
					PAN_valP v = (PAN_valP)(unitval);
					fprintf(file, "\tGO4K_PAN\tPANNING(%d)\n", v->panning);
				}
                else if (unitval->id == M_OUT)
				{
					OUT_valP v = (OUT_valP)(unitval);
					fprintf(file, "\tGO4K_OUT\tGAIN(%d), AUXSEND(%d)\n", v->gain, v->auxsend);
				}
                else if (unitval->id == M_ACC)
				{
					ACC_valP v = (ACC_valP)(unitval);
					if (v->flags == ACC_OUT)
						fprintf(file, "\tGO4K_ACC\tACCTYPE(OUTPUT)\n");
					else
						fprintf(file, "\tGO4K_ACC\tACCTYPE(AUX)\n");
				}
                else if (unitval->id == M_FLD)
				{
					FLD_valP v = (FLD_valP)(unitval);
					fprintf(file, "\tGO4K_FLD\tVALUE(%d)\n", v->value);
				}
                else if (unitval->id == M_SLW)
				{
					SLW_valP v = (SLW_valP)(unitval);
					if (v->delayidx < delay_indices.size())
					{
						fprintf(file, "\tGO4K_SLW\tDELAYIDX(%d),CALCMODE(%d)\n", 
							delay_indices[v->delayidx],
                            v->calcmode
                        );
					}					
					else
					{
						fprintf(file, "\t ERROR IN SLW UNIT, INSTRUMENT %d, UNIT %d\n", i, u);
					}
				}

                else if (unitval->id == M_FLDFROM)
				{
					FLDFROM_valP v = (FLDFROM_valP)(unitval);

					// local storage
					if (v->source_stack == -1 || v->source_stack == i)
					{
						// skip empty units on the way to target
						int emptySkip = 0;
                        for (int e = 0; e < v->source_unit; e++)
						{
							if (SynthObj.InstrumentValues[i][e][0] == M_NONE)
								emptySkip++;
						}
						fprintf(file, "\tGO4K_FLDFROM\tSOURCE(%d*MAX_WORKSPACE_SLOTS+%d)\n", v->source_unit-emptySkip, v->source_slot);
					}
					// global storage
					else
					{
                        // missing feature, todo ...
						fprintf(file, "\tERRRRRORRRRR MISSING FEATURE AAAARGH GO4K_FLDFROM CROSS-INSTRUMENT CASE\n");
					}
				} // M_FLDFROM
                else if (unitval->id == M_NONE)
                { // empty unit, do nothing
                }
                else
                    fprintf(file, "\t UNKNOWN UNIT ID %d, INSTRUMENT %d, UNIT %d\n", unitval->id, i, u);

			}
			fprintf(file, "GO4K_END_PARAMDEF\n");
		}
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// the global data
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		{
			fprintf(file, "GO4K_BEGIN_PARAMDEF(Global)\n");
			for (int u = 0; u < MAX_UNITS; u++)
			{
				if (SynthObj.GlobalValues[u][0] == M_ENV)
				{
					ENV_valP v = (ENV_valP)(SynthObj.GlobalValues[u]);
					fprintf(file, "\tGO4K_ENV\tATTAC(%d),DECAY(%d),SUSTAIN(%d),RELEASE(%d),GAIN(%d)\n", v->attac, v->decay, v->sustain, v->release, v->gain);
				}
				else if (SynthObj.GlobalValues[u][0] == M_VCO)
				{
					VCO_valP v = (VCO_valP)(SynthObj.GlobalValues[u]);
					char type[10]; type[0] = 0;
					char lfo[10]; lfo[0] = 0;
					if (v->flags & VCO_SINE)
						sprintf(type, "SINE");
					if (v->flags & VCO_TRISAW)
						sprintf(type, "TRISAW");
					if (v->flags & VCO_PULSE)
						sprintf(type, "PULSE");
					if (v->flags & VCO_NOISE)
						sprintf(type, "NOISE");
					if (v->flags & VCO_GATE)
						sprintf(type, "GATE");
					if (v->flags & VCO_LFO)
						sprintf(lfo, "|LFO");
					fprintf(file, "\tGO4K_VCO\tTRANSPOSE(%d),DETUNE(%d),PHASE(%d),GATES(%d),COLOR(%d),SHAPE(%d),GAIN(%d),FLAGS(%s%s)\n", v->transpose, v->detune, v->phaseofs, v->gate, v->color, v->shape, v->gain, type, lfo);
				}
				else if (SynthObj.GlobalValues[u][0] == M_VCF)
				{
					VCF_valP v = (VCF_valP)(SynthObj.GlobalValues[u]);
					char type[10]; type[0] = 0;
					if (v->type == VCF_LOWPASS)
						sprintf(type, "LOWPASS");
					if (v->type == VCF_HIGHPASS)
						sprintf(type, "HIGHPASS");
					if (v->type == VCF_BANDPASS)
						sprintf(type, "BANDPASS");
					if (v->type == VCF_BANDSTOP)
						sprintf(type, "BANDSTOP");
					if (v->type == VCF_ALLPASS)
						sprintf(type, "ALLPASS");
					if (v->type == VCF_PEAK)
						sprintf(type, "PEAK");
					fprintf(file, "\tGO4K_VCF\tFREQUENCY(%d),RESONANCE(%d),VCFTYPE(%s)\n", v->freq, v->res, type);
				}
				else if (SynthObj.GlobalValues[u][0] == M_DST)
				{
					DST_valP v = (DST_valP)(SynthObj.GlobalValues[u]);
					fprintf(file, "\tGO4K_DST\tDRIVE(%d), SNHFREQ(%d)\n", v->drive, v->snhfreq);
				}
				else if (SynthObj.GlobalValues[u][0] == M_DLL)
				{
					DLL_valP v = (DLL_valP)(SynthObj.GlobalValues[u]);
					if (v->delayidx < delay_indices.size())
					{
						fprintf(file, "\tGO4K_DLL\tPREGAIN(%d),DRY(%d),FEEDBACK(%d),DAMP(%d),FREQUENCY(%d),DEPTH(%d),DELAY(%d),COUNT(%d)\n", 
							v->pregain, v->dry, v->feedback, v->damp, v->freq, v->depth, delay_indices[v->delayidx], v->count);	
					}
					// error handling in case indices are fucked up
					else
					{
						fprintf(file, "\tGO4K_DLL\tPREGAIN(%d),DRY(%d),FEEDBACK(%d),DAMP(%d),FREQUENCY(%d),DEPTH(%d),DELAY(%d),COUNT(%d) ; ERROR\n", 
							v->pregain, v->dry, v->feedback, v->damp, v->freq, v->depth, v->delayidx, v->count);
					}
				}
				else if (SynthObj.GlobalValues[u][0] == M_FOP)
				{
					FOP_valP v = (FOP_valP)(SynthObj.GlobalValues[u]);
					char type[16]; type[0] = 0;
					if (v->flags == FOP_POP)
						sprintf(type, "FOP_POP");
					if (v->flags == FOP_PUSH)
						sprintf(type, "FOP_PUSH");
					if (v->flags == FOP_XCH)
						sprintf(type, "FOP_XCH");
					if (v->flags == FOP_ADD)
						sprintf(type, "FOP_ADD");
					if (v->flags == FOP_ADDP)
						sprintf(type, "FOP_ADDP");
					if (v->flags == FOP_MUL)
						sprintf(type, "FOP_MUL");
					if (v->flags == FOP_MULP)
						sprintf(type, "FOP_MULP");
					if (v->flags == FOP_ADDP2)
						sprintf(type, "FOP_ADDP2");
					if (v->flags == FOP_LOADNOTE)
						sprintf(type, "FOP_LOADNOTE");
					fprintf(file, "\tGO4K_FOP\tOP(%s)\n", type);
				}
				else if (SynthObj.GlobalValues[u][0] == M_FSTx)
				{
					FSTx_valP v = (FSTx_valP)(SynthObj.GlobalValues[u]);
					// local storage
					if (v->dest_stack == -1 || v->dest_stack == MAX_INSTRUMENTS)
					{
						// skip empty units on the way to target
						int emptySkip = 0;
						for (int e = 0; e < v->dest_unit; e++)
						{
							if (SynthObj.GlobalValues[e][0] == M_NONE)
								emptySkip++;
						}
						fprintf(file, "\tGO4K_FST\tAMOUNT(%d),DEST(%d*MAX_WORKSPACE_SLOTS+%d)\n", v->amount, v->dest_unit-emptySkip, v->dest_slot );
					}
					// global storage
					else
					{
						int storestack = InstrumentIndex[v->dest_stack];
						// skip empty units on the way to target
						int emptySkip = 0;
						for (int e = 0; e < v->dest_unit; e++)
						{
							if (SynthObj.InstrumentValues[v->dest_stack][e][0] == M_NONE)
								emptySkip++;
						}
						// invalid store target, possibly due non usage of the target instrument
						if (storestack == -1)
						{
							fprintf(file, "\tGO4K_FSTG\tAMOUNT(0),DEST(7*4+go4k_instrument.workspace)\n");
						}
						else
						{
							fprintf(file, "\tGO4K_FSTG\tAMOUNT(%d),DEST(%d*go4k_instrument.size*MAX_VOICES+%d*MAX_WORKSPACE_SLOTS*4+%d*4+go4k_instrument.workspace)\n", v->amount, storestack, v->dest_unit-emptySkip, v->dest_slot );
						}
					}
				}
				else if (SynthObj.GlobalValues[u][0] == M_PAN)
				{
					PAN_valP v = (PAN_valP)(SynthObj.GlobalValues[u]);
					fprintf(file, "\tGO4K_PAN\tPANNING(%d)\n", v->panning);
				}
				else if (SynthObj.GlobalValues[u][0] == M_OUT)
				{
					OUT_valP v = (OUT_valP)(SynthObj.GlobalValues[u]);
					fprintf(file, "\tGO4K_OUT\tGAIN(%d), AUXSEND(%d)\n", v->gain, v->auxsend);
				}
				else if (SynthObj.GlobalValues[u][0] == M_ACC)
				{
					ACC_valP v = (ACC_valP)(SynthObj.GlobalValues[u]);
					if (v->flags == ACC_OUT)
						fprintf(file, "\tGO4K_ACC\tACCTYPE(OUTPUT)\n");
					else
						fprintf(file, "\tGO4K_ACC\tACCTYPE(AUX)\n");
				}
				else if (SynthObj.GlobalValues[u][0] == M_FLD)
				{
					FLD_valP v = (FLD_valP)(SynthObj.GlobalValues[u]);
					fprintf(file, "\tGO4K_FLD\tVALUE(%d)\n", v->value);
				}
				else if (SynthObj.GlobalValues[u][0] == M_FLDFROM)
				{
					FLD_valP v = (FLD_valP)(SynthObj.GlobalValues[u]);
					fprintf(file, "\tGO4K_FLD\tVALUE(%d)\n", v->value);
				}
                else if (SynthObj.GlobalValues[u][0] != M_NONE)
                    fprintf(file, "\t ERRRRRRORRRR UNKNOWN OR UNSUPPORTED UNIT ID %d IN GLOBAL STACK\n", SynthObj.GlobalValues[u][0]);
			}
			fprintf(file, "GO4K_END_PARAMDEF\n");
		}
		fprintf(file, "go4k_synth_parameter_values_end\n");
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// delay line times
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// TODO: Reduction here
		fprintf(file, "%%ifdef USE_SECTIONS\n");
		fprintf(file, "section .g4kmuc5 data align=1\n");
		fprintf(file, "%%else\n");
		fprintf(file, "section .data\n");
		fprintf(file, "%%endif\n");
		fprintf(file, "%%ifdef GO4K_USE_DLL\n"); // aaa todo: 
		fprintf(file, "global _go4k_delay_times\n");
		fprintf(file, "_go4k_delay_times\n");
		for (int i = 0; i < delay_times.size(); i++)
		{
			fprintf(file, "\tdw %d\n", delay_times[i]);
		}
		fprintf(file, "%%endif\n");

		fclose(file);
	}
#ifdef EXPORT_OBJECT_FILE
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// extract resources and compile song
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	std::string compfile = path + "c4klang.exe";
	std::string asmfile = path + "4klang.asm";

	// First find and load the nasm compiler resource           
	HRSRC hResource = FindResource(hInst, MAKEINTRESOURCE(IDR_COMP_RES), "BINARY");         
	HGLOBAL hFileResource = LoadResource(hInst, hResource);
	// Now open and map this to a disk file           
	unsigned char *pcompres = (unsigned char *)LockResource(hFileResource);           
	DWORD dwSize = SizeofResource(hInst, hResource);            
	// write content
	FILE *fcompfile = fopen(compfile.c_str(), "wb");
	int tindex = 0;
	while (dwSize--)
	{
		unsigned char r = *pcompres++;
		r -= CryptTable[tindex++];
		fwrite(&r, 1, 1, fcompfile);
	}
	fclose(fcompfile);

	// First find and load the asm resource
	hResource = FindResource(hInst, MAKEINTRESOURCE(IDR_ASM_RES), "BINARY");         
	hFileResource = LoadResource(hInst, hResource);
	// Now open and map this to a disk file           
	unsigned char *pasmres = (unsigned char *)LockResource(hFileResource);           
	dwSize = SizeofResource(hInst, hResource);            
	// write content
	FILE *fasmfile = fopen(asmfile.c_str(), "wb");
	tindex = 0;
	while (dwSize--)
	{
		unsigned char r = *pasmres++;
		r -= CryptTable[tindex++];
		fwrite(&r, 1, 1, fasmfile);
	}
	fclose(fasmfile);
	
	// compile object
	char olddir[1024];
	GetCurrentDirectory(1024, olddir);
	SetCurrentDirectory(path.c_str());

	std::string comparg;
	if (objformat == 0) // windows obj
	{
		comparg = "c4klang.exe 4klang.asm -fwin32 -o\"" + fpath + "\"";
	}
	if (objformat == 1) // linux obj
	{
		comparg = "c4klang.exe 4klang.asm -felf32 -o\"" + fpath + "\"";
	}
	if (objformat == 2) // macosx obj
	{
		comparg = "c4klang.exe 4klang.asm -fmacho32 -o\"" + fpath + "\"";
	}
	STARTUPINFO si;
    PROCESS_INFORMATION pi;
    ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);
    ZeroMemory( &pi, sizeof(pi) );
	si.dwFlags = STARTF_USESHOWWINDOW;
	si.wShowWindow = SW_HIDE;
    // Start the child process. 
    if( !CreateProcess( NULL,   // No module name (use command line). 
        (LPSTR)comparg.c_str(), // Command line. 
        NULL,             // Process handle not inheritable. 
        NULL,             // Thread handle not inheritable. 
        FALSE,            // Set handle inheritance to FALSE. 
        0,                // No creation flags. 
        NULL,             // Use parent's environment block. 
        NULL,             // Use parent's starting directory. 
        &si,              // Pointer to STARTUPINFO structure.
        &pi )             // Pointer to PROCESS_INFORMATION structure.
    ) 
    {
		// delete temp files
		DeleteFile(compfile.c_str());	
		DeleteFile(asmfile.c_str());	
		DeleteFile(incfile.c_str());	
		SetCurrentDirectory(olddir);
		free(CryptTable);
        MessageBox( 0, "Compile failed", "Uuuups", MB_OK | MB_SETFOREGROUND);
        return;
    }
    // Wait until child process exits.
    WaitForSingleObject( pi.hProcess, INFINITE );
    // Close process and thread handles. 
    CloseHandle( pi.hProcess );
    CloseHandle( pi.hThread );
	// delete temp files
	DeleteFile(compfile.c_str());	
	DeleteFile(asmfile.c_str());	
	DeleteFile(incfile.c_str());
	// restore old directory
	SetCurrentDirectory(olddir);
	free(CryptTable);
#endif
	// write song info file
	std::string infofile;
	if (objformat != 0) // not windows obj
		infofile = fpath.substr(0, fpath.size()-1) + "h";
	else
		infofile = fpath.substr(0, fpath.size()-3) + "h";
	FILE *fnfofile = fopen(infofile.c_str(), "w");
	fprintf(fnfofile, "// some useful song defines for 4klang\n");
	fprintf(fnfofile, "#define SAMPLE_RATE %d\n", 44100);
	fprintf(fnfofile, "#define BPM %f\n", BeatsPerMinute);
	fprintf(fnfofile, "#define MAX_INSTRUMENTS %d\n", maxinst);
	fprintf(fnfofile, "#define MAX_PATTERNS %d\n", PatternIndices[0].size());
	fprintf(fnfofile, "#define PATTERN_SIZE_SHIFT %d\n", GetShift2(PatternSize));
	fprintf(fnfofile, "#define PATTERN_SIZE (1 << PATTERN_SIZE_SHIFT)\n");
	fprintf(fnfofile, "#define MAX_TICKS (MAX_PATTERNS*PATTERN_SIZE)\n");
	fprintf(fnfofile, "#define SAMPLES_PER_TICK %d\n", ((int)(44100.0f*4.0f*60.0f/(BeatsPerMinute*16.0f))));
	fprintf(fnfofile, "#define MAX_SAMPLES (SAMPLES_PER_TICK*MAX_TICKS)\n");
	fprintf(fnfofile, "#define POLYPHONY %d\n", SynthObj.Polyphony);
	if (output16)
	{
		fprintf(fnfofile, "#define INTEGER_16BIT\n");
		fprintf(fnfofile, "#define SAMPLE_TYPE short\n");
	}
	else
	{
		fprintf(fnfofile, "#define FLOAT_32BIT\n");
		fprintf(fnfofile, "#define SAMPLE_TYPE float\n");
	}
	if (objformat == 0)
	{
		fprintf(fnfofile, "\n#define WINDOWS_OBJECT\n\n");
		fprintf(fnfofile, "// declaration of the external synth render function, you'll always need that\n");
		fprintf(fnfofile, "extern \"C\" void  __stdcall	_4klang_render(void*);\n");
		fprintf(fnfofile, "// declaration of the external envelope buffer. access only if you're song was exported with that option\n");
		fprintf(fnfofile, "extern \"C\" float _4klang_envelope_buffer;\n");
		fprintf(fnfofile, "// declaration of the external note buffer. access only if you're song was exported with that option\n");
		fprintf(fnfofile, "extern \"C\" int   _4klang_note_buffer;\n");
	}
	if (objformat == 1)
	{
		fprintf(fnfofile, "\n#define LINUX_OBJECT\n\n");
		fprintf(fnfofile, "// declaration of the external synth render function, you'll always need that\n");
		fprintf(fnfofile, "extern void* __4klang_render(void*);\n");
		fprintf(fnfofile, "// declaration of the external envelope buffer. access only if you're song was exported with that option\n");
		fprintf(fnfofile, "extern float __4klang_envelope_buffer;\n");
		fprintf(fnfofile, "// declaration of the external note buffer. access only if you're song was exported with that option\n");
		fprintf(fnfofile, "extern int   __4klang_note_buffer;\n");
	}
	if (objformat == 2)
	{
		fprintf(fnfofile, "\n#define MACOSX_OBJECT\n\n");
		fprintf(fnfofile, "// declaration of the external synth render function, you'll always need that\n");
		fprintf(fnfofile, "extern void* _4klang_render(void*);\n");
		fprintf(fnfofile, "// declaration of the external envelope buffer. access only if you're song was exported with that option\n");
		fprintf(fnfofile, "extern float _4klang_envelope_buffer;\n");
		fprintf(fnfofile, "// declaration of the external note buffer. access only if you're song was exported with that option\n");
		fprintf(fnfofile, "extern int   _4klang_note_buffer;\n");
	}
	
	fclose(fnfofile);


    // save the patch file, just in case the VST host's music/project file gets lost
	std::string patchfile;
	if (objformat != 0) // not windows obj
		patchfile = fpath.substr(0, fpath.size()-1) + "4kp";
	else
		patchfile = fpath.substr(0, fpath.size()-3) + "4kp";

    Go4kVSTi_SavePatch(patchfile.c_str());
}

SynthObjectP Go4kVSTi_GetSynthObject()
{
	return &SynthObj;
}

GlobalSettingsP Go4kVSTi_GetGlobalSettings()
{
	return &GlobalSettings;
}
