#include <stdio.h>

#include "windows.h"



#define FPU_PROBLEM_INCREMENT 8



// init synth
void Go4kVSTi_Init();


// destroy synth
void Go4kVSTi_Destroy();


// clear instrument slot
void Go4kVSTi_ClearInstrumentSlot(int instrument_index, int slot);
// clear instrument workspace
void Go4kVSTi_ClearInstrumentWorkspace(int instrument_index);
// clear global slot
void Go4kVSTi_ClearGlobalSlot(int slot);
// clear global workspace
void Go4kVSTi_ClearGlobalWorkspace();

// reset instrument to default values
void Go4kVSTi_ResetInstrument(int instrument_index, bool reset_all_delay_buffers);
// reset global units (the "global instrument") to default values
void Go4kVSTi_ResetGlobalUnits(bool reset_all_delay_buffers);
void Go4kVSTi_ResetGlobalSettings();
// reset the synth to default values
void Go4kVSTi_ResetPatch();

// flip 2 neighbour instrument slots
void Go4kVSTi_FlipInstrumentSlots(int instrument_index, int a, int b);
// flip 2 neighbour global slots
void Go4kVSTi_FlipGlobalSlots(int a, int b);

// init a slot with given type with default values
void Go4kVSTi_InitSlot(BYTE* slot, int instrument_index, int type);
// init a instrument slot
void Go4kVSTi_InitInstrumentSlot(int instrument_index, int s, int type);
// init a global slot
void Go4kVSTi_InitGlobalSlot(int s, int type);

// set global bpm
void Go4kVSTi_SetBPM(float bpm);
float Go4kVSTi_GetBPM();

// activate solo mode
void Go4kVSTi_Solo(int channel, int solo);
// stream recording activation/deactivation
void Go4kVSTi_Record(bool record, int patternsize, float patternquant);
// panic
void Go4kVSTi_Panic();
// update dll times (e.g. sync to current bpm)
void Go4kVSTi_UpdateDelayAndSlewTimes();
// clear delay lines
void Go4kVSTi_ClearDelayLinesAndSlews();

// one tick the whole synth pipeline. results are left and right output sample
void Go4kVSTi_Tick(float *oleft, float *oright, int samples);
// add a voice to the first instrument (in ascending order) that matches given (channel, note, velocity) combination
void Go4kVSTi_AddVoice(int channel, int note, int velocity);
// stop the first instrument (in ascending order) that matches the given channel, and has the given note being held down
void Go4kVSTi_StopVoice(int channel, int note);

// stop all voices with given MIDI channel, regardless of their velocity and note ranges
void Go4kVSTi_StopAllVoicesOnChannel(int channel);

void Go4kVSTi_StopVoicesFromAllInstruments();

// load binary patch data
void Go4kVSTi_LoadPatch(char *filename);
// save binary patch data
void Go4kVSTi_SavePatch(const char *filename);
// load instrumen data to specified channel
void Go4kVSTi_LoadInstrument(char* filename, int instrument_index);
// save instrument data from current channel
void Go4kVSTi_SaveInstrument(char* filename, int instrument_index);
// load unit data into specified slot
void Go4kVSTi_LoadUnit(char* filename, BYTE* slot);
// save unit date from specified slot
void Go4kVSTi_SaveUnit(char* filename, BYTE* slot);

#define EXPORT_OBJECT_FILE_
void Go4kVSTi_SaveByteStream(HINSTANCE hInst, char* filename, int useenvlevels, int useenotevalues, int clipoutput, int undenormalize, int objformat, int output16, int swing_strength);

#define MAX_POLYPHONY		2
#define MAX_INSTRUMENTS		48

#define MAX_UNITS			32

// #define MAX_SLOTS			32


// In the VST plugin, we have max 16 values per unit (old 4klang versions had 32),
// even though in the 4k export asm/inc we only have 8.
#define MAX_SLOT_VALUES		16

#define GLOBAL_INSTRUMENT_IDX (MAX_INSTRUMENTS)

#define DLL_SYNCTYPE_ABSOLUTE 0
#define DLL_SYNCTYPE_BPM 1
#define DLL_SYNCTYPE_NOTE_PITCH 2


#define SLW_SYNCTYPE_ABSOLUTE_MS 0
#define SLW_SYNCTYPE_ABSOLUTE_SMP 1
#define SLW_SYNCTYPE_BPM 2
#define SLW_SYNCTYPE_NOTE_PITCH 3


enum UnitID
{
	M_NONE = 0,
	M_ENV,
	M_VCO,
	M_VCF,
	M_DST,
	M_DLL,
	M_FOP,
	M_FSTx, // this is a general store, which results in either FST or FSTG in the 4k export
	M_PAN,
	M_OUT,
	M_ACC,
	M_FLD,
    M_SLW,
    M_FLDFROM,
    // M_ARP,
	NUM_MODULES
};




/////////////////////////////////////////////////////////////////////////////////////
// value definitions
/////////////////////////////////////////////////////////////////////////////////////


#define ENV_STATE_ATTAC		0
#define ENV_STATE_DECAY		1
#define ENV_STATE_SUSTAIN	2
#define ENV_STATE_RELEASE	3
#define ENV_STATE_OFF		4

typedef struct ENV_val
{
	BYTE	id;
	BYTE	attac;
	BYTE	decay;
	BYTE	sustain;
	BYTE	release;
	BYTE	gain;
// GUI STUFF
} *ENV_valP;

#define VCO_SINE		0x01
#define VCO_TRISAW		0x02
#define VCO_PULSE		0x04
#define VCO_NOISE		0x08
#define VCO_LFO			0x10
#define VCO_GATE		0x20
typedef struct VCO_val
{
	BYTE	id;
	BYTE	transpose;
	BYTE	detune;
	BYTE	phaseofs;
	BYTE	gate;
	BYTE	color;
	BYTE	shape;
	BYTE	gain;
	BYTE	flags;
// GUI STUFF
} *VCO_valP;

typedef struct VCO11_val
{
	BYTE	id;
	BYTE	transpose;
	BYTE	detune;
	BYTE	phaseofs;
	BYTE	color;
	BYTE	shape;
	BYTE	gain;
	BYTE	flags;
// GUI STUFF
} *VCO11_valP;

#define VCF_LOWPASS		0x1
#define VCF_HIGHPASS	0x2
#define VCF_BANDPASS	0x4
#define	VCF_BANDSTOP	0x3
#define VCF_ALLPASS		0x7
#define	VCF_PEAK		0x8
typedef struct VCF_val
{
	BYTE	id;
	BYTE	freq;
	BYTE	res;
	BYTE	type;
// GUI STUFF
} *VCF_valP;

// Generic "base class"
typedef struct SOME_UNIT_val
{
    BYTE    id;
    BYTE    data[MAX_SLOT_VALUES-1];
} *SOME_UNIT_valP;

typedef struct DST_val
{
	BYTE	id;
	BYTE	drive;
	BYTE	snhfreq;
// GUI STUFF
} *DST_valP;

typedef struct DLL_val
{
	BYTE	id;
	BYTE	pregain;
	BYTE	dry;
	BYTE	feedback;
	BYTE	damp;
	BYTE	freq;
	BYTE	depth;
	BYTE	delayidx;
	BYTE	count;
// GUI STUFF
	BYTE	guidelay;
	BYTE	synctype;
	BYTE	leftreverb;
	BYTE	reverb;
} *DLL_valP;

typedef struct DLL10_val
{
	BYTE	id;
	BYTE	pregain;
	BYTE	dry;
	BYTE	feedback;
	BYTE	damp;
	BYTE	delay;
	BYTE	count;
// GUI STUFF
	BYTE	guidelay;
	BYTE	synctype;
	BYTE	leftreverb;
	BYTE	reverb;
} *DLL10_valP;

#define FOP_POP			0x1
#define FOP_ADDP		0x2
#define FOP_MULP		0x3
#define FOP_PUSH		0x4
#define FOP_XCH			0x5
#define FOP_ADD			0x6
#define FOP_MUL			0x7
#define FOP_ADDP2		0x8
#define FOP_LOADNOTE	0x9
typedef struct FOP_val
{
	BYTE	id;
	BYTE	flags;
} *FOP_valP;

// *********************************************************************************
// NOTE:    The FSTx_val struct's layout in this VSTi plugin is NOT THE SAME as  
//          either FST or FSTG's byte data values in the 4k-exported version.
//          The "FSTx" unit can result in either an FST or FSTG unit in the 4k export,
//          depending on whether the store was local (inside the same instrument)
//          or across different instruments.
//          In Go4kVSTi_Tick() there's completely separate handling code for M_FSTx,
//          and the functions in 4klang.asm are not called.
// *********************************************************************************
// 
// In the exported asm file, the FST flags are stored in the target workspace slot reference number like this:
// %define	FST_SET			0x0000
// %define	FST_ADD			0x4000
// %define	FST_POP			0x8000
//
// so the maximum unit slot index is 0x3fff, meaning
//   max 64 instruments * 32 units * 8 workspace slots
// ..
// or 16 instruments * 64 units * 16 workspace slots
// 
// and we cannot suppot "FST_MUL" because that would have to be something like 0x2000
// which would cut something in half. OR the %2 parameter of the GO4K_FST macro has to grow to "dd".
// or there has to be a %3 parameter.

#define	FST_OPERATION_SET			0x00
#define	FST_OPERATION_ADD			0x10
#define	FST_OPERATION_MUL			0x20
#define	FST_OPERATION_POP			0x40

#define	FST_CONDITION_ALWAYS			0x00
#define	FST_CONDITION_IF_SMALLER		0x01
#define	FST_CONDITION_IF_LARGER			0x02
#define	FST_CONDITION_IF_SIGN_CHANGED   0x04

// This whole unit type and struct is GUI STUFF.
typedef struct FSTx_val
{
	BYTE	id;
	BYTE	amount;
	BYTE	operation; // todo: vaihda kentn nimeksi "operation"
	char	dest_stack; // mink takia n on "char" eik "BYTE"?????
	char	dest_unit; // hoi laari laa voisko tss olla 255 eli -1 ja sill vois storeta vaikka voicen pitchiin???
	char	dest_slot;
	char	dest_id;
	BYTE	condition;
} *FSTx_valP;

typedef struct FSTx_wrk
{
	float	previn;
} *FSTx_wrkP;


typedef struct PAN_val
{
	BYTE	id;
	BYTE	panning;
// GUI STUFF
} *PAN_valP;

typedef struct OUT_val
{
	BYTE	id;
	BYTE	gain;
	BYTE	auxsend;
// GUI STUFF
} *OUT_valP;

#define ACC_OUT				0
#define	ACC_AUX				8
typedef struct ACC_val
{
	BYTE	id;
	BYTE	flags;
} *ACC_valP;

typedef struct FLD_val
{
	BYTE	id;
	BYTE	value;
// GUI STUFF
} *FLD_valP;

/* Slew limiter, glide, smooth, slide, integrator, envelope follower, ... */
#define SLW_CALCMODE_LINEAR 0
#define SLW_CALCMODE_RMS 1
typedef struct SLW_val
{
	BYTE	id;
	BYTE	delayidx; // index in delay table
    BYTE	calcmode; // 0=linear, 1=RMS, 2=...?
// GUI STUFF
	BYTE	guidelay;
	BYTE	synctype; // absolute milliseconds, absolute samples, musical BPM, or musical note pitch
    WORD    delaytimesmp; // delay time in samples. this copy of the value is only used for showing it in the gui
} *SLW_valP;


/* Load value from somewhere */
// This whole unit type and struct is GUI STUFF.
typedef struct FLDFROM_val
{
	BYTE	id;
	BYTE	amount;
	char	source_stack; // note: loading from other stacks i.e. instruments might be problematic if polyphony = 1, because there are multiple voices
	char	source_unit;
	char	source_slot;
	char	source_id;
} *FLDFROM_valP;




/////////////////////////////////////////////////////////////////////////////////////
// workspace definitions
/////////////////////////////////////////////////////////////////////////////////////

typedef struct InstrumentWorkspace
{
	DWORD	release;
	DWORD	note;
    // in 4klang 3.01 the workspace size used to be 256, without explaining where it came from, which was 32*8
    // in 4klang 3.11 the workspace size is [MAX_UNITS*MAX_UNIT_SLOTS], 
    // and "MAX_UNIT_SLOTS" used to be "MAX_WORKSPACE_SLOTS" in 4klang 3.01
    float	unit_workspace[MAX_UNITS*MAX_SLOT_VALUES];

	float	dlloutl;
	float	dlloutr;
	float	outl;
	float	outr;
} *InstrumentWorkspaceP;

typedef struct InstrumentInputSettingsS {
    // Instrument settings, i.e. MIDI channel, velocity and note limits are not stored in the single instrument file
    DWORD MIDIChannel; 
    DWORD VelocityLo;
    DWORD VelocityHi;
    DWORD NoteLo;
    DWORD NoteHi;
    DWORD MapVelocityToNote;
    BYTE DummyReservedSpace[40];
} *InstrumentInputSettingsP;

typedef struct SynthObject
{
	DWORD Polyphony;
	char InstrumentNames[MAX_INSTRUMENTS][64];
	BYTE InstrumentValues[MAX_INSTRUMENTS][MAX_UNITS][MAX_SLOT_VALUES];
	BYTE GlobalValues[MAX_UNITS][MAX_SLOT_VALUES];								
	InstrumentWorkspace InstrumentWork[MAX_INSTRUMENTS*MAX_POLYPHONY];
	InstrumentWorkspace GlobalWork;
	DWORD InstrumentSignalValid[MAX_INSTRUMENTS];
	DWORD GlobalSignalValid;
	float SignalTrace[MAX_INSTRUMENTS];
	int ControlInstrument[MAX_INSTRUMENTS];
	int VoiceIndex[MAX_INSTRUMENTS]; // polyphony "toggle": what voice is active (basically a round-robin thing? but max polyphony 2, so 0..1)
    InstrumentInputSettingsS InstrumentInputSettings[MAX_INSTRUMENTS];
    int HeldNote[MAX_INSTRUMENTS]; // What MIDI note is being held down. -1 = no note is held. This is not the same thing as InstrumentWorkspace.note.

    int InstrumentFPUProblemsCount[MAX_INSTRUMENTS+1]; // UpdateVoiceDisplay sets these to the sum of each instrument's units' FPU problem counts
    int UnitFPUProblemsCount[(MAX_INSTRUMENTS+1)*MAX_UNITS]; // just one counter for both voices
    int AutoPanicNotificationCounter; // auto-panic writes this and the GUI reads it to show the user when auto-panic has worked
    int PanicCount; // each panic incident increases this, so you can leave the song running and see if there has been problems.

    // int audio_renderings_active; // for diagnosing how the VST host calls our stuff
    HANDLE synth_mutex;
} *SynthObjectP;

SynthObjectP Go4kVSTi_GetSynthObject();

// auto-panic mask bits, correspond to x87 FPU 
#define AUTO_PANIC_ON_INVALID_OPERATION 1
#define AUTO_PANIC_ON_DENORMAL 2
#define AUTO_PANIC_ON_DIVIDE_BY_ZERO 4
#define AUTO_PANIC_ON_OVERFLOW 8
#define AUTO_PANIC_ON_UNDERFLOW 16
#define AUTO_PANIC_ON_PRECISION_LOSS 32
#define AUTO_PANIC_ON_STACK_FAULT 64
#define AUTO_PANIC_ON_EXCEPTION_FLAG 128



typedef struct GlobalSettingsS
{   // 4k export settings go here
    DWORD SwingStrength; // 50-75
    DWORD PatternSize; // pattern size as notes, .g. 8, 16, 32, 64
    DWORD NoteQuantization; // 0:1/4, 1:1/8, 2:1/16, 3:1/32, 4:1/64, 5:1/128
    DWORD ClipOutput; // 0:no, 1:yes
    DWORD Undenormalize; // 0:no, 1:yes
    DWORD RecordEnvs; // 0:no, 1:yes
    DWORD RecordNotes; // 0:no, 1:yes
    DWORD Output16Bit; // 0:no, 1:yes
    DWORD ObjectFormat; // 0:Windows OBJ, 1:Linux ELF, 2:OSX MACHO
    DWORD AlwaysOnTop; // 0:no, 1:yes

    DWORD AutoPanicMask; // 0:off    
    DWORD AutoPanicDoPanic; // 0:do nothing on fpu problems, default:1

    // the Polyphony setting is stored in the SynthObject
} *GlobalSettingsP;

GlobalSettingsP Go4kVSTi_GetGlobalSettings();


// -- "chunk i/o" is used for reading to/from memory instead of fwrite(), etc. --
#define MAX_CHUNK_SIZE 100000

typedef struct Chunk {
    size_t size; // size and writepos
    size_t readpos;
    BYTE data[MAX_CHUNK_SIZE];
} *ChunkP;

extern void chunk_init(ChunkP c);
// chunk_write() is like fwrite()
extern size_t chunk_write(const void *str, size_t element_size, size_t element_count, ChunkP chunk);
// chunk_read() is like fread()
extern size_t chunk_read(void *str, size_t element_size, size_t element_count, ChunkP chunk);
// chunk_write_FILE() writes a FILE into the memory chunk (i.e. it loads the file to the memory chunk)
extern size_t chunk_write_FILE(FILE *stream, ChunkP chunk);

// ---------------------------------------------------------------------------

extern bool Go4kVSTi_LoadPatchFromDataChunk(ChunkP chunk);
extern int Go4kVSTi_SavePatchToDataChunk(ChunkP chunk);



//extern bool IsModulationLinkUnit(SOME_UNIT_valP v);
#define IsModulationLinkUnit(x) (((x)->id == M_FSTx) || ((x)->id == M_FLDFROM))


extern int GetModulationLinkInstrument(SOME_UNIT_valP v);
extern char GetModulationLinkUnit(SOME_UNIT_valP v);
extern char GetModulationLinkSlot(SOME_UNIT_valP v);
extern char GetModulationLinkID(SOME_UNIT_valP v);

extern void SetModulationLinkUnit(SOME_UNIT_valP v, char value);
extern void SetModulationLinkSlot(SOME_UNIT_valP v, char value);
extern void SetModulationLinkID(SOME_UNIT_valP v, char value);

extern bool GetSynthMutex();
extern bool ReleaseSynthMutex();

