// 4klang was quite modest in its Windows requirements:
// #define _WIN32_WINNT	0x0400
// SetDBBrushColor needs at least Windows 2000 Professional.
#define _WIN32_WINNT 0x0500

#include <windows.h>
#include <winuser.h>
#include <commctrl.h>
#include "..\..\win\resource.h"
#include "Go4kVSTiGUI.h"
#include "Go4kVSTiCore.h"
#include <stdio.h>
#include <vector>
#include <map>
#include <string>
#include <math.h>

#define T_INSTRUMENT	0
#define T_GLOBAL		1
#define NUM_TABS		2

static HINSTANCE hInstance;
static UINT_PTR timer = 1;
static UINT_PTR backupTimer = 2;
static UINT_PTR timerID = 0;
static UINT_PTR backupTimerID = 0;
static HWND DialogWnd = NULL;
static HWND ModuleWnd[NUM_MODULES];
static HWND TabWnd[NUM_TABS];
static HWND SetWnd;
static HWND ScrollWnd;
static int SelectedIUnit = 0;
static int SelectedGUnit = 0;
static int SelectedTab = T_INSTRUMENT;
static int SelectedInstrument = 0;
static int SetUnit = -1;
static char SliderValTxt[128];
static DWORD UnitCopyBuffer[MAX_SLOT_VALUES];
static int InstrumentScrollPos[MAX_INSTRUMENTS];
static int GlobalScrollPos;
static int InitDone = false;

static SynthObjectP SynthObjP = NULL;
static GlobalSettingsP GlobalSetsP = NULL;

static char* UnitName[NUM_MODULES] =
{
	"-",
	"Envelope",
	"Oscillator",
	"Filter",
	"Distortion",
	"DelayLine",
	"Arithmetic",
	"Store",
	"Panning",
	"Output",
	"Accumulate",
	"Load",
    "Slew",
    "Load From",
};

// minimal signal precondition for each unit
static int UnitPreSignals[NUM_MODULES] = 
{
	0,	// none
	0,	// env
	0,	// vco
	1,	// vcf
	1,	// dst
	1,	// dll
	2,  // fop, signal count for arithmetic depends on mode
	1,	// fst
	1,	// pan
	2,	// out
	0,	// acc
	0,  // fld
    1,	// slw
    0,  // fldfrom
};

// signal post condition (relative to precondition)
static int UnitPostSignals[NUM_MODULES] = 
{
	0,	// none
	1,	// env
	1,	// vco
	0,	// vcf
	0,	// dst
	0,	// dll
	0,  // fop, signal count for arithmetic depends on mode
	0,	// fst
	1,	// pan
	-2,	// out
	2,	// acc
	1,	// fld
    0,  // slw
    1,  // fldfrom
};

// Modulation targets map to _wrk structures specified in 4klang.inc, for example 
// struc	go4kVCO_wrk
//
// The _wrk strucs have two kinds of fields: actual work variables, and modulation targets.
// The actual work variables, e.g. go4kVCO_wrk.phase, are "master" copies of the unit's
// state, and were not meant to be directly modulated in 4klang. We reveal them as 
// modulation targets and sources, so we can do e.g. oscillator sync.
//
// Actual work variables are in (Parentheses).
//
// The DLL and SLW unit work strucs, e.g. go4kDLL_wrk, are special cases, because 
// there the actual state variables are not stored on the normal work data, 
// but in a separate delay work buffer. For modulation targets, there's a separate
// "_wrk2" structure.
// 
static char* UnitModulationTargetNames[NUM_MODULES][8] = 
{
	{ "", "", "", "", "", "", "", "" },
	{ "(State)", "(Level)", "Gain", "Attack", "Decay", "", "Release", "" }, // ENV
	{ "(Phase)", "Transpose", "Detune", "Frequency", "Phase", "Color", "Shape", "Gain"}, // VCO
	{ "(Low)", "(High)", "(Band)", "(Freq)", "Cutoff Frequency", "Resonance", "", "" }, // VCF
	{ "(Out)", "(S&H Phase)", "Drive", "Sample&Hold Frequency", "", "", "", ""}, // DST
	{ "Pregain", "Feedback", "Dry", "Damp", "LFO Frequency", "LFO Depth", "", "" }, // DLL_wrk2
	{ "", "", "", "", "", "", "", "" }, // FOP
	{ "", "", "", "", "", "", "", "" }, // FST
	{ "Panning", "", "", "", "", "", "", "" }, // PAN
	{ "AUX", "Gain", "", "", "", "", "", "" }, // OUT
	{ "", "", "", "", "", "", "", "" }, // ACC
	{ "Value", "", "", "", "", "", "", "" }, // FLD
	{ "", "", "", "", "", "", "", "" }, // SLW
	{ "", "", "", "", "", "", "", "" }, // FLDFROM
};

static char* UnitModulationTargetShortNames[NUM_MODULES][8] = 
{
	{ "", "", "", "", "", "", "", "" },
	{ "(State)", "(Level)", "Gain", "Attack", "Decay", "", "Release", "" }, // ENV
	{ "(Phase)", "Transp", "Detune", "Freq", "Phase", "Color", "Shape", "Gain"}, // VCO
	{ "(Low)", "(High)", "(Band)", "(Freq)", "Cutoff", "Res", "", "" }, // VCF
	{ "(Out)", "(SnHPh)", "Drive", "Freq", "", "", "", ""}, // DST
	{ "Pregain", "Feedb", "Dry", "Damp", "LFO F", "LFO D", "", "" }, // DLL_wrk2
	{ "", "", "", "", "", "", "", "" }, // FOP
	{ "", "", "", "", "", "", "", "" }, // FST
	{ "Pan", "", "", "", "", "", "", "" }, // PAN
	{ "AUX", "Gain", "", "", "", "", "", "" }, // OUT
	{ "", "", "", "", "", "", "", "" }, // ACC
	{ "Value", "", "", "", "", "", "", "" }, // FLD
	{ "", "", "", "", "", "", "", "" }, // SLW
	{ "", "", "", "", "", "", "", "" }, // FLDFROM
};

static char* delayName[33] = 
{	
	"1/32T",
	"1/32",
	"1/32D",
	"1/16T",
	"1/16",
	"1/16D",
	"1/8T",
	"1/8",
	"1/8D",
	"1/4T",
	"1/4",
	"1/4D",
	"1/2T",
	"1/2",
	"1/2D",
	"1T",
	"1",
	"1D",
	"2T",
	"2",
	"2D",
	"3/8",
	"5/8",
	"7/8",
	"9/8",
	"11/8",
	"13/8",
	"15/8",
	"3/4",
	"5/4",
	"7/4",
	"3/2",
	"3/2",
};

static char* NoteNames[128] = 
{	
    "C-0", "C#0", "D-0", "D#0", "E-0", "F-0", "F#0", "G-0", "G#0", "A-0", "A#0", "B-0", 
    "C-1", "C#1", "D-1", "D#1", "E-1", "F-1", "F#1", "G-1", "G#1", "A-1", "A#1", "B-1", 
    "C-2", "C#2", "D-2", "D#2", "E-2", "F-2", "F#2", "G-2", "G#2", "A-2", "A#2", "B-2", 
    "C-3", "C#3", "D-3", "D#3", "E-3", "F-3", "F#3", "G-3", "G#3", "A-3", "A#3", "B-3", 
    "C-4", "C#4", "D-4", "D#4", "E-4", "F-4", "F#4", "G-4", "G#4", "A-4", "A#4", "B-4", 
    "C-5", "C#5", "D-5", "D#5", "E-5", "F-5", "F#5", "G-5", "G#5", "A-5", "A#5", "B-5", 
    "C-6", "C#6", "D-6", "D#6", "E-6", "F-6", "F#6", "G-6", "G#6", "A-6", "A#6", "B-6", 
    "C-7", "C#7", "D-7", "D#7", "E-7", "F-7", "F#7", "G-7", "G#7", "A-7", "A#7", "B-7", 
    "C-8", "C#8", "D-8", "D#8", "E-8", "F-8", "F#8", "G-8", "G#8", "A-8", "A#8", "B-8", 
    "C-9", "C#9", "D-9", "D#9", "E-9", "F-9", "F#9", "G-9", "G#9", "A-9", "A#9", "B-9", 
    "C-10", "C#10", "D-10", "D#10", "E-10", "F-10", "F#10", "G-10"
};


char* GetUnitString(BYTE* unit, char* unitname)
{
	char UnitDesc[128];
	UnitDesc[0]=0;
	
	if (unit[0] == M_VCO)
	{
		VCO_valP val = (VCO_valP)unit;
		char lfo[5];
		lfo[0] = 0;
		if (val->flags & VCO_LFO)
			sprintf(lfo, "-LFO");
		if (val->flags & VCO_SINE)
			sprintf(UnitDesc, "    (%s%s)","Sine",lfo);
		if (val->flags & VCO_TRISAW)
			sprintf(UnitDesc, "    (%s%s)","TriSaw",lfo);
		if (val->flags & VCO_PULSE)
			sprintf(UnitDesc, "    (%s%s)","Pulse",lfo);
		if (val->flags & VCO_NOISE)
			sprintf(UnitDesc, "    (%s%s)","Noise",lfo);
		if (val->flags & VCO_GATE)
			sprintf(UnitDesc, "    (%s%s)","Gate",lfo);
	}
	if (unit[0] == M_VCF)
	{
		VCF_valP val = (VCF_valP)unit;
		if (val->type == VCF_LOWPASS)
			sprintf(UnitDesc, "    (%s F:%d)","Low", val->freq);
		if (val->type == VCF_HIGHPASS)
			sprintf(UnitDesc, "    (%s F:%d)","High", val->freq);
		if (val->type == VCF_BANDPASS)
			sprintf(UnitDesc, "    (%s F:%d)","Band", val->freq);
		if (val->type == VCF_BANDSTOP)
			sprintf(UnitDesc, "    (%s F:%d)","Notch", val->freq);
		if (val->type == VCF_PEAK)
			sprintf(UnitDesc, "    (%s F:%d)","Peak", val->freq);
		if (val->type == VCF_ALLPASS)
			sprintf(UnitDesc, "    (%s F:%d)","All", val->freq);		
	}
	if (unit[0] == M_FOP)
	{
		FOP_valP val = (FOP_valP)unit;
		if (val->flags == FOP_ADD)
			sprintf(UnitDesc, "    (%s)","+");
		if (val->flags == FOP_ADDP)
			sprintf(UnitDesc, "    (%s)","+/Pop");
		if (val->flags == FOP_MUL)
			sprintf(UnitDesc, "    (%s)","*");
		if (val->flags == FOP_MULP)
			sprintf(UnitDesc, "    (%s)","*/Pop");
		if (val->flags == FOP_POP)
			sprintf(UnitDesc, "    (%s)","Pop");
		if (val->flags == FOP_PUSH)
			sprintf(UnitDesc, "    (%s)","Push");
		if (val->flags == FOP_XCH)
			sprintf(UnitDesc, "    (%s)","Xch");
		if (val->flags == FOP_ADDP2)
			sprintf(UnitDesc, "    (%s)","2+/Pop");
		if (val->flags == FOP_LOADNOTE)
			sprintf(UnitDesc, "    (%s)","LoadNote");
	}
	if (unit[0] == M_DLL)
	{
		DLL_valP val = (DLL_valP)unit;
		if (val->reverb == 0)
			sprintf(UnitDesc, "    (%s)","Delay");
		else
			sprintf(UnitDesc, "    (%s)","Reverb");
	}
	if (unit[0] == M_FSTx)
	{
		FSTx_valP val = (FSTx_valP)unit;
		if (val->dest_unit == -1 || val->dest_slot == -1 || val->dest_id == -1)
		{
			sprintf(UnitDesc, "    (No Target)");
		}
		else
		{
			std::string st;
			if (val->operation & FST_OPERATION_ADD)
				st = "+";
			if (val->operation & FST_OPERATION_MUL)
				st = "*";
			if (val->operation & FST_OPERATION_POP)
				st += ",Pop";

			if ((val->dest_stack != -1) && (val->dest_stack != SelectedInstrument))
			{
				if (val->dest_stack == MAX_INSTRUMENTS)
					sprintf(UnitDesc, "    (GU%d %s)%s", val->dest_unit+1, UnitModulationTargetShortNames[val->dest_id][val->dest_slot], st.c_str());
				else
					sprintf(UnitDesc, "    (%dU%d %s)%s",val->dest_stack+1, val->dest_unit+1, UnitModulationTargetShortNames[val->dest_id][val->dest_slot], st.c_str());
			}
			else
			{
				sprintf(UnitDesc, "    (U%d %s)%s", val->dest_unit+1, UnitModulationTargetShortNames[val->dest_id][val->dest_slot], st.c_str());
			}

		}
	} // M_FST
	if (unit[0] == M_DST)
	{
		DST_valP val = (DST_valP)unit;
		sprintf(UnitDesc, "    (D:%d)", val->drive-64, val->snhfreq);
	}
	if (unit[0] == M_PAN)
	{
		PAN_valP val = (PAN_valP)unit;
		sprintf(UnitDesc, "    (%d)", val->panning-64);
	}
	if (unit[0] == M_ACC)
	{
		ACC_valP val = (ACC_valP)unit;
		if (val->flags == ACC_OUT)
			sprintf(UnitDesc, "    (%s)","Out");
		else
			sprintf(UnitDesc, "    (%s)","Aux");
	}
	if (unit[0] == M_OUT)
	{
		OUT_valP val = (OUT_valP)unit;
		sprintf(UnitDesc, "    (G:%d A:%d)",val->gain, val->auxsend);
	}
	if (unit[0] == M_FLD)
	{
		FLD_valP val = (FLD_valP)unit;
		sprintf(UnitDesc, "    (%d)", val->value-64);
	}
	if (unit[0] == M_SLW)
	{
		SLW_valP val = (SLW_valP)unit;
		sprintf(UnitDesc, "    (%s)", "Morjes");
	}

	if (unit[0] == M_FLDFROM)
	{
		FLDFROM_valP val = (FLDFROM_valP)unit;
		if (val->source_unit == -1 || val->source_slot == -1 || val->source_id == -1)
		{
			sprintf(UnitDesc, "    (No Source)");
		}
		else
		{
			std::string st;
			// st = "";

			if ((val->source_stack != -1) && (val->source_stack != SelectedInstrument))
			{
				if (val->source_stack == MAX_INSTRUMENTS)
					sprintf(UnitDesc, "    (GU%d %s)%s", val->source_unit+1, UnitModulationTargetShortNames[val->source_id][val->source_slot], st.c_str());
				else
					sprintf(UnitDesc, "    (%dU%d %s)%s",val->source_stack+1, val->source_unit+1, UnitModulationTargetShortNames[val->source_id][val->source_slot], st.c_str());
			}
			else
			{
				sprintf(UnitDesc, "    (U%d %s)%s", val->source_unit+1, UnitModulationTargetShortNames[val->source_id][val->source_slot], st.c_str());
			}

		}
	} // M_FLDFROM


	sprintf(unitname, "%s%s", UnitName[unit[0]], UnitDesc);
	return unitname;
} // GetUnitString()

void UpdateDelayTimes(DLL_valP unit)
{	
	// if reverb skip
	if (unit->reverb)
		return;

	Go4kVSTi_UpdateDelayAndSlewTimes();

	int delay;
	char text[10];
	DLL_valP v = (DLL_valP)unit;
	if (v->synctype == 2)
	{
	}
	else
	{
		if (v->synctype == 1)
		{
			sprintf(text, "%s", delayName[v->guidelay>>2]);
		}
		else
		{
			delay = v->guidelay*16;
			sprintf(text, "%.2f ms", (float)delay/44.100f);
		}
		sprintf(SliderValTxt, "%s", text);
		SetWindowText(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), SliderValTxt);
	}
} // UpdateDelayTimes

void UpdateSlewTimes(SLW_valP unit) // this is called from the GUI widget callbacks
{	
	Go4kVSTi_UpdateDelayAndSlewTimes();

	int delay;
	char text[10];
	SLW_valP v = (SLW_valP)unit;
	//if (v->synctype == 2)
	//{
        // mik tn synctypen asettaa hh
	//}
	//else
	{

// #define SYNCTYPE_ABSOLUTE_MS 0
// #define SYNCTYPE_ABSOLUTE_SMP 1
// #define SYNCTYPE_BPM 2
// #define SYNCTYPE_NOTE_PITCH 3

		if (v->synctype == 2)
		{ // 
			sprintf(text, "%s", delayName[v->guidelay>>2]);
		}
		else
		{
			delay = v->guidelay*16;
			sprintf(text, "%.2f ms", (float)delay/44.100f);
		}        

		sprintf(SliderValTxt, "%s", text);
		SetWindowText(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_VAL), SliderValTxt);

        sprintf(SliderValTxt, "%d", v->delaytimesmp);
        SetWindowText(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_SMP), SliderValTxt);
	}
} // UpdateDelayTimes

static char InstrumentComboboxItemTexts[MAX_INSTRUMENTS][200];

static const char* MIDIChannelComboboxTexts[16] = {"1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13", "14", "15", "16"};


void UpdateInstrumentSelectionComboboxItemTexts()
{
	for (int i = 0; i < MAX_INSTRUMENTS; i++)
    {
        //sprintf(InstrumentComboboxItemTexts[i], "%d: %s (ch %d)", i + 1, (LPSTR)&(SynthObjP->InstrumentNames[i]), SynthObjP);
        sprintf(InstrumentComboboxItemTexts[i], "%d: %s (ch:%d, v:%d..%d, n:%s..%s)", i + 1, 
              SynthObjP->InstrumentNames[i], SynthObjP->InstrumentInputSettings[i].MIDIChannel + 1,
              SynthObjP->InstrumentInputSettings[i].VelocityLo,
              SynthObjP->InstrumentInputSettings[i].VelocityHi,
              NoteNames[SynthObjP->InstrumentInputSettings[i].NoteLo],
              NoteNames[SynthObjP->InstrumentInputSettings[i].NoteHi]
        );
    }
}

void UpdateInstrumentSelectionCombobox(HWND hwndDlg)
{
    if (!SynthObjP)
        return;

        UpdateInstrumentSelectionComboboxItemTexts();
        
        SendDlgItemMessage(hwndDlg, IDC_INSTRUMENT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)(0));

		for (int i = 0; i < MAX_INSTRUMENTS; i++)
        {
            SendDlgItemMessage(hwndDlg, IDC_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)InstrumentComboboxItemTexts[i]);
        }
        
        SendDlgItemMessage(hwndDlg, IDC_INSTRUMENT_MIDI_CHANNEL, CB_SETCURSEL, (WPARAM)(SynthObjP->InstrumentInputSettings[SelectedInstrument].MIDIChannel), (LPARAM)0);

        SendDlgItemMessage(hwndDlg, IDC_INSTRUMENT, CB_SETCURSEL, (WPARAM)SelectedInstrument, (LPARAM)0);
}


int GetNumberFromControl(HWND hwndDlg, int nIDDlgItem, int minVal, int maxVal)
{
    static char temp[4];
	GetDlgItemText(hwndDlg, nIDDlgItem, temp, 4);
    temp[3] = 0;
    int val = atoi(temp);
    if (val < minVal) val = 0;
    if (val > maxVal) val = 127;
                        
    return val;
}

int ConvertPatternSizeFromGUI(int guival)
{
    if (guival == 0) return 8;
    else if (guival == 1) return 16;
    else if (guival == 2) return 32;
    else if (guival == 3) return 64;
    else
        return 16;
}

int ConvertPatternSizeToGUI(int patternsize)
{
    if (patternsize == 8) return 0;
    else if (patternsize == 16) return 1;
    else if (patternsize == 32) return 2;
    else if (patternsize == 64) return 3;
    else
        return 1;
}


BOOL SwitchSelectedInstrument(HWND hwndDlg, int NewSelectedInstrument)
{
    SelectedInstrument = NewSelectedInstrument;

	if (SendDlgItemMessage(hwndDlg, IDC_SOLO, BM_GETCHECK, 0, 0)==BST_CHECKED) 
		Go4kVSTi_Solo(SelectedInstrument, true);
	else
		Go4kVSTi_Solo(SelectedInstrument, false);
	// get values from synth and set gui accordingly ...
	UpdateControls(SelectedInstrument);
	// reset scroll pos
	int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
	SCROLLINFO sinfo;
	sinfo.cbSize = sizeof(SCROLLINFO);
	sinfo.fMask = SIF_POS;
	sinfo.nPos = InstrumentScrollPos[SelectedInstrument];
	SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);
	ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);

    return TRUE;
}

// CB for the main DLG
INT_PTR CALLBACK MainDialogProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{ 
	switch (message) 
    { 
		case WM_INITDIALOG:
		{			
			HWND tab = GetDlgItem(hwndDlg, IDC_MAINTAB);
			TCITEM item;
			item.mask = TCIF_TEXT;
			item.pszText = "Instrument";
			SendDlgItemMessage(hwndDlg, IDC_MAINTAB, TCM_INSERTITEM, (WPARAM)0, (LPARAM)(&item));
			item.pszText = "Global";
			SendDlgItemMessage(hwndDlg, IDC_MAINTAB, TCM_INSERTITEM, (WPARAM)2, (LPARAM)(&item));
            
            // the synth object doesn't exist yet, so we can't call this
            // UpdateInstrumentSelectionCombobox(hwndDlg);

            // MIDI Channel combobox
            for (int i=0; i<16; i++)
                SendDlgItemMessage(hwndDlg, IDC_INSTRUMENT_MIDI_CHANNEL, CB_ADDSTRING, (WPARAM)0, (LPARAM)(MIDIChannelComboboxTexts[i]));    		

			const char* swingstrength[] = {"50", "51", "52", "53", "54", "55", "56", "57", "58", "59", 
                                           "60", "61", "62", "63", "64", "65", "66", "67", "68", "69", 
                                           "70", "71", "72", "73", "74", "75"};
			for (int i = 0; i < 26; i++)
				SendDlgItemMessage(hwndDlg, IDC_SWING_STRENGTH, CB_ADDSTRING, (WPARAM)0, (LPARAM)(swingstrength[i]));
			SendDlgItemMessage(hwndDlg, IDC_SWING_STRENGTH, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);

			const char* patternsize[] = {"8", "16", "32", "64"};
			for (int i = 0; i < 4; i++)
				SendDlgItemMessage(hwndDlg, IDC_PATTERN_SIZE, CB_ADDSTRING, (WPARAM)0, (LPARAM)(patternsize[i]));
			SendDlgItemMessage(hwndDlg, IDC_PATTERN_SIZE, CB_SETCURSEL, (WPARAM)1, (LPARAM)0);

			const char* patternquant[] = {"1/4", "1/8", "1/16", "1/32", "1/64", "1/128"};
			for (int i = 0; i < 6; i++)
				SendDlgItemMessage(hwndDlg, IDC_PATTERN_QUANT, CB_ADDSTRING, (WPARAM)0, (LPARAM)(patternquant[i]));
			SendDlgItemMessage(hwndDlg, IDC_PATTERN_QUANT, CB_SETCURSEL, (WPARAM)2, (LPARAM)0);

			const char* objformat[] = {"Windows OBJ", "Linux ELF", "OSX MACHO"};
			for (int i = 0; i < 3; i++)
				SendDlgItemMessage(hwndDlg, IDC_OBJFORMAT, CB_ADDSTRING, (WPARAM)0, (LPARAM)(objformat[i]));
			SendDlgItemMessage(hwndDlg, IDC_OBJFORMAT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);

			const char* polyphony[] = { "1xPolyphone", "2xPolyphone" };
			for (int i = 0; i < MAX_POLYPHONY; i++)
				SendDlgItemMessage(hwndDlg, IDC_POLYPHONY, CB_ADDSTRING, (WPARAM)0, (LPARAM)(polyphony[i]));
			SendDlgItemMessage(hwndDlg, IDC_POLYPHONY, CB_SETCURSEL, (WPARAM)1, (LPARAM)0);

			SendDlgItemMessage(hwndDlg, IDC_CLIPOUTPUT, BM_SETCHECK, 1, 0);

            SendDlgItemMessage(hwndDlg, IDC_16BIT, BM_SETCHECK, 1, 0); // by default, 16 bit output

			return TRUE;
		}
		case WM_COMMAND:
		{
			// channel combo box 
			if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_INSTRUMENT))
			{
				if (HIWORD(wParam) == CBN_SELCHANGE)
                    return SwitchSelectedInstrument(hwndDlg, SendDlgItemMessage(hwndDlg, IDC_INSTRUMENT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0));
			}
			// Input MIDI Channel for the selected instrument
            else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_INSTRUMENT_MIDI_CHANNEL))
			{
				if (HIWORD(wParam) == CBN_SELCHANGE)
				{
                    SynthObjP->InstrumentInputSettings[SelectedInstrument].MIDIChannel = SendDlgItemMessage(hwndDlg, IDC_INSTRUMENT_MIDI_CHANNEL, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
					// get values from synth and set gui accordingly ...
					UpdateControls(SelectedInstrument);
					Go4kVSTi_UpdateDelayAndSlewTimes();
					Go4kVSTi_ClearDelayLinesAndSlews();
				}
				return TRUE;
			}

			// Polyphony
			else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_POLYPHONY))
			{
				if (HIWORD(wParam) == CBN_SELCHANGE)
				{
					SynthObjP->Polyphony = 1+SendDlgItemMessage(hwndDlg, IDC_POLYPHONY, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
					// get values from synth and set gui accordingly ...
					UpdateControls(SelectedInstrument);
					Go4kVSTi_UpdateDelayAndSlewTimes();
					Go4kVSTi_ClearDelayLinesAndSlews();
				}
				return TRUE;
			}
			// solo check box
			else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_SOLO))
			{
				if (SendDlgItemMessage(hwndDlg, IDC_SOLO, BM_GETCHECK, 0, 0)==BST_CHECKED) 
					Go4kVSTi_Solo(SelectedInstrument, true);
				else
					Go4kVSTi_Solo(SelectedInstrument, false);
				return TRUE;
			}
			else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_INSTRUMENT_NAME))
			{
				if (HIWORD(wParam) == EN_CHANGE)
				{
					GetDlgItemText(hwndDlg, IDC_INSTRUMENT_NAME, (LPSTR)&(SynthObjP->InstrumentNames[SelectedInstrument]), 127);
					int stsel = SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
					SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
					SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"Local");
					for (int i = 0; i < MAX_INSTRUMENTS; i++)
					{
						char instname[128];
                        if (SynthObjP)
						    sprintf(instname, "%d: %s", i+1, SynthObjP->InstrumentNames[i]);
                        else
                            sprintf(instname, "NULL hevoinen");
						SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)instname);
					}
					SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"Global");
					SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_SETCURSEL, (WPARAM)stsel, (LPARAM)0);

                    UpdateInstrumentSelectionCombobox(hwndDlg);
					return TRUE;
				}
			}
            else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_INSTRUMENT_NOTE_LO))
			{
				if (HIWORD(wParam) == EN_CHANGE)
				{
                    SynthObjP->InstrumentInputSettings[SelectedInstrument].NoteLo = GetNumberFromControl(hwndDlg, IDC_INSTRUMENT_NOTE_LO, 0, 127);
                    UpdateInstrumentSelectionCombobox(hwndDlg);
					return TRUE;
				}
			}
            else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_INSTRUMENT_NOTE_HI))
			{
				if (HIWORD(wParam) == EN_CHANGE)
				{
                    SynthObjP->InstrumentInputSettings[SelectedInstrument].NoteHi = GetNumberFromControl(hwndDlg, IDC_INSTRUMENT_NOTE_HI, 0, 127);
                    UpdateInstrumentSelectionCombobox(hwndDlg);
					return TRUE;
				}
			}
            else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_INSTRUMENT_VELOCITY_LO))
			{
				if (HIWORD(wParam) == EN_CHANGE)
				{
                    SynthObjP->InstrumentInputSettings[SelectedInstrument].VelocityLo = GetNumberFromControl(hwndDlg, IDC_INSTRUMENT_VELOCITY_LO, 1, 127);
                    UpdateInstrumentSelectionCombobox(hwndDlg);
					return TRUE;
				}
			}
            else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_INSTRUMENT_VELOCITY_HI))
			{
				if (HIWORD(wParam) == EN_CHANGE)
				{
                    SynthObjP->InstrumentInputSettings[SelectedInstrument].VelocityHi = GetNumberFromControl(hwndDlg, IDC_INSTRUMENT_VELOCITY_HI, 1, 127);
                    UpdateInstrumentSelectionCombobox(hwndDlg);
					return TRUE;
				}
			}
            else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_INSTRUMENT_MAP_VELOCITY_TO_NOTE))
			{
				if (SendDlgItemMessage(hwndDlg, IDC_INSTRUMENT_MAP_VELOCITY_TO_NOTE, BM_GETCHECK, 0, 0)==BST_CHECKED)
					SynthObjP->InstrumentInputSettings[SelectedInstrument].MapVelocityToNote = 1;
				else
					SynthObjP->InstrumentInputSettings[SelectedInstrument].MapVelocityToNote = 0;
				return TRUE;
			}

            else if ((((HWND)lParam) == GetDlgItem(hwndDlg, IDC_AUTO_PANIC_ON_INVALID_OPERATION)) ||
                     (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_AUTO_PANIC_ON_DENORMAL)) ||
                     (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_AUTO_PANIC)))
            {
                UpdateAutoPanicFromGUI();
				return TRUE;
			}

			// always on top check box
			else if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_ALWAYSONTOP))
			{
				if (SendDlgItemMessage(hwndDlg, IDC_ALWAYSONTOP, BM_GETCHECK, 0, 0)==BST_CHECKED)
				{
					SetWindowPos(hwndDlg, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED);					
				}
				else
				{
					SetWindowPos(hwndDlg, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED);
				}
				return TRUE;
			}
			else
			{
				// global Buttons
				switch (LOWORD(wParam)) 
			    { 
					case IDC_LOAD_PATCH:
					{
						char lpstrFilter[256] =
						{'4','k','l','a','n','g',' ','P','a','t','c','h', 0 ,
						'*','.','4','k','p', 0, 
						0};
						char lpstrFile[4096] = "*.4kp";
						char lpstrDirectory[4096];
						GetCurrentDirectory(4096, lpstrDirectory);

						OPENFILENAME ofn =
						{
							sizeof(OPENFILENAME),
							DialogWnd,
							GetModuleHandle(NULL),
							lpstrFilter,
							NULL,
							0,
							1,
							lpstrFile,
							256,
							NULL,
							0,
							lpstrDirectory,
							NULL,
							0,
							0,
							0,
							"4kp",
							0,
							NULL,
							NULL		
						};
						
						if (GetOpenFileName(&ofn))
						{
							// reset scroll pos
							int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
							SCROLLINFO sinfo;
							sinfo.cbSize = sizeof(SCROLLINFO);
							sinfo.fMask = SIF_POS;
							sinfo.nPos = 0;
							SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);
							ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);
							for (int i=0; i<MAX_INSTRUMENTS; i++)
								InstrumentScrollPos[i] = 0;
							GlobalScrollPos = 0;
							
							for (int i = 0; i < MAX_INSTRUMENTS; i++)
							{
								SynthObjP->InstrumentSignalValid[i] = 0;
							}
							SynthObjP->GlobalSignalValid = 0;
							Go4kVSTi_LoadPatch(ofn.lpstrFile);
							//instrument check
							for (int i = 0; i < MAX_INSTRUMENTS; i++)
							{
								UpdateSignalCount(i);
							}
							UpdateControls(SelectedInstrument);
						}
						return TRUE;
					}
					case IDC_SAVE_PATCH:
					{
						char lpstrFilter[256] =
						{'4','k','l','a','n','g',' ','P','a','t','c','h', 0 ,
						'*','.','4','k','p', 0, 
						0};
						char lpstrFile[4096] = "*.4kp";
						char lpstrDirectory[4096];
						GetCurrentDirectory(4096, lpstrDirectory);

						OPENFILENAME ofn =
						{
							sizeof(OPENFILENAME),
							DialogWnd,
							GetModuleHandle(NULL),
							lpstrFilter,
							NULL,
							0,
							1,
							lpstrFile,
							256,
							NULL,
							0,
							lpstrDirectory,
							NULL,
							OFN_OVERWRITEPROMPT,
							0,
							0,
							"4kp",
							0,
							NULL,
							NULL		
						};

						if (GetSaveFileName(&ofn))
						{
							Go4kVSTi_SavePatch(ofn.lpstrFile);
						}
						return TRUE;
					}
					case IDC_RESET_PATCH:
					{
						if (MessageBox(0, "Do you really want to reset the patch?", "Info", MB_YESNO) == IDYES)
						{
							// reset scroll pos
							int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
							SCROLLINFO sinfo;
							sinfo.cbSize = sizeof(SCROLLINFO);
							sinfo.fMask = SIF_POS;
							sinfo.nPos = 0;
							SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);
							ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);
							for (int i=0; i<MAX_INSTRUMENTS; i++)
								InstrumentScrollPos[i] = 0;
							GlobalScrollPos = 0;

							for (int i = 0; i < MAX_INSTRUMENTS; i++)
							{
								SynthObjP->InstrumentSignalValid[i] = 0;
							}
							SynthObjP->GlobalSignalValid = 0;
							Go4kVSTi_ResetPatch();
							UpdateControls(SelectedInstrument);						
						}
						return TRUE;
					}
					case IDC_LOAD_INSTRUMENT:
					{
						char lpstrFilter[256] =
						{'4','k','l','a','n','g',' ','I','n','s','t','r','u','m','e','n','t', 0 ,
						'*','.','4','k','i', 0, 
						0};
						char lpstrFile[4096] = "*.4ki";
						char lpstrDirectory[4096];
						GetCurrentDirectory(4096, lpstrDirectory);

						OPENFILENAME ofn =
						{
							sizeof(OPENFILENAME),
							DialogWnd,
							GetModuleHandle(NULL),
							lpstrFilter,
							NULL,
							0,
							1,
							lpstrFile,
							256,
							NULL,
							0,
							lpstrDirectory,
							NULL,
							0,
							0,
							0,
							"4ki",
							0,
							NULL,
							NULL		
						};
						
						if (GetOpenFileName(&ofn))
						{
							// reset scroll pos
							int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
							SCROLLINFO sinfo;
							sinfo.cbSize = sizeof(SCROLLINFO);
							sinfo.fMask = SIF_POS;
							sinfo.nPos = 0;
							SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);
							ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);
														
							UpdateControls(SelectedInstrument);
							if (SelectedTab == T_INSTRUMENT)
							{
								InstrumentScrollPos[SelectedInstrument] = 0;
								SynthObjP->InstrumentSignalValid[SelectedInstrument] = 0;
								Go4kVSTi_LoadInstrument(ofn.lpstrFile, (char)SelectedInstrument);
							}
							else
							{
								GlobalScrollPos = 0;
								SynthObjP->GlobalSignalValid = 0;
								Go4kVSTi_LoadInstrument(ofn.lpstrFile, (char)16);
							}
							UpdateControls(SelectedInstrument);			
						}
						return TRUE;
					}
					case IDC_SAVE_INSTRUMENT:
					{
						char lpstrFilter[256] =
						{'4','k','l','a','n','g',' ','I','n','s','t','r','u','m','e','n','t', 0 ,
						'*','.','4','k','i', 0, 
						0};
						char lpstrFile[4096] = "*.4ki";
						char lpstrDirectory[4096];
						GetCurrentDirectory(4096, lpstrDirectory);

						OPENFILENAME ofn =
						{
							sizeof(OPENFILENAME),
							DialogWnd,
							GetModuleHandle(NULL),
							lpstrFilter,
							NULL,
							0,
							1,
							lpstrFile,
							256,
							NULL,
							0,
							lpstrDirectory,
							NULL,
							OFN_OVERWRITEPROMPT,
							0,
							0,
							"4ki",
							0,
							NULL,
							NULL		
						};

						if (GetSaveFileName(&ofn))
						{
							if (SelectedTab == T_INSTRUMENT)
								Go4kVSTi_SaveInstrument(ofn.lpstrFile, (char)SelectedInstrument);
							else
								Go4kVSTi_SaveInstrument(ofn.lpstrFile, (char)16);
						}
						return TRUE;
					}
					case IDC_RESET_INSTRUMENT:
					{
						if (MessageBox(0, "Do you really want to reset the instrument?", "Info", MB_YESNO) == IDYES)
						{
							if (SelectedTab == T_INSTRUMENT)
							{
								// reset scroll pos
								int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
								SCROLLINFO sinfo;
								sinfo.cbSize = sizeof(SCROLLINFO);
								sinfo.fMask = SIF_POS;
								sinfo.nPos = 0;
								SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);
								ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);
								InstrumentScrollPos[SelectedInstrument] = 0;
								
								SynthObjP->InstrumentSignalValid[SelectedInstrument] = 0;
								Go4kVSTi_ResetInstrument(SelectedInstrument, true);
							}
							else
							{
								// reset scroll pos
								int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
								SCROLLINFO sinfo;
								sinfo.cbSize = sizeof(SCROLLINFO);
								sinfo.fMask = SIF_POS;
								sinfo.nPos = 0;
								SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);
								ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);
								GlobalScrollPos = 0;

								SynthObjP->GlobalSignalValid = 0;
								Go4kVSTi_ResetGlobalUnits(true);
							}
							UpdateControls(SelectedInstrument);						
						}
						return TRUE;
					}
					case IDC_UNIT_LOAD:
					{
						char lpstrFilter[256] =
						{'4','k','l','a','n','g',' ','U','n','i','t', 0 ,
						'*','.','4','k','u', 0, 
						0};
						char lpstrFile[4096] = "*.4ku";
						char lpstrDirectory[4096];
						GetCurrentDirectory(4096, lpstrDirectory);

						OPENFILENAME ofn =
						{
							sizeof(OPENFILENAME),
							DialogWnd,
							GetModuleHandle(NULL),
							lpstrFilter,
							NULL,
							0,
							1,
							lpstrFile,
							256,
							NULL,
							0,
							lpstrDirectory,
							NULL,
							0,
							0,
							0,
							"4ku",
							0,
							NULL,
							NULL		
						};
						
						if (GetOpenFileName(&ofn))
						{
							BYTE* uslot;
							if (SelectedTab == T_INSTRUMENT)
							{
								SynthObjP->InstrumentSignalValid[SelectedInstrument] = 0;
								uslot = SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit];
							}
							else
							{
								SynthObjP->GlobalSignalValid = 0;
								uslot = SynthObjP->GlobalValues[SelectedGUnit];
							}
							Go4kVSTi_LoadUnit(ofn.lpstrFile, uslot);
							UpdateControls(SelectedInstrument);
						}
						return TRUE;
					}
					case IDC_UNIT_SAVE:
					{
						char lpstrFilter[256] =
						{'4','k','l','a','n','g',' ','U','n','i','t', 0 ,
						'*','.','4','k','u', 0, 
						0};
						char lpstrFile[4096] = "*.4ku";
						char lpstrDirectory[4096];
						GetCurrentDirectory(4096, lpstrDirectory);

						OPENFILENAME ofn =
						{
							sizeof(OPENFILENAME),
							DialogWnd,
							GetModuleHandle(NULL),
							lpstrFilter,
							NULL,
							0,
							1,
							lpstrFile,
							256,
							NULL,
							0,
							lpstrDirectory,
							NULL,
							OFN_OVERWRITEPROMPT,
							0,
							0,
							"4ku",
							0,
							NULL,
							NULL		
						};

						if (GetSaveFileName(&ofn))
						{
							BYTE* uslot;
							if (SelectedTab == T_INSTRUMENT)
							{
								uslot = SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit];
							}
							else
							{
								uslot = SynthObjP->GlobalValues[SelectedGUnit];
							}
							Go4kVSTi_SaveUnit(ofn.lpstrFile, uslot);
						}
						return TRUE;
					}
					case IDC_RECORD_BUTTON:
					{
						EnableWindow(GetDlgItem(DialogWnd, IDC_RECORD_BUTTON), false);
						EnableWindow(GetDlgItem(DialogWnd, IDC_STOP_BUTTON), true);
						int psize = SendDlgItemMessage(hwndDlg, IDC_PATTERN_SIZE, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
                        int patternsize = ConvertPatternSizeFromGUI(psize);
						float patternquant = 1.0;
						int pquant = SendDlgItemMessage(hwndDlg, IDC_PATTERN_QUANT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
						if (pquant == 0)
							patternquant = 0.25;
						if (pquant == 1)
							patternquant = 0.5;
						if (pquant == 3)
							patternquant = 2.0;
						if (pquant == 4)
							patternquant = 4.0;
						if (pquant == 5)
							patternquant = 8.0;
						Go4kVSTi_Record(true, patternsize, patternquant);
						return TRUE;
					}
					case IDC_STOP_BUTTON:
					{
						EnableWindow(GetDlgItem(DialogWnd, IDC_RECORD_BUTTON), true);
						EnableWindow(GetDlgItem(DialogWnd, IDC_STOP_BUTTON), false);
						Go4kVSTi_Record(false, 0, 0);
						return TRUE;
					}
					case IDC_PANIC:
					{
                        GetSynthMutex();
						Go4kVSTi_Panic();
                        SynthObjP->PanicCount = 0; // reset panic count on manual panic button press
                        ReleaseSynthMutex();
						return TRUE;
					}
					case IDC_UNIT_RESET:
					{
						if (SelectedTab == T_INSTRUMENT)
							Go4kVSTi_InitInstrumentSlot(SelectedInstrument, SelectedIUnit, SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit][0]);
						else
							Go4kVSTi_InitGlobalSlot(SelectedGUnit, SynthObjP->GlobalValues[SelectedGUnit][0]);
						UpdateControls(SelectedInstrument);	
						return TRUE;
					}
					case IDC_UNIT_COPY:
					{
						if (SelectedTab == T_INSTRUMENT)
							memcpy(UnitCopyBuffer, SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit], MAX_SLOT_VALUES);
						else
							memcpy(UnitCopyBuffer, SynthObjP->GlobalValues[SelectedGUnit], MAX_SLOT_VALUES);
						return TRUE;
					}
					case IDC_UNIT_PASTE:
					{
						if (SelectedTab == T_INSTRUMENT)
							memcpy(SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit], UnitCopyBuffer, MAX_SLOT_VALUES);
						else
							memcpy(SynthObjP->GlobalValues[SelectedGUnit], UnitCopyBuffer, MAX_SLOT_VALUES);
						UpdateControls(SelectedInstrument);	
						return TRUE;
					}
                    case IDC_VOICECOUNT1 : return SwitchSelectedInstrument(hwndDlg, 0);
                    case IDC_VOICECOUNT2 : return SwitchSelectedInstrument(hwndDlg, 1);
                    case IDC_VOICECOUNT3 : return SwitchSelectedInstrument(hwndDlg, 2);
                    case IDC_VOICECOUNT4 : return SwitchSelectedInstrument(hwndDlg, 3);
                    case IDC_VOICECOUNT5 : return SwitchSelectedInstrument(hwndDlg, 4);
                    case IDC_VOICECOUNT6 : return SwitchSelectedInstrument(hwndDlg, 5);
                    case IDC_VOICECOUNT7 : return SwitchSelectedInstrument(hwndDlg, 6);
                    case IDC_VOICECOUNT8 : return SwitchSelectedInstrument(hwndDlg, 7);
                    case IDC_VOICECOUNT9 : return SwitchSelectedInstrument(hwndDlg, 8);
                    case IDC_VOICECOUNT10 : return SwitchSelectedInstrument(hwndDlg, 9);
                    case IDC_VOICECOUNT11 : return SwitchSelectedInstrument(hwndDlg, 10);
                    case IDC_VOICECOUNT12 : return SwitchSelectedInstrument(hwndDlg, 11);
                    case IDC_VOICECOUNT13 : return SwitchSelectedInstrument(hwndDlg, 12);
                    case IDC_VOICECOUNT14 : return SwitchSelectedInstrument(hwndDlg, 13);
                    case IDC_VOICECOUNT15 : return SwitchSelectedInstrument(hwndDlg, 14);
                    case IDC_VOICECOUNT16 : return SwitchSelectedInstrument(hwndDlg, 15);
                    case IDC_VOICECOUNT17 : return SwitchSelectedInstrument(hwndDlg, 16);
                    case IDC_VOICECOUNT18 : return SwitchSelectedInstrument(hwndDlg, 17);
                    case IDC_VOICECOUNT19 : return SwitchSelectedInstrument(hwndDlg, 18);
                    case IDC_VOICECOUNT20 : return SwitchSelectedInstrument(hwndDlg, 19);
                    case IDC_VOICECOUNT21 : return SwitchSelectedInstrument(hwndDlg, 20);
                    case IDC_VOICECOUNT22 : return SwitchSelectedInstrument(hwndDlg, 21);
                    case IDC_VOICECOUNT23 : return SwitchSelectedInstrument(hwndDlg, 22);
                    case IDC_VOICECOUNT24 : return SwitchSelectedInstrument(hwndDlg, 23);
                    case IDC_VOICECOUNT25 : return SwitchSelectedInstrument(hwndDlg, 24);
                    case IDC_VOICECOUNT26 : return SwitchSelectedInstrument(hwndDlg, 25);
                    case IDC_VOICECOUNT27 : return SwitchSelectedInstrument(hwndDlg, 26);
                    case IDC_VOICECOUNT28 : return SwitchSelectedInstrument(hwndDlg, 27);
                    case IDC_VOICECOUNT29 : return SwitchSelectedInstrument(hwndDlg, 28);
                    case IDC_VOICECOUNT30 : return SwitchSelectedInstrument(hwndDlg, 29);
                    case IDC_VOICECOUNT31 : return SwitchSelectedInstrument(hwndDlg, 30);
                    case IDC_VOICECOUNT32 : return SwitchSelectedInstrument(hwndDlg, 31);
                    case IDC_VOICECOUNT33 : return SwitchSelectedInstrument(hwndDlg, 32);
                    case IDC_VOICECOUNT34 : return SwitchSelectedInstrument(hwndDlg, 33);
                    case IDC_VOICECOUNT35 : return SwitchSelectedInstrument(hwndDlg, 34);
                    case IDC_VOICECOUNT36 : return SwitchSelectedInstrument(hwndDlg, 35);
                    case IDC_VOICECOUNT37 : return SwitchSelectedInstrument(hwndDlg, 36);
                    case IDC_VOICECOUNT38 : return SwitchSelectedInstrument(hwndDlg, 37);
                    case IDC_VOICECOUNT39 : return SwitchSelectedInstrument(hwndDlg, 38);
                    case IDC_VOICECOUNT40 : return SwitchSelectedInstrument(hwndDlg, 39);
                    case IDC_VOICECOUNT41 : return SwitchSelectedInstrument(hwndDlg, 40);
                    case IDC_VOICECOUNT42 : return SwitchSelectedInstrument(hwndDlg, 41);
                    case IDC_VOICECOUNT43 : return SwitchSelectedInstrument(hwndDlg, 42);
                    case IDC_VOICECOUNT44 : return SwitchSelectedInstrument(hwndDlg, 43);
                    case IDC_VOICECOUNT45 : return SwitchSelectedInstrument(hwndDlg, 44);
                    case IDC_VOICECOUNT46 : return SwitchSelectedInstrument(hwndDlg, 45);
                    case IDC_VOICECOUNT47 : return SwitchSelectedInstrument(hwndDlg, 46);
                    case IDC_VOICECOUNT48 : return SwitchSelectedInstrument(hwndDlg, 47);

					return FALSE;
				} // switch (LOWORD(wParam)) 
				return FALSE;
			} // else, global buttons
			return FALSE;
		}
/*		case WM_MOUSEWHEEL:
		{
			if (InitDone)
			{
				int scrollw = 16*((int)((short)HIWORD(wParam)))/120;

				int oldpos = GetScrollPos(ScrollWnd, SB_VERT);				
				SetScrollPos(ScrollWnd, SB_VERT, oldpos-scrollw, true);
				int newpos = GetScrollPos(ScrollWnd, SB_VERT);
				ScrollWindow(ScrollWnd, 0, oldpos - newpos, NULL, NULL);		
				if (SelectedTab == T_INSTRUMENT)
				{
					InstrumentScrollPos[SelectedInstrument] = newpos;
				}
				else if (SelectedTab == T_GLOBAL)
				{
					GlobalScrollPos = newpos;
				}
				return TRUE;
			}
			break;
		}
*/
		case WM_NOTIFY:
		{
			if (((LPNMHDR)lParam)->hwndFrom == GetDlgItem(hwndDlg, IDC_MAINTAB))
			{
				if (((LPNMHDR)lParam)->code == TCN_SELCHANGE)
				{
					int i = TabCtrl_GetCurSel(GetDlgItem(hwndDlg, IDC_MAINTAB));
					TabChanged(i);
					return TRUE;
				}					
			}
			break;
		}
		case WM_CLOSE:
		{
			ShowWindow( DialogWnd, SW_SHOWMINIMIZED );
			return TRUE;
		}
		case WM_DROPFILES:
		{
			char DropFileName[4096];
			DragQueryFile((HDROP)wParam, 0, DropFileName, 4096);
			std::string filename = DropFileName;
			int slen = filename.length();
			bool invalidFile = false;
			if (slen > 4)
			{
				std::string extension = filename.substr(slen-4, 4);
				// dropped patch?
				if (!extension.compare(".4kp"))
				{
					// reset scroll pos
					int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
					SCROLLINFO sinfo;
					sinfo.cbSize = sizeof(SCROLLINFO);
					sinfo.fMask = SIF_POS;
					sinfo.nPos = 0;
					SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);
					ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);
					for (int i=0; i<MAX_INSTRUMENTS; i++)
						InstrumentScrollPos[i] = 0;
					GlobalScrollPos = 0;

					for (int i = 0; i < MAX_INSTRUMENTS; i++)
					{
						SynthObjP->InstrumentSignalValid[i] = 0;
					}
					SynthObjP->GlobalSignalValid = 0;
					Go4kVSTi_LoadPatch(DropFileName);
					//instrument check
					for (int i = 0; i < MAX_INSTRUMENTS; i++)
					{
						UpdateSignalCount(i);
					}
					UpdateControls(SelectedInstrument);
				}
				// dropped instrument?
				else if (!extension.compare(".4ki"))
				{
					// reset scroll pos
					int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
					SCROLLINFO sinfo;
					sinfo.cbSize = sizeof(SCROLLINFO);
					sinfo.fMask = SIF_POS;
					sinfo.nPos = 0;
					SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);
					ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);
												
					UpdateControls(SelectedInstrument);
					if (SelectedTab == T_INSTRUMENT)
					{
						InstrumentScrollPos[SelectedInstrument] = 0;
						SynthObjP->InstrumentSignalValid[SelectedInstrument] = 0;
						Go4kVSTi_LoadInstrument(DropFileName, (char)SelectedInstrument);
					}
					else
					{
						GlobalScrollPos = 0;
						SynthObjP->GlobalSignalValid = 0;
						Go4kVSTi_LoadInstrument(DropFileName, (char)16);
					}
					UpdateControls(SelectedInstrument);
				}
				// dropped shit!
				else
				{
					invalidFile = true;
				}
			}
			else
			{
				invalidFile = true;
			}
			if (invalidFile)
			{
				MessageBox(0,DropFileName,"Only patches(4kp) and instruments(4ki) allowed!",MB_OK);
			}
			return TRUE;
		} // case WM_DROPFILES
/*		case WM_CTLCOLORSTATIC:
		{
            //SetBkMode((HDC)wParam,TRANSPARENT);
            SetBkMode((HDC)wParam,OPAQUE);
            //SetTextColor((HDC)wParam, RGB(255,rand()%256,rand()%256));
            SetTextColor((HDC)wParam, GetSysColor(COLOR_BTNTEXT));
            SetBkColor((HDC)wParam, RGB(0,rand()%256,rand()%256));
            return (INT_PTR)GetSysColorBrush(COLOR_MENU);            
        }  */

    case WM_DRAWITEM:
        // owner-draw voice count buttons
        if ((wParam >= IDC_VOICECOUNT1) && (wParam <= IDC_VOICECOUNT48))
        {
            int instr_idx = wParam - IDC_VOICECOUNT1;
            LPDRAWITEMSTRUCT lpDIS = (LPDRAWITEMSTRUCT)lParam;
            int c = SynthObjP->InstrumentFPUProblemsCount[instr_idx];
            if (c > (FPU_PROBLEM_INCREMENT * 2)) c = (FPU_PROBLEM_INCREMENT * 2);
            char buttontext[3];
            HWND hwndButton = GetDlgItem(hwndDlg, wParam);
            int buttontextlen = GetWindowText(hwndButton, buttontext, 3);
            if (!buttontextlen)
                buttontext[0] = 0;

            // highlight FPU problems
            if (c)
            {
                SetTextColor(lpDIS->hDC, RGB(255,200+c*10,200+c*10));
                SetBkColor(lpDIS->hDC, RGB(255, 0, 0));
                SetDCBrushColor(lpDIS->hDC, RGB(255, 0, 0));
            }
            else            
            {
                SetTextColor(lpDIS->hDC, RGB(200,200,200));
                SetBkColor(lpDIS->hDC, RGB(0, 0, 0));
                SetDCBrushColor(lpDIS->hDC, RGB(0, 0, 0));
            }

            SelectObject(lpDIS->hDC, GetStockObject(DC_BRUSH));
            RoundRect(lpDIS->hDC, lpDIS->rcItem.left, lpDIS->rcItem.top, lpDIS->rcItem.right, lpDIS->rcItem.bottom, 5, 5);
            TextOut(lpDIS->hDC, lpDIS->rcItem.left + (3 - buttontextlen) * 3, lpDIS->rcItem.top + 3, buttontext, buttontextlen);

            return TRUE;
        }
        break;

	} // switch (message) 
    return FALSE; 
} // MainDialogProc()

// CB for the tabs
BOOL CALLBACK ModuleSettingsDialogProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) 
{ 
	switch (message) 
    { 
        case WM_COMMAND: 
			return ButtonPressed(wParam, lParam);

		case WM_HSCROLL:
			return ScrollbarChanged(hwndDlg, wParam, lParam);		
    } 
    return FALSE; 
} 

// CB for the tabs
BOOL CALLBACK ScrollDialogProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) 
{ 
	switch (message) 
    {
		case WM_VSCROLL:
		{
			RECT rect;
			GetWindowRect(ScrollWnd, &rect);
			int pagesize = 1+rect.bottom - rect.top;
			int linesize = 16;
			int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
			switch (LOWORD(wParam))
			{	
				
				case SB_BOTTOM: // Scrolls to the lower right. 
					break;
				case SB_ENDSCROLL: // Ends scroll. 
					break;
				case SB_LINEDOWN: // Scrolls one line down. 
					SetScrollPos(ScrollWnd, SB_VERT, oldpos+linesize, true);
					break;
				case SB_LINEUP: // Scrolls one line up. 
					SetScrollPos(ScrollWnd, SB_VERT, oldpos-linesize, true);
					break;
				case SB_PAGEDOWN: // Scrolls one page down. 
					SetScrollPos(ScrollWnd, SB_VERT, oldpos+pagesize, true);
					break;
				case SB_PAGEUP: // Scrolls one page up. 
					SetScrollPos(ScrollWnd, SB_VERT, oldpos-pagesize, true);
					break;
				case SB_THUMBPOSITION: // The user has dragged the scroll box (thumb) and released the mouse button. The high-order word indicates the position of the scroll box at the end of the drag operation. 
				case SB_THUMBTRACK: // The user is dragging the scroll box. This message is sent repeatedly until the user releases the mouse button. The high-order word indicates the position that the scroll box has been dragged to. 
					SetScrollPos(ScrollWnd, SB_VERT, HIWORD(wParam), true);
					break;
				case SB_TOP: // Scrolls to the upper left.
					break;
			}
			int newpos = GetScrollPos(ScrollWnd, SB_VERT);
			ScrollWindow(ScrollWnd, 0, oldpos - newpos, NULL, NULL);		

			if (SelectedTab == T_INSTRUMENT)
			{
				InstrumentScrollPos[SelectedInstrument] = newpos;
			}
			else if (SelectedTab == T_GLOBAL)
			{
				GlobalScrollPos = newpos;
			}
			return TRUE;
		}
/*		case WM_MOUSEWHEEL:
		{
			if (InitDone)
			{
				int scrollw = 16*((int)((short)HIWORD(wParam)))/120;

				int oldpos = GetScrollPos(ScrollWnd, SB_VERT);				
				SetScrollPos(ScrollWnd, SB_VERT, oldpos-scrollw, true);
				int newpos = GetScrollPos(ScrollWnd, SB_VERT);
				ScrollWindow(ScrollWnd, 0, oldpos - newpos, NULL, NULL);		
				if (SelectedTab == T_INSTRUMENT)
				{
					InstrumentScrollPos[SelectedInstrument] = newpos;
				}
				else if (SelectedTab == T_GLOBAL)
				{
					GlobalScrollPos = newpos;
				}
				return TRUE;
			}
			break;
		}
*/
		default:
			break;
	} 
    return FALSE; 
}

// CB for the tabs
BOOL CALLBACK StackDialogProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) 
{ 
    bool shift_pressed = GetKeyState(VK_SHIFT) & 0x8000;

	switch (message) 
    { 
        case WM_COMMAND: 
			return StackButtonPressed(wParam, shift_pressed);

        case WM_DRAWITEM:
            // owner-draw signal count buttons
            if ((wParam >= IDC_ISTACK_SIGNALCOUNT1) && (wParam <= IDC_ISTACK_SIGNALCOUNT32) && ((wParam - IDC_ISTACK_SIGNALCOUNT1) % 6 == 0))
            {
                int unit_idx = (wParam - IDC_ISTACK_SIGNALCOUNT1) / 6;
                LPDRAWITEMSTRUCT lpDIS = (LPDRAWITEMSTRUCT)lParam;

                int c = SynthObjP->UnitFPUProblemsCount[SelectedInstrument*MAX_UNITS+unit_idx];
                if (c > (FPU_PROBLEM_INCREMENT * 2)) c = (FPU_PROBLEM_INCREMENT * 2);
                char buttontext[4];
                HWND hwndButton = GetDlgItem(hwndDlg, wParam);
                int buttontextlen = GetWindowText(hwndButton, buttontext, 4);
                if (!buttontextlen)
                    buttontext[0] = 0;

                // highlight FPU problems
                if (c)
                {
                    SetTextColor(lpDIS->hDC, RGB(255,200+c*10,200+c*10));
                    SetBkColor(lpDIS->hDC, RGB(255, 0, 0));
                    SetDCBrushColor(lpDIS->hDC, RGB(255, 0, 0));
                }
                else if (buttontextlen > 2) // "inv"
                {
                    SetTextColor(lpDIS->hDC, RGB(255,255,200));
                    SetBkColor(lpDIS->hDC, RGB(80, 40, 0));
                    SetDCBrushColor(lpDIS->hDC, RGB(80, 40, 0));
                }
                else
                {
                    SetTextColor(lpDIS->hDC, RGB(200,200,200));
                    SetBkColor(lpDIS->hDC, RGB(0, 0, 0));
                    SetDCBrushColor(lpDIS->hDC, RGB(0, 0, 0));
                }

                SelectObject(lpDIS->hDC, GetStockObject(DC_BRUSH));
                RoundRect(lpDIS->hDC, lpDIS->rcItem.left, lpDIS->rcItem.top, lpDIS->rcItem.right, lpDIS->rcItem.bottom, 5, 5);
                TextOut(lpDIS->hDC, lpDIS->rcItem.left + (4 - buttontextlen) * 3, lpDIS->rcItem.top + 3, buttontext, buttontextlen);

                return TRUE;
            }
            break;
    } // switch (message)
	return FALSE; 
} 

// CB for the set dialog
BOOL CALLBACK SetDialogProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
    {
        case WM_INITDIALOG:
		{
			for (int i = 1; i < NUM_MODULES; i++)
			{
				if (SelectedTab == T_INSTRUMENT && i == M_ACC)
					SendDlgItemMessage(hwndDlg, IDC_SET_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"");
				else
					SendDlgItemMessage(hwndDlg, IDC_SET_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)(UnitName[i]));
			}
			SendDlgItemMessage(hwndDlg, IDC_SET_UNIT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
			SetUnit = 1;
			return TRUE;
		}
        case WM_COMMAND:
		{
			if (((HWND)lParam) == GetDlgItem(hwndDlg, IDC_SET_UNIT))
			{
				if (HIWORD(wParam) == CBN_SELCHANGE)
				{
					SetUnit = 1+SendDlgItemMessage(hwndDlg, IDC_SET_UNIT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
				}
				return TRUE;
			}
            switch(LOWORD(wParam))
            {
                case IDOK:
					if (SelectedTab == T_INSTRUMENT && SetUnit == M_ACC)
						EndDialog(hwndDlg, IDCANCEL);
					else
						EndDialog(hwndDlg, IDOK);
                break;
                case IDCANCEL:
                    EndDialog(hwndDlg, IDCANCEL);
                break;
            }
			break;
		}
        default:
            return FALSE;
    }
    return TRUE;
}

VOID CALLBACK TimerProc(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
{    
	for (int i = 0; i < MAX_INSTRUMENTS; i++)
    {
		UpdateVoiceDisplay(i);
    }
    static int last_panic_count = -1;
    if ((SynthObjP->AutoPanicNotificationCounter) || (SynthObjP->PanicCount != last_panic_count))
    {
        if (SynthObjP->AutoPanicNotificationCounter > 1) 
            // when auto panic has worked, show it in the button
            SetWindowText(GetDlgItem(DialogWnd, IDC_PANIC), "** Auto Panic **");
        else
        {
	        char text[100];
            sprintf(text, "PANIC!\n(%d auto panics)", SynthObjP->PanicCount);
            SetWindowText(GetDlgItem(DialogWnd, IDC_PANIC), text);
        }
        SynthObjP->AutoPanicNotificationCounter--;
    }
    last_panic_count = SynthObjP->PanicCount;
}

VOID CALLBACK BackupTimerProc(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
{
	// Go4kVSTi_SavePatch("C:\\temp\\4klang.4kp");
}

void Go4kVSTiGUI_Create(HINSTANCE hInst)
{
	timer = 1;
	backupTimer = 2;
	timerID = 0;
	backupTimerID = 0;
	DialogWnd = NULL;
	SelectedIUnit = 0;
	SelectedGUnit = 0;
	SelectedTab = T_INSTRUMENT;
	SelectedInstrument = 0;
	SetUnit = -1;
	hInstance = hInst;
	InitCommonControls();
	DialogWnd = CreateDialog(hInst,MAKEINTRESOURCE(IDD_GO4KVSTIDIALOG),NULL,(DLGPROC)MainDialogProc);

	// tab dialogs
	HWND tmp = GetDlgItem(DialogWnd, IDC_MODULE_SETTINGS);
	for (int i = 0; i < NUM_MODULES; i++)
	{	
		ModuleWnd[M_NONE+i] = CreateDialog(hInst,MAKEINTRESOURCE(IDD_NONE+i),tmp,(DLGPROC)ModuleSettingsDialogProc);
		MoveWindow(ModuleWnd[M_NONE+i], 7, 43, 450, 280, true);
		ShowWindow(ModuleWnd[M_NONE+i], SW_HIDE);
	}

	tmp = GetDlgItem(DialogWnd, IDC_MAINTAB);
	ScrollWnd = CreateDialog(hInst,MAKEINTRESOURCE(IDD_SCROLLWINDOW),tmp,(DLGPROC)ScrollDialogProc);
	MoveWindow(ScrollWnd, 3, 23, 392, 410, true);
	ShowWindow(ScrollWnd, SW_SHOW);

	tmp = ScrollWnd;
	RECT rect;
	// instrument stack dialog
	TabWnd[T_INSTRUMENT] = CreateDialogA(hInst,MAKEINTRESOURCE(IDD_INSTRUMENT_STACK),tmp,(DLGPROC)StackDialogProc);
	GetWindowRect(TabWnd[T_INSTRUMENT], &rect);
	int virtualsizey = rect.bottom - rect.top;
	MoveWindow(TabWnd[T_INSTRUMENT], 5, 0, 365, virtualsizey, true);
	ShowWindow(TabWnd[T_INSTRUMENT], SW_SHOW);
	EnableWindow(GetDlgItem(TabWnd[T_INSTRUMENT], IDC_ISTACK_UNIT1), false);
	ShowWindow(GetDlgItem(DialogWnd, IDC_ISTACK_VALID), true);
	
	// global stack dialog
	TabWnd[T_GLOBAL] = CreateDialog(hInst,MAKEINTRESOURCE(IDD_GLOBAL_STACK),tmp,(DLGPROC)StackDialogProc);
	MoveWindow(TabWnd[T_GLOBAL], 5, 0, 365, virtualsizey, true);
	ShowWindow(TabWnd[T_GLOBAL], SW_HIDE);
	EnableWindow(GetDlgItem(TabWnd[T_GLOBAL], IDC_GSTACK_UNIT1), false);
	ShowWindow(GetDlgItem(DialogWnd, IDC_GSTACK_VALID), false);

	// setup scrollwindow
	GetWindowRect(ScrollWnd, &rect);
	int sizey = rect.bottom - rect.top;
	SCROLLINFO sinfo;
	sinfo.cbSize = sizeof(SCROLLINFO);
	sinfo.fMask = SIF_PAGE | SIF_RANGE | SIF_POS;
	sinfo.nPos = 0;
	sinfo.nPage = sizey;
	sinfo.nMin = 0;
	sinfo.nMax = virtualsizey+2;
	SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true); 

	for (int i=0; i<MAX_INSTRUMENTS; i++)
		InstrumentScrollPos[i] = 0;
	GlobalScrollPos = 0;

	Go4kVSTi_Init();
	SynthObjP = Go4kVSTi_GetSynthObject();
    GlobalSetsP = Go4kVSTi_GetGlobalSettings();    
	UpdateControls(SelectedInstrument);

	timerID = SetTimer(DialogWnd, timer, 200, TimerProc);
	backupTimerID = SetTimer(DialogWnd, backupTimer, 60*1000, BackupTimerProc);

	memset(UnitCopyBuffer, 0, MAX_SLOT_VALUES);

	// set fst instrument elements
	const char* idest[] = {"Local", "Instrument  1", "Instrument  2", "Instrument  3", "Instrument  4", "Instrument  5", "Instrument  6", "Instrument  7", "Instrument  8",
				"Instrument  9", "Instrument 10", "Instrument 11", "Instrument 12", "Instrument 13", "Instrument 14", "Instrument 15", "Instrument 16", "Global" };
	for (int i = 0; i < 18; i++)
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)(idest[i]));

	InitDone = true;
}

void Go4kVSTiGUI_Show(int showCommand)
{
	ShowWindow(DialogWnd, showCommand);
}

void Go4kVSTiGUI_Destroy()
{
	// Go4kVSTi_SavePatch("C:\\4klang.4kp");
/*
	if (MessageBox(NULL,"If you didnt save your 4klang patch yet, this information will be lost.\n"
					"Do you want to save the patch now?",
					"Reminder",
					MB_YESNO) == IDYES)
	{
		char lpstrFilter[256] =
		{'4','k','l','a','n','g',' ','P','a','t','c','h', 0 ,
		'*','.','4','k','p', 0, 
		0};
		char lpstrFile[4096] = "*.4kp";
		char lpstrDirectory[4096];
		GetCurrentDirectory(4096, lpstrDirectory);

		OPENFILENAME ofn =
		{
			sizeof(OPENFILENAME),
			DialogWnd,
			GetModuleHandle(NULL),
			lpstrFilter,
			NULL,
			0,
			1,
			lpstrFile,
			256,
			NULL,
			0,
			lpstrDirectory,
			NULL,
			OFN_OVERWRITEPROMPT,
			0,
			0,
			"4kp",
			0,
			NULL,
			NULL		
		};

		if (GetSaveFileName(&ofn))
		{
			Go4kVSTi_SavePatch(ofn.lpstrFile);
		}
	}	
*/
	KillTimer(DialogWnd, timerID);
	KillTimer(DialogWnd, backupTimerID);
	DestroyWindow(DialogWnd);


} // Go4kVSTiGUI_Destroy()


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//																		//
//                          Button Events								//
//																		//
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void SetButtonParams(int uid, BYTE* val, WPARAM id, LPARAM lParam)
{
	DWORD res;
	if (uid == M_VCO)
	{
		VCO_valP v = (VCO_valP)val;
		MaybeButtonGroupChanged(IDC_VCO_SINE, IDC_VCO_GATE, LOWORD(id), M_VCO, res);
		if (res)
		{
			if (SendDlgItemMessage(ModuleWnd[M_VCO], IDC_VCO_LFO, BM_GETCHECK, 0, 0)==BST_CHECKED)
				v->flags = VCO_LFO;
			else
				v->flags = 0;

			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_TRANSPOSE), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_TRANSPOSE_VAL), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_DETUNE), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_DETUNE_VAL), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_PHASE), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_PHASE_VAL), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR_VAL), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_SHAPE), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_SHAPE_VAL), true);
			EnableButtonGroup(IDC_VCO_GATE1, IDC_VCO_GATE16, M_VCO, false);
			if (IDC_VCO_GATE == res)
			{
				v->flags |= VCO_GATE;
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR_VAL), false);
				EnableButtonGroup(IDC_VCO_GATE1, IDC_VCO_GATE16, M_VCO, true);
				// adjust gate bits stored in color and shape with current selection of checkboxes
				WORD gatebits;
				GetCheckboxGroupBitmask(IDC_VCO_GATE1, IDC_VCO_GATE16, IDC_VCO_GATE1, M_VCO, gatebits, res);
				v->gate = gatebits & 0xff;
				v->color = (gatebits >> 8) & 0xff;
				// adjust color and shape sliders
				InitSliderCenter(ModuleWnd[M_VCO], IDC_VCO_COLOR, 0, 128, v->color);
			}
			else
			{
				// set color at least to valid range
				if (v->color > 128) v->color = 128;
				InitSliderCenter(ModuleWnd[M_VCO], IDC_VCO_COLOR, 0, 128, v->color);
			}
			if (IDC_VCO_SINE == res)
			{
				v->flags |= VCO_SINE;
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR_VAL), false);
			}
			if (IDC_VCO_TRISAW == res)
			{
				v->flags |= VCO_TRISAW;
			}
			if (IDC_VCO_PULSE == res)
			{
				v->flags |= VCO_PULSE;
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_SHAPE), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_SHAPE_VAL), false);
			}
			if (IDC_VCO_NOISE == res)
			{
				v->flags |= VCO_NOISE;
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_TRANSPOSE), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_TRANSPOSE_VAL), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_DETUNE), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_DETUNE_VAL), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_PHASE), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_PHASE_VAL), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR_VAL), false);
			}			
		}
		if (LOWORD(id) == IDC_VCO_LFO)
		{
			if (SendDlgItemMessage(ModuleWnd[M_VCO], IDC_VCO_LFO, BM_GETCHECK, 0, 0)==BST_CHECKED) 
				v->flags |= VCO_LFO;
			else
				v->flags &= ~VCO_LFO;
		}
		// gate bits changed?
		WORD gatebits;
		GetCheckboxGroupBitmask(IDC_VCO_GATE1, IDC_VCO_GATE16, LOWORD(id), M_VCO, gatebits, res);
		if (res)
		{
			v->gate = gatebits & 0xff;
			v->color = (gatebits >> 8) & 0xff;
			// adjust color slider
			InitSliderCenter(ModuleWnd[M_VCO], IDC_VCO_COLOR, 0, 128, v->color);
		}
	}
	else if (uid == M_VCF)
	{
		VCF_valP v = (VCF_valP)val;
		MaybeButtonGroupChanged(IDC_VCF_LOW, IDC_VCF_ALL, LOWORD(id), M_VCF, res);
		if (res)
		{
			if (res == IDC_VCF_LOW)
				v->type = VCF_LOWPASS;
			else if (res == IDC_VCF_HIGH)
				v->type = VCF_HIGHPASS;
			else if (res == IDC_VCF_BAND)
				v->type = VCF_BANDPASS;
			else if (res == IDC_VCF_NOTCH)
				v->type = VCF_BANDSTOP;
			else if (res == IDC_VCF_PEAK)
				v->type = VCF_PEAK;
			else if (res == IDC_VCF_ALL)
				v->type = VCF_ALLPASS;
		}
	}
	else if (uid == M_DLL)
	{
		DLL_valP v = (DLL_valP)val;
		MaybeButtonGroupChanged(IDC_DLL_DELAY, IDC_DLL_REVERB, LOWORD(id), M_DLL, res);
		if (res)
		{
			if (res == IDC_DLL_DELAY)
			{
				v->reverb = 0;
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_LEFTREVERB), SW_HIDE);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), SW_SHOW);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), SW_SHOW);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_NOSYNC), SW_SHOW);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_BPMSYNC), SW_SHOW);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_NOTESYNC), SW_SHOW);
			}
			else if (res == IDC_DLL_REVERB)
			{
				v->reverb = 1;
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_LEFTREVERB), SW_SHOW);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), SW_HIDE);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), SW_HIDE);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_NOSYNC), SW_HIDE);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_BPMSYNC), SW_HIDE);
				ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_NOTESYNC), SW_HIDE);
			}
		}
		MaybeButtonGroupChanged(IDC_DLL_NOSYNC, IDC_DLL_NOTESYNC, LOWORD(id), M_DLL, res);
		if (res)
		{
			if (res == IDC_DLL_NOSYNC)
			{
				v->synctype = 0;
				EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), true);
				EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), true);
			}
			else if (res == IDC_DLL_BPMSYNC)
			{
				v->synctype = 1;
				EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), true);
				EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), true);
			}
			else if (res == IDC_DLL_NOTESYNC)
			{
				v->synctype = 2;
				EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), false);
				EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), false);
			}
		}
		if (LOWORD(id) == IDC_DLL_LEFTREVERB)
		{			
			if (SendDlgItemMessage(ModuleWnd[M_DLL], IDC_DLL_LEFTREVERB, BM_GETCHECK, 0, 0)==BST_CHECKED) 
				v->leftreverb = 1;
			else
				v->leftreverb = 0;
		}
		UpdateDelayTimes(v);
	} // if (uid == M_DLL)
	else if (uid == M_FOP)
	{
		// invalidate stack first
		if (SelectedTab == T_INSTRUMENT)
		{
			SynthObjP->InstrumentSignalValid[SelectedInstrument] = 0;
		}
		else
		{
			SynthObjP->GlobalSignalValid = 0;
		}
		FOP_valP v = (FOP_valP)val;
		MaybeButtonGroupChanged(IDC_FOP_POP, IDC_FOP_LOADNOTE, LOWORD(id), M_FOP, res);
		if (res)
		{
			v->flags = 1 + res - IDC_FOP_POP;
		}		
		// update signalcount
		UpdateSignalCount(SelectedInstrument);
	}
	else if (uid == M_FSTx)
	{
		FSTx_valP v = (FSTx_valP)val;
		// changed instrument
		if (LOWORD(id) == IDC_FST_DESTINATION_INSTRUMENT)
		{
			int inst = SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0) - 1;
			if (v->dest_stack != inst)
			{
				v->dest_stack = inst;

				// fill unit combobox with current units and set to nothing
				BYTE *units;
				if (SelectedTab == T_INSTRUMENT)
				{
					if (v->dest_stack == -1)
						units = SynthObjP->InstrumentValues[SelectedInstrument][0];
					else if (v->dest_stack == MAX_INSTRUMENTS)
						units = SynthObjP->GlobalValues[0];
					else
						units = SynthObjP->InstrumentValues[v->dest_stack][0];
				}
				else
				{
					if (v->dest_stack == -1)
						units = SynthObjP->GlobalValues[0];
					else if (v->dest_stack == MAX_INSTRUMENTS)
						units = SynthObjP->GlobalValues[0];
					else
						units = SynthObjP->InstrumentValues[v->dest_stack][0];
				}

				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_UNIT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
				for (int i = 0; i < MAX_UNITS; i++)
				{
					char unitname[128], unitname2[128];
					sprintf(unitname, "%d: %s", i+1, GetUnitString(&units[i*MAX_SLOT_VALUES], unitname2));
					SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)unitname);
				}
				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_UNIT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
				v->dest_unit = -1;

				// set slot combobox to nothing and disable
				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
				EnableWindow(GetDlgItem(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT), false);
				v->dest_slot = -1;
				v->dest_id = -1;
			}
		}
		else if (LOWORD(id) == IDC_FST_DESTINATION_UNIT)
		{
			int unit = SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_UNIT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0) - 1;
			if (v->dest_unit != unit)
			{
				v->dest_unit = unit;

				// fill slot combobox with current targets and set to nothing
				BYTE *slots;
				if (SelectedTab == T_INSTRUMENT)
				{
					if (v->dest_stack == -1)
						slots = SynthObjP->InstrumentValues[SelectedInstrument][unit];
					else if (v->dest_stack == MAX_INSTRUMENTS)
						slots = SynthObjP->GlobalValues[unit];
					else
						slots = SynthObjP->InstrumentValues[v->dest_stack][unit];
				}
				else
				{
					if (v->dest_stack == -1)
						slots = SynthObjP->GlobalValues[unit];
					else if (v->dest_stack == MAX_INSTRUMENTS)
						slots = SynthObjP->GlobalValues[unit];
					else
						slots = SynthObjP->InstrumentValues[v->dest_stack][unit];
				}

				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
				for (int i = 0; i < 8; i++)
				{
					SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)UnitModulationTargetNames[slots[0]][i]);
				}
				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
				v->dest_slot = -1;
				v->dest_id = -1;

				if (v->dest_unit == -1)
					EnableWindow(GetDlgItem(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT), false);
				else
					EnableWindow(GetDlgItem(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT), true);
			}
		}
		else if (LOWORD(id) == IDC_FST_DESTINATION_SLOT)
		{
			int slot = SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0) - 1;
			v->dest_slot = slot;

			// fill slot combobox with current targets and set to nothing
			BYTE *slots;
			if (SelectedTab == T_INSTRUMENT)
			{
				if (v->dest_stack == -1)
					slots = SynthObjP->InstrumentValues[SelectedInstrument][v->dest_unit];
				else if (v->dest_stack == MAX_INSTRUMENTS)
					slots = SynthObjP->GlobalValues[v->dest_unit];
				else
					slots = SynthObjP->InstrumentValues[v->dest_stack][v->dest_unit];
			}
			else
			{
				if (v->dest_stack == -1)
					slots = SynthObjP->GlobalValues[v->dest_unit];
				else if (v->dest_stack == MAX_INSTRUMENTS)
					slots = SynthObjP->GlobalValues[v->dest_unit];
				else
					slots = SynthObjP->InstrumentValues[v->dest_stack][v->dest_unit];
			}

			v->dest_id = slots[0];
		}
		MaybeButtonGroupChanged(IDC_FST_SET, IDC_FST_MUL, LOWORD(id), M_FSTx, res);
		if (res)
		{
			int pop = v->operation & FST_OPERATION_POP;
			if (res == IDC_FST_SET)
				v->operation = FST_OPERATION_SET;
			else if (res == IDC_FST_ADD)
				v->operation = FST_OPERATION_ADD;
			else if (res == IDC_FST_MUL)
				v->operation = FST_OPERATION_MUL;
			if (pop)
				v->operation |= FST_OPERATION_POP;
		}
		if (LOWORD(id) == IDC_FST_POP)
		{			
			if (SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_POP, BM_GETCHECK, 0, 0)==BST_CHECKED) 
				v->operation |= FST_OPERATION_POP;
			else
				v->operation &= ~FST_OPERATION_POP;
		}

        if ((LOWORD(id) >= IDC_FST_COND_ALWAYS) && (LOWORD(id) <= IDC_FST_COND_IF_SIGN_CHANGED))
		{ // one of the condition checkboxes            
            int always = (SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_ALWAYS, BM_GETCHECK, 0, 0)==BST_CHECKED);
            int smaller = (SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_IF_SMALLER, BM_GETCHECK, 0, 0)==BST_CHECKED);
            int larger = (SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_IF_LARGER, BM_GETCHECK, 0, 0)==BST_CHECKED);
            int signchanged = (SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_IF_SIGN_CHANGED, BM_GETCHECK, 0, 0)==BST_CHECKED);

            if ((LOWORD(id) == IDC_FST_COND_ALWAYS) && (always))
                smaller = larger = signchanged = 0;
            else if ((LOWORD(id) == IDC_FST_COND_IF_LARGER) && (larger))
                smaller = 0;
            else if ((LOWORD(id) == IDC_FST_COND_IF_SMALLER) && (smaller))
                larger = 0;

            if (!smaller && !larger && !signchanged)
                always = 1;
            else if (smaller || larger || signchanged)
                always = 0;

            SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_ALWAYS, BM_SETCHECK, always, 0);
            SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_IF_SMALLER, BM_SETCHECK, smaller, 0);
            SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_IF_LARGER, BM_SETCHECK, larger, 0);
            SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_IF_SIGN_CHANGED, BM_SETCHECK, signchanged, 0);

            v->condition = FST_CONDITION_ALWAYS;
            if (smaller) v->condition |= FST_CONDITION_IF_SMALLER;
            if (larger) v->condition |= FST_CONDITION_IF_LARGER;
            if (signchanged) v->condition |= FST_CONDITION_IF_SIGN_CHANGED;
		}


		// update signalcount
		UpdateSignalCount(SelectedInstrument);
	} // uid == M_FSTx
	else if (uid == M_ACC)
	{
		ACC_valP v = (ACC_valP)val;
		MaybeButtonGroupChanged(IDC_ACC_OUT, IDC_ACC_AUX, LOWORD(id), M_ACC, res);
		if (res)
		{
			if (res == IDC_ACC_OUT)
				v->flags = ACC_OUT;
			else
				v->flags = ACC_AUX;
		}
	}
	if (uid == M_SLW)
    {
        SLW_valP v = (SLW_valP)val;

        // the MaybeButtonGroupChanged macro enables and disables buttons
        MaybeButtonGroupChanged(IDC_SLW_NOSYNC_MS, IDC_SLW_NOTESYNC, LOWORD(id), M_SLW, res);
        if (res)
        {
            if (res == IDC_SLW_NOSYNC_MS)
			{
                v->synctype = SLW_SYNCTYPE_ABSOLUTE_MS;                
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME), true); // slider
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_VAL), true); // value text
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_SMP), true); // value in samples text
            }
            else if (res == IDC_SLW_NOSYNC_SMP)
			{
                v->synctype = SLW_SYNCTYPE_ABSOLUTE_SMP;
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME), true); // slider
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_VAL), true); // value text
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_SMP), true); // value in samples text
            }
            else if (res == IDC_SLW_BPMSYNC)
            {
                v->synctype = SLW_SYNCTYPE_BPM;
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME), true); // slider
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_VAL), true); // value text
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_SMP), true); // value in samples text
            }
            else if (res == IDC_SLW_NOTESYNC)
            {
                v->synctype = SLW_SYNCTYPE_NOTE_PITCH;
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME), false); // slider
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_VAL), false); // value text
                EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_SMP), false); // value in samples text
            }            
        }
        // the MaybeButtonGroupChanged macro enables and disables buttons
        MaybeButtonGroupChanged(IDC_SLW_CALCMODE_LINEAR, IDC_SLW_CALCMODE_RMS, LOWORD(id), M_SLW, res);
        if (res)
        { // linear/RMS calcmode setting changed
            if (res == IDC_SLW_CALCMODE_LINEAR)
            {
                v->calcmode = SLW_CALCMODE_LINEAR;
            }
            else if (res == IDC_SLW_CALCMODE_RMS)
            {
                v->calcmode = SLW_CALCMODE_RMS;
            }

        }

        UpdateSlewTimes(v);
    } // if (uid == M_SLW)
	else if (uid == M_FLDFROM)
	{
		FLDFROM_valP v = (FLDFROM_valP)val;
		// changed instrument
		if (LOWORD(id) == IDC_FLDFROM_SOURCE_INSTRUMENT)
		{
			int inst = SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_INSTRUMENT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0) - 1;
			if (v->source_stack != inst)
			{
				v->source_stack = inst;

				// fill unit combobox with current units and set to nothing
				BYTE *units;
				if (SelectedTab == T_INSTRUMENT)
				{
					if (v->source_stack == -1)
						units = SynthObjP->InstrumentValues[SelectedInstrument][0];
					else if (v->source_stack == MAX_INSTRUMENTS)
						units = SynthObjP->GlobalValues[0];
					else
						units = SynthObjP->InstrumentValues[v->source_stack][0];
				}
				else
				{
					if (v->source_stack == -1)
						units = SynthObjP->GlobalValues[0];
					else if (v->source_stack == MAX_INSTRUMENTS)
						units = SynthObjP->GlobalValues[0];
					else
						units = SynthObjP->InstrumentValues[v->source_stack][0];
				}

				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_UNIT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
				for (int i = 0; i < MAX_UNITS; i++)
				{
					char unitname[128], unitname2[128];
					sprintf(unitname, "%d: %s", i+1, GetUnitString(&units[i*MAX_SLOT_VALUES], unitname2));
					SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)unitname);
				}
				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_UNIT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
				v->source_unit = -1;

				// set slot combobox to nothing and disable
				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
				EnableWindow(GetDlgItem(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT), false);
				v->source_slot = -1;
				v->source_id = -1;
			}
		}
		else if (LOWORD(id) == IDC_FLDFROM_SOURCE_UNIT)
		{
			int unit = SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_UNIT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0) - 1;
			if (v->source_unit != unit)
			{
				v->source_unit = unit;

				// fill slot combobox with current targets and set to nothing
				BYTE *slots;
				if (SelectedTab == T_INSTRUMENT)
				{
					if (v->source_stack == -1)
						slots = SynthObjP->InstrumentValues[SelectedInstrument][unit];
					else if (v->source_stack == MAX_INSTRUMENTS)
						slots = SynthObjP->GlobalValues[unit];
					else
						slots = SynthObjP->InstrumentValues[v->source_stack][unit];
				}
				else
				{
					if (v->source_stack == -1)
						slots = SynthObjP->GlobalValues[unit];
					else if (v->source_stack == MAX_INSTRUMENTS)
						slots = SynthObjP->GlobalValues[unit];
					else
						slots = SynthObjP->InstrumentValues[v->source_stack][unit];
				}

				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
				for (int i = 0; i < 8; i++)
				{
					SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)UnitModulationTargetNames[slots[0]][i]);
				}
				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
				v->source_slot = -1;
				v->source_id = -1;

				if (v->source_unit == -1)
					EnableWindow(GetDlgItem(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT), false);
				else
					EnableWindow(GetDlgItem(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT), true);
			}
		}
		else if (LOWORD(id) == IDC_FLDFROM_SOURCE_SLOT)
		{
			int slot = SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0) - 1;
			v->source_slot = slot;

			// fill slot combobox with current targets and set to nothing
			BYTE *slots;
			if (SelectedTab == T_INSTRUMENT)
			{
				if (v->source_stack == -1)
					slots = SynthObjP->InstrumentValues[SelectedInstrument][v->source_unit];
				else if (v->source_stack == MAX_INSTRUMENTS)
					slots = SynthObjP->GlobalValues[v->source_unit];
				else
					slots = SynthObjP->InstrumentValues[v->source_stack][v->source_unit];
			}
			else
			{
				if (v->source_stack == -1)
					slots = SynthObjP->GlobalValues[v->source_unit];
				else if (v->source_stack == MAX_INSTRUMENTS)
					slots = SynthObjP->GlobalValues[v->source_unit];
				else
					slots = SynthObjP->InstrumentValues[v->source_stack][v->source_unit];
			}

			v->source_id = slots[0];
		}
		// update signalcount
		UpdateSignalCount(SelectedInstrument);
	} // uid == M_FLDFROM
} // SetButtonParams()

bool ButtonPressed(WPARAM id, LPARAM lParam)
{
	// instrument tab is active
	if (SelectedTab == T_INSTRUMENT)
	{
		UnitID uid = (UnitID)(SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit][0]);
		SetButtonParams(uid, SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit], id, lParam);
		char unitname[128];
		GetUnitString(SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit], unitname);
		SetWindowText(GetDlgItem(TabWnd[T_INSTRUMENT], IDC_ISTACK_UNIT1+SelectedIUnit*6), unitname);
	}
	else if (SelectedTab == T_GLOBAL)
	{		
		UnitID uid = (UnitID)(SynthObjP->GlobalValues[SelectedGUnit][0]);
		SetButtonParams(uid, SynthObjP->GlobalValues[SelectedGUnit], id, lParam);
		char unitname[128];
		GetUnitString(SynthObjP->GlobalValues[SelectedGUnit], unitname);
		SetWindowText(GetDlgItem(TabWnd[T_GLOBAL], IDC_GSTACK_UNIT1+SelectedGUnit*6), unitname);
	}
	return TRUE;
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//																		//
//                        Scrollbar Events								//
//																		//
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void SetSliderParams(int uid, BYTE* val, LPARAM lParam)
{
	if (uid == M_ENV)
	{
		ENV_valP v = (ENV_valP)val;
		// attack
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_ENV_ATTACK))
		{
			UpdateSliderValue(IDC_ENV_ATTACK, v->attac);
		}
		//decay
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_ENV_DECAY))
		{
			UpdateSliderValue(IDC_ENV_DECAY, v->decay);
		}
		// sustain
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_ENV_SUSTAIN))
		{
			UpdateSliderValue(IDC_ENV_SUSTAIN, v->sustain);
		}
		// release
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_ENV_RELEASE))
		{
			UpdateSliderValue(IDC_ENV_RELEASE, v->release);
		}
		// gain
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_ENV_GAIN))
		{
			UpdateSliderValue(IDC_ENV_GAIN, v->gain);
		}
	}
	else if (uid == M_VCO)
	{
		VCO_valP v = (VCO_valP)val;
		// transpose
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_VCO_TRANSPOSE))
		{
			UpdateSliderValueCenter(IDC_VCO_TRANSPOSE, v->transpose);
		}
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_VCO_DETUNE))
		{
			UpdateSliderValueCenter(IDC_VCO_DETUNE, v->detune);
		}
		// phaseofs
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_VCO_PHASE))
		{
			UpdateSliderValue(IDC_VCO_PHASE, v->phaseofs);
		}
		// color
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_VCO_COLOR))
		{
			UpdateSliderValueCenter(IDC_VCO_COLOR, v->color);
			// gate bits
			WORD gatebits = ((WORD)v->color << 8) | (WORD)v->gate ;
			SetCheckboxGroupBitmask(IDC_VCO_GATE1, IDC_VCO_GATE16, M_VCO, gatebits);
		}
		// shape
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_VCO_SHAPE))
		{
			v->shape = SendMessage(GetDlgItem(ModuleWnd[uid], IDC_VCO_SHAPE), TBM_GETPOS, 0, 0);
			if (v->shape == 0)
				v->shape = 1;
			if (v->shape == 128)
				v->shape = 127;
			sprintf(SliderValTxt, "%d", v->shape-64);
			SetWindowText(GetDlgItem(ModuleWnd[uid], IDC_VCO_SHAPE_VAL), SliderValTxt);
		}
		// gain
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_VCO_GAIN))
		{
			UpdateSliderValue(IDC_VCO_GAIN, v->gain);
		}
	}
	else if (uid == M_VCF)
	{
		VCF_valP v = (VCF_valP)val;
		// frequency
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_VCF_FREQUENCY))
		{
			UpdateSliderValue(IDC_VCF_FREQUENCY, v->freq);
		}
		// resonance
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_VCF_RESONANCE))
		{
			UpdateSliderValue(IDC_VCF_RESONANCE, v->res);
		}
	}
	else if (uid == M_DST)
	{
		DST_valP v = (DST_valP)val;
		// drive
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_DST_DRIVE))
		{
			v->drive = SendMessage(GetDlgItem(ModuleWnd[uid], IDC_DST_DRIVE), TBM_GETPOS, 0, 0);
			if (v->drive == 0)
				v->drive = 1;
			if (v->drive == 128)
				v->drive = 127;
			sprintf(SliderValTxt, "%d", v->drive-64);
			SetWindowText(GetDlgItem(ModuleWnd[uid], IDC_DST_DRIVE_VAL), SliderValTxt);
		}		
		// snhfreq
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_DST_SNH))
		{
			UpdateSliderValue(IDC_DST_SNH, v->snhfreq);
		}	
	}
	else if (uid == M_DLL)
	{
		DLL_valP v = (DLL_valP)val;
		// pregain
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_DLL_PREGAIN))
		{
			UpdateSliderValue(IDC_DLL_PREGAIN, v->pregain);
		}
		// dry
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_DLL_DRY))
		{
			UpdateSliderValue(IDC_DLL_DRY, v->dry);
		}		
		// feedback
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_DLL_FEEDBACK))
		{
			UpdateSliderValue(IDC_DLL_FEEDBACK, v->feedback);
		}
		// frequency
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_DLL_FREQUENCY))
		{
			UpdateSliderValue(IDC_DLL_FREQUENCY, v->freq);
		}
		// depth
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_DLL_DEPTH))
		{
			UpdateSliderValue(IDC_DLL_DEPTH, v->depth);
		}
		// damp
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_DLL_DAMP))
		{
			UpdateSliderValue(IDC_DLL_DAMP, v->damp);
		}
		// delay
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_DLL_DTIME))
		{
			v->guidelay = SendMessage(GetDlgItem(ModuleWnd[uid], IDC_DLL_DTIME), TBM_GETPOS, 0, 0);
			UpdateDelayTimes(v);
		}
	}
	else if (uid == M_FSTx)
	{
		FSTx_valP v = (FSTx_valP)val;
		// gain
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_FST_AMOUNT))
		{
			UpdateSliderValueCenter2(IDC_FST_AMOUNT, v->amount);
		}
	}
	else if (uid == M_PAN)
	{
		PAN_valP v = (PAN_valP)val;
		// panning
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_PAN_PANNING))
		{
			UpdateSliderValueCenter(IDC_PAN_PANNING, v->panning);
		}
	}
	else if (uid == M_OUT)
	{
		OUT_valP v = (OUT_valP)val;
		// gain
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_OUT_GAIN))
		{
			UpdateSliderValue(IDC_OUT_GAIN, v->gain);
		}		
		// auxsend
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_OUT_AUXSEND))
		{
			UpdateSliderValue(IDC_OUT_AUXSEND, v->auxsend);
		}
	}
	else if (uid == M_FLD)
	{
		FLD_valP v = (FLD_valP)val;
		// panning
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_FLD_VALUE))
		{
			UpdateSliderValueCenter(IDC_FLD_VALUE, v->value);
		}
	}
	else if (uid == M_SLW)
	{ // aaa todo
		SLW_valP v = (SLW_valP)val;

		// delay time
		if ((HWND)lParam == GetDlgItem(ModuleWnd[uid], IDC_SLW_DTIME))
		{
			v->guidelay = SendMessage(GetDlgItem(ModuleWnd[uid], IDC_SLW_DTIME), TBM_GETPOS, 0, 0);

			UpdateSlewTimes(v);
		}
	} // if (uid == M_SLW)

	else if (uid == M_FLDFROM)
	{
		FLDFROM_valP v = (FLDFROM_valP)val;
		// FLDFROM doesn't have any sliders
	} // if (uid == M_FLDFROM)

} // SetSliderParams()

bool ScrollbarChanged(HWND hwndDlg, WPARAM wParam, LPARAM lParam)
{
	// instrument tab is active
	if (SelectedTab == T_INSTRUMENT)
	{
		UnitID uid = (UnitID)(SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit][0]);
		SetSliderParams(uid, SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit], lParam);
		char unitname[128];
		GetUnitString(SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit], unitname);
		SetWindowText(GetDlgItem(TabWnd[T_INSTRUMENT], IDC_ISTACK_UNIT1+SelectedIUnit*6), unitname);
	}
	else if (SelectedTab == T_GLOBAL)
	{		
		UnitID uid = (UnitID)(SynthObjP->GlobalValues[SelectedGUnit][0]);
		SetSliderParams(uid, SynthObjP->GlobalValues[SelectedGUnit], lParam);
		char unitname[128];
		GetUnitString(SynthObjP->GlobalValues[SelectedGUnit], unitname);
		SetWindowText(GetDlgItem(TabWnd[T_GLOBAL], IDC_GSTACK_UNIT1+SelectedGUnit*6), unitname);
	}
	return TRUE;
} // ScrollbarChanged()

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//																		//
//                       Stack Button Events							//
//																		//
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////





void CheckModulations(int unit1, int unit2 = -1)
{
	// a instrument stack changed
	if (SelectedTab == T_INSTRUMENT)
	{
		// look in all instruments if a store unit had its target on one of the changed units
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			for (int u = 0; u < MAX_UNITS; u++)
			{
                SOME_UNIT_valP v = (SOME_UNIT_valP)(SynthObjP->InstrumentValues[i][u]);
				//if (SynthObjP->InstrumentValues[i][u][0] == M_FSTx)
                if (IsModulationLinkUnit(v))
				{
					//FSTx_valP v = (FSTx_valP)(SynthObjP->InstrumentValues[i][u]);

					// the store points to global, so skip
					//if (v->dest_stack == MAX_INSTRUMENTS)
                    if (GetModulationLinkInstrument(v) == MAX_INSTRUMENTS)
						continue;
					
					int target_inst = GetModulationLinkInstrument(v);
                    if (target_inst == -1)
						target_inst = i;
					//if (v->dest_stack == -1) 
					//	target_inst = i;
					//else
					//	target_inst = v->dest_stack;

					// if the store points to the current instrument
					if (target_inst == SelectedInstrument)
					{
						// a up/down process 
						if (unit2 != -1)
						{
							// if the target unit was unit1 or unit2 realign target unit
							//if (v->dest_unit == unit1)
                            if (GetModulationLinkUnit(v) == unit1)
							{
								// v->dest_unit = unit2;
                                SetModulationLinkUnit(v, unit2);
							}
							//else if (v->dest_unit == unit2)
                            else if (GetModulationLinkUnit(v) == unit2)
							{
								//v->dest_unit = unit1;
                                SetModulationLinkUnit(v, unit1);
							}
						}
						// a set/reset process
						else
						{
							// if the target unit was unit1 reset store
							//if (v->dest_unit == unit1)
                            if (GetModulationLinkUnit(v) == unit1)
							{
								char text[128];
								sprintf(text, "Instrument %d had a modulation target in the changed unit\nIt's no longer valid and will be reset", i);
								MessageBox(DialogWnd, text, "Info", MB_OK);

//								v->dest_unit = -1;
//								v->dest_slot = -1;
//								v->dest_id = -1;
                                SetModulationLinkUnit(v, -1);                                
                                SetModulationLinkSlot(v, -1);
                                SetModulationLinkID(v, -1);
							}
						}
					}
				}
			}
		}
		// now check the global slots
		for (int u = 0; u < MAX_UNITS; u++)
		{
            SOME_UNIT_valP v = (SOME_UNIT_valP)(SynthObjP->GlobalValues[u]);
			//if (SynthObjP->GlobalValues[u][0] == M_FSTx)
            if (IsModulationLinkUnit(v))
			{
				//FSTx_valP v = (FSTx_valP)(SynthObjP->GlobalValues[u]);
				// the store points to global, so skip
				//if (v->dest_stack == -1 || v->dest_stack == MAX_INSTRUMENTS)
                if (GetModulationLinkInstrument(v) == -1 || GetModulationLinkInstrument(v) == MAX_INSTRUMENTS)
					continue;

				// if the store points to the current instrument
				//if (v->dest_stack == SelectedInstrument)
                if (GetModulationLinkInstrument(v) == SelectedInstrument)
				{
					// a up/down process 
					if (unit2 != -1)
					{
						// if the target unit was unit1 or unit2 realign target unit
						//if (v->dest_unit == unit1)
                        if (GetModulationLinkUnit(v) == unit1)
						{
							//v->dest_unit = unit2;
                            SetModulationLinkUnit(v, unit2);
						}
						//else if (v->dest_unit == unit2)
                        else if (GetModulationLinkUnit(v) == unit2)
						{
							//v->dest_unit = unit1;
                            SetModulationLinkUnit(v, unit1);
						}
					}
					// a set/reset process
					else
					{
						// if the target unit was unit1 reset store
						// if (v->dest_unit == unit1)
                        if (GetModulationLinkUnit(v) == unit1)
						{
							char text[128];
							sprintf(text, "Global had a modulation target in the changed unit\nIt's no longer valid and will be reset");
							MessageBox(DialogWnd, text, "Info", MB_OK);
							//v->dest_unit = -1;
							//v->dest_slot = -1;
							//v->dest_id = -1;
                            SetModulationLinkUnit(v, -1);
                            SetModulationLinkSlot(v, -1);
                            SetModulationLinkID(v, -1);
						}
					}
				}
			}
		}
	}
	// the global stack changed
	else
	{
		// look in all instruments if a store unit had its target on a global unit
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			for (int u = 0; u < MAX_UNITS; u++)
			{
				//if (SynthObjP->InstrumentValues[i][u][0] == M_FSTx)
                SOME_UNIT_valP v = (SOME_UNIT_valP)(SynthObjP->InstrumentValues[i][u]);
                if (IsModulationLinkUnit(v))
				{
					//FSTx_valP v = (FSTx_valP)(SynthObjP->InstrumentValues[i][u]);
					// the store doesnt point to global, so skip
					// if (v->dest_stack != MAX_INSTRUMENTS)
                    if (GetModulationLinkInstrument(v) != MAX_INSTRUMENTS)
						continue;

					// a up/down process 
					if (unit2 != -1)
					{
						// if the target unit was unit1 or unit2 realign target unit
						//if (v->dest_unit == unit1)
                        if (GetModulationLinkUnit(v) == unit1)
						{
							//v->dest_unit = unit2;
                            SetModulationLinkUnit(v, unit2);
						}
						//else if (v->dest_unit == unit2)
                        else if (GetModulationLinkUnit(v) == unit2)
						{
							//v->dest_unit = unit1;
                            SetModulationLinkUnit(v, unit1);
						}
					}
					// a set/reset process
					else
					{
						// if the target unit was unit1 reset store
						// if (v->dest_unit == unit1)
                        if (GetModulationLinkUnit(v) == unit1)
						{
							char text[128];
							sprintf(text, "Instrument %d had a modulation target in the changed unit\nIt's no longer valid and will be reset", i);
							MessageBox(DialogWnd, text, "Info", MB_OK);
							//v->dest_unit = -1;
							//v->dest_slot = -1;
							//v->dest_id = -1;
                            SetModulationLinkUnit(v, -1);
                            SetModulationLinkSlot(v, -1);
                            SetModulationLinkID(v, -1);
						}
					}
				}
			}
		}
		// now check the global slots
		for (int u = 0; u < MAX_UNITS; u++)
		{
            SOME_UNIT_valP v = (SOME_UNIT_valP)(SynthObjP->GlobalValues[u]);
			// if (SynthObjP->GlobalValues[u][0] == M_FSTx)
            if (IsModulationLinkUnit(v))
			{
				//FSTx_valP v = (FSTx_valP)(SynthObjP->GlobalValues[u]);
				// the store points to global, so check
				//if (v->dest_stack == -1 || v->dest_stack == MAX_INSTRUMENTS)
                if (GetModulationLinkInstrument(v) == -1 || GetModulationLinkInstrument(v) == MAX_INSTRUMENTS)
				{
					// a up/down process 
					if (unit2 != -1)
					{
						// if the target unit was unit1 or unit2 realign target unit
						//if (v->dest_unit == unit1)
						//{
							//v->dest_unit = unit2;
						//}
						//else if (v->dest_unit == unit2)
						//{
							//v->dest_unit = unit1;
						//}

						// if the target unit was unit1 or unit2 realign target unit
						//if (v->dest_unit == unit1)
                        if (GetModulationLinkUnit(v) == unit1)
						{
							//v->dest_unit = unit2;
                            SetModulationLinkUnit(v, unit2);
						}
						//else if (v->dest_unit == unit2)
                        else if (GetModulationLinkUnit(v) == unit2)
						{
							//v->dest_unit = unit1;
                            SetModulationLinkUnit(v, unit1);
						}


					}
					// a set/reset process
					else
					{
						// if the target unit was unit1 reset store
                        if (GetModulationLinkUnit(v) == unit1)
						{
							char text[128];
							sprintf(text, "Global had a modulation target in the changed unit\nIt's no longer valid and will be reset");
							MessageBox(DialogWnd, text, "Info", MB_OK);
							//v->dest_unit = -1;
							//v->dest_slot = -1;
							//v->dest_id = -1;
                            SetModulationLinkUnit(v, -1);
                            SetModulationLinkSlot(v, -1);
                            SetModulationLinkID(v, -1);
						}
					}
				}
			}
		}
	}
} // CheckModulations()

bool StackButtonPressed(WPARAM id, bool shift_pressed)
{
	int res;

	// Instrument Stack Buttons
	if (SelectedTab == T_INSTRUMENT)
	{
		// check unit click
		InterleavedButtonGroupChanged(IDC_ISTACK_UNIT1, MAX_UNITS, LOWORD(id), res);
		if (res >= 0)
		{
			SelectedIUnit = res;
			UpdateModuleParamWindow(T_INSTRUMENT, res);
		}
		
		// check down click
		for (int i = 0, found = 0; (i < MAX_UNITS) && !found; i++)
		{
			WORD tid = IDC_ISTACK_DOWN1+i*6;
			if (LOWORD(id) == tid)
			{
				SynthObjP->InstrumentSignalValid[SelectedInstrument] = 0;
				EnableWindow(GetDlgItem(TabWnd[T_INSTRUMENT], tid), false);

                // Down button is clicked while holding down Shift
                if (shift_pressed)
                {
                    // Find the first free slot below the one that was clicked
                    int first_free_slot = -1;
                    for (int j=i+1; (j<MAX_UNITS) && (first_free_slot < 0); j++)
                        if (SynthObjP->InstrumentValues[SelectedInstrument][j][0] == M_NONE)
                            first_free_slot = j;

                    // If a free slot was found, move it one slot at a time up to the clicked slot
                    if (first_free_slot >= 0)
                    {
                        for (int j=first_free_slot; j>i; j--)
                        {
                            Go4kVSTi_FlipInstrumentSlots(SelectedInstrument, j-1, j);
				            CheckModulations(j-1, j);
                        }
                    }                    
                }
                else
                { // Shift not held down: just flip two slots
				    Go4kVSTi_FlipInstrumentSlots(SelectedInstrument, i, i+1);
				    CheckModulations(i, i+1);
                }

				UpdateControls(SelectedInstrument);
				EnableWindow(GetDlgItem(TabWnd[T_INSTRUMENT], tid), true);

                found = 1;
			}			
		}

		// check up click
		for (int i = 0; i < MAX_UNITS; i++)
		{
			WORD tid = IDC_ISTACK_UP1+i*6;
			if (LOWORD(id) == tid)
			{
				SynthObjP->InstrumentSignalValid[SelectedInstrument] = 0;
				EnableWindow(GetDlgItem(TabWnd[T_INSTRUMENT], tid), false);

                // Up button is clicked while holding down Shift
                if (shift_pressed)
                {
                    // Find the first free slot above the one that was clicked
                    int first_free_slot = -1;
                    for (int j=i-1; (j>=0) && (first_free_slot < 0); j--)
                        if (SynthObjP->InstrumentValues[SelectedInstrument][j][0] == M_NONE)
                            first_free_slot = j;

                    // If a free slot was found, move it one slot at a time up to the clicked slot
                    if (first_free_slot >= 0)
                    {
                        for (int j=first_free_slot; j<i; j++)
                        {
                            Go4kVSTi_FlipInstrumentSlots(SelectedInstrument, j, j+1);
				            CheckModulations(j, j+1);
                        }
                    }                    
                }
                else
                { // Shift not held down: just flip two slots
				    Go4kVSTi_FlipInstrumentSlots(SelectedInstrument, i, i-1);
				    CheckModulations(i, i-1);
                }


				UpdateControls(SelectedInstrument);
				EnableWindow(GetDlgItem(TabWnd[T_INSTRUMENT], tid), true);
			}
		}

		// check set click
		for (int i = 0; i < MAX_UNITS; i++)
		{
			WORD tid = IDC_ISTACK_SET1+i*6;
			if (LOWORD(id) == tid)
			{
				EnableWindow(GetDlgItem(TabWnd[T_INSTRUMENT], tid), false);
				if (IDOK == DialogBox(hInstance, MAKEINTRESOURCE(IDD_SET_DIALOG), DialogWnd, (DLGPROC)SetDialogProc))
				{
					if (SetUnit != SynthObjP->InstrumentValues[SelectedInstrument][i][0])
					{
						SynthObjP->InstrumentSignalValid[SelectedInstrument] = 0;
						Go4kVSTi_InitInstrumentSlot(SelectedInstrument, i, SetUnit);
						CheckModulations(i);
						SelectedIUnit = i;
						UpdateControls(SelectedInstrument);
					}
				}
				EnableWindow(GetDlgItem(TabWnd[T_INSTRUMENT], tid), true);
				return TRUE;
			}
		}

		// check reset click
		for (int i = 0; i < MAX_UNITS; i++)
		{
			WORD tid = IDC_ISTACK_RESET1+i*6;
			if (LOWORD(id) == tid)
			{
				SynthObjP->InstrumentSignalValid[SelectedInstrument] = 0;
				EnableWindow(GetDlgItem(TabWnd[T_INSTRUMENT], tid), false);			
				Go4kVSTi_ClearInstrumentSlot(SelectedInstrument, i);
				CheckModulations(i);
				UpdateControls(SelectedInstrument);
				EnableWindow(GetDlgItem(TabWnd[T_INSTRUMENT], tid), true);
			}			
		}
	}
	else if (SelectedTab == T_GLOBAL)
	{
		// check unit click
		InterleavedButtonGroupChanged(IDC_GSTACK_UNIT1, MAX_UNITS, LOWORD(id), res);
		if (res >= 0)
		{
			SelectedGUnit = res;
			UpdateModuleParamWindow(T_GLOBAL, res);
		}

		// check down click
		for (int i = 0; i < MAX_UNITS; i++)
		{
			WORD tid = IDC_GSTACK_DOWN1+i*6;
			if (LOWORD(id) == tid)
			{
				SynthObjP->GlobalSignalValid = 0;
				EnableWindow(GetDlgItem(TabWnd[T_GLOBAL], tid), false);
				Go4kVSTi_FlipGlobalSlots(i, i+1);
				CheckModulations(i, i+1);
				UpdateControls(SelectedInstrument);
				EnableWindow(GetDlgItem(TabWnd[T_GLOBAL], tid), true);
			}
		}

		// check up click
		for (int i = 0; i < MAX_UNITS; i++)
		{
			WORD tid = IDC_GSTACK_UP1+i*6;
			if (LOWORD(id) == tid)
			{
				SynthObjP->GlobalSignalValid = 0;
				EnableWindow(GetDlgItem(TabWnd[T_GLOBAL], tid), false);
				Go4kVSTi_FlipGlobalSlots(i, i-1);
				CheckModulations(i, i-1);
				UpdateControls(SelectedInstrument);
				EnableWindow(GetDlgItem(TabWnd[T_GLOBAL], tid), true);
			}
		}

		// check set click
		for (int i = 0; i < MAX_UNITS; i++)
		{
			WORD tid = IDC_GSTACK_SET1+i*6;
			if (LOWORD(id) == tid)
			{
				EnableWindow(GetDlgItem(TabWnd[T_GLOBAL], tid), false);
				if (IDOK == DialogBox(hInstance, MAKEINTRESOURCE(IDD_SET_DIALOG), DialogWnd, (DLGPROC)SetDialogProc))
				{
					if (SetUnit != SynthObjP->GlobalValues[i][0])
					{
						SynthObjP->GlobalSignalValid = 0;
						Go4kVSTi_InitGlobalSlot(i, SetUnit);
						CheckModulations(i);
						SelectedGUnit = i;
						UpdateControls(SelectedInstrument);
					}
				}
				EnableWindow(GetDlgItem(TabWnd[T_GLOBAL], tid), true);
			}
		}

		// check reset click
		for (int i = 0; i < MAX_UNITS; i++)
		{
			WORD tid = IDC_GSTACK_RESET1+i*6;
			if (LOWORD(id) == tid)
			{
				SynthObjP->GlobalSignalValid = 0;
				EnableWindow(GetDlgItem(TabWnd[T_GLOBAL], tid), false);
				Go4kVSTi_ClearGlobalSlot(i);
				CheckModulations(i);
				UpdateControls(SelectedInstrument);
				EnableWindow(GetDlgItem(TabWnd[T_GLOBAL], tid), true);
			}
		}
	}
	return TRUE;
} // StackButtonPressed()

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//																		//
//                        TabChange Event 								//
//																		//
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void TabChanged(int index)
{
	for (int i = 0; i < NUM_TABS; i++)
	{
		if (i == index)
			ShowWindow(TabWnd[i], SW_SHOW);		
		else
			ShowWindow(TabWnd[i], SW_HIDE);			
	}
	if (index == T_INSTRUMENT)
	{
		ShowWindow(GetDlgItem(DialogWnd, IDC_ISTACK_VALID), true);
		ShowWindow(GetDlgItem(DialogWnd, IDC_GSTACK_VALID), false);
		// reset scroll pos
		int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
		SCROLLINFO sinfo;
		sinfo.cbSize = sizeof(SCROLLINFO);
		sinfo.fMask = SIF_POS;
		sinfo.nPos = InstrumentScrollPos[SelectedInstrument];
		SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);				
		ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);		
	}
	if (index == T_GLOBAL)
	{
		ShowWindow(GetDlgItem(DialogWnd, IDC_ISTACK_VALID), false);
		ShowWindow(GetDlgItem(DialogWnd, IDC_GSTACK_VALID), true);
		// reset scroll pos
		int oldpos = GetScrollPos(ScrollWnd, SB_VERT);
		SCROLLINFO sinfo;
		sinfo.cbSize = sizeof(SCROLLINFO);
		sinfo.fMask = SIF_POS;
		sinfo.nPos = GlobalScrollPos;
		SetScrollInfo(ScrollWnd, SB_VERT, &sinfo, true);
		ScrollWindow(ScrollWnd, 0, oldpos - sinfo.nPos, NULL, NULL);
	}

	SelectedTab = index;
	UpdateModuleParamWindow(SelectedTab, -1);
} // TabChanged()

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//																		//
//                        GUI Update Events								//
//																		//
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void UpdateSignalCount(int channel)
{
	//////////////////////////////////////////////////////////////
	//  instrument
	//////////////////////////////////////////////////////////////

	// disable for safety first
	SynthObjP->InstrumentSignalValid[channel] = 0;
	int valid = 1;
	int signalcount = 0;
	// check stack validity
	for (int i = 0; i < MAX_UNITS; i++)
	{
		// check unit precondition
		int precond = UnitPreSignals[SynthObjP->InstrumentValues[channel][i][0]];
		// adjust for arithmetic unit (depending on mode)
		if (SynthObjP->InstrumentValues[channel][i][0] == M_FOP)
		{
			FOP_valP v = (FOP_valP)(SynthObjP->InstrumentValues[channel][i]);
			if (v->flags == FOP_LOADNOTE)
				precond = 0;
			if (v->flags == FOP_POP)
				precond = 1;
			if (v->flags == FOP_PUSH)
				precond = 1;
			if (v->flags == FOP_ADDP2)
				precond = 4;
		}
		if (signalcount < precond)
		{
			valid = 0;
			Go4kVSTi_ClearInstrumentWorkspace(channel);
			for (int j = i; j < MAX_UNITS; j++)
			{
				char txt[10];
				sprintf(txt, "%d", signalcount);
				SetWindowText(GetDlgItem(TabWnd[T_INSTRUMENT], IDC_ISTACK_SIGNALCOUNT1+j*6), "inv");
			}
			char err[128];
			sprintf(err, "Signal INVALID! Unit %d needs >= %d signals", i+1, precond);
			SetWindowText(GetDlgItem(DialogWnd, IDC_ISTACK_VALID), err);
			break;
		}
		// precondition ok
		else
		{
			signalcount += UnitPostSignals[SynthObjP->InstrumentValues[channel][i][0]];
			// adjust for arithmetic unit (depending on mode)
			if (SynthObjP->InstrumentValues[channel][i][0] == M_FOP)
			{
				FOP_valP v = (FOP_valP)(SynthObjP->InstrumentValues[channel][i]);
				if (v->flags == FOP_POP)
					signalcount--;
				if (v->flags == FOP_PUSH)
					signalcount++;
				if (v->flags == FOP_ADDP)
					signalcount--;
				if (v->flags == FOP_MULP)
					signalcount--;
				if (v->flags == FOP_ADDP2)
					signalcount-=2;
				if (v->flags == FOP_LOADNOTE)
					signalcount++;
			}
			// check stack undeflow
			if (signalcount < 0)
			{
				valid = 0;
				Go4kVSTi_ClearInstrumentWorkspace(channel);
				for (int j = i; j < MAX_UNITS; j++)
				{
					char txt[10];
					sprintf(txt, "%d", signalcount);
					SetWindowText(GetDlgItem(TabWnd[T_INSTRUMENT], IDC_ISTACK_SIGNALCOUNT1+j*6), "inv");
				}
				char err[128];
				sprintf(err, "Signal INVALID! Unit %d Stack Underflow %d", i+1, signalcount);
				SetWindowText(GetDlgItem(DialogWnd, IDC_ISTACK_VALID), err);
				break;
			}
			// check stack overflow
			else if (signalcount > 8)
			{
				valid = 0;
				Go4kVSTi_ClearInstrumentWorkspace(channel);
				for (int j = i; j < MAX_UNITS; j++)
				{
					char txt[10];
					sprintf(txt, "%d", signalcount);
					SetWindowText(GetDlgItem(TabWnd[T_INSTRUMENT], IDC_ISTACK_SIGNALCOUNT1+j*6), "inv");
				}
				char err[128];
				sprintf(err, "Signal INVALID! Unit %d Stack Overflow %d", i+1, signalcount);
				SetWindowText(GetDlgItem(DialogWnd, IDC_ISTACK_VALID), err);
				break;
			}
			// TODO: check module internal overflow (like needing additional FP slots internally)
			else 
			{
			}
			// update signal counter
			char txt[10];
			sprintf(txt, "%d", signalcount);
			SetWindowText(GetDlgItem(TabWnd[T_INSTRUMENT], IDC_ISTACK_SIGNALCOUNT1+i*6), txt);
		}
	}
	if (valid)
	{
		if (signalcount == 0)
		{
			SynthObjP->InstrumentSignalValid[channel] = 1;
			SetWindowText(GetDlgItem(DialogWnd, IDC_ISTACK_VALID), "Signal OK");
		}
		else
		{
			Go4kVSTi_ClearInstrumentWorkspace(channel);
			SetWindowText(GetDlgItem(DialogWnd, IDC_ISTACK_VALID), "Signal INVALID! Signal Count != 0 after last unit");
		}
	}

	//////////////////////////////////////////////////////////////
	//  global
	//////////////////////////////////////////////////////////////

	// disable for safety first
	SynthObjP->GlobalSignalValid = 0;
	valid = 1;
	signalcount = 0;
	// check stack validity
	for (int i = 0; i < MAX_UNITS; i++)
	{
		// check unit precondition
		int precond = UnitPreSignals[SynthObjP->GlobalValues[i][0]];
		// adjust for arithmetic unit (depending on mode)
		if (SynthObjP->GlobalValues[i][0] == M_FOP)
		{
			FOP_valP v = (FOP_valP)(SynthObjP->GlobalValues[i]);
			if (v->flags == FOP_LOADNOTE)
				precond = 0;
			if (v->flags == FOP_POP)
				precond = 1;
			if (v->flags == FOP_PUSH)
				precond = 1;
			if (v->flags == FOP_ADDP2)
				precond = 4;
		}
		if (signalcount < precond)
		{
			valid = 0;
			Go4kVSTi_ClearGlobalWorkspace();
			for (int j = i; j < MAX_UNITS; j++)
			{
				char txt[10];
				sprintf(txt, "%d", signalcount);
				SetWindowText(GetDlgItem(TabWnd[T_GLOBAL], IDC_GSTACK_SIGNALCOUNT1+j*6), "inv");
			}
			char err[128];
			sprintf(err, "Signal INVALID! Unit %d needs >= %d signals", i+1, precond);
			SetWindowText(GetDlgItem(DialogWnd, IDC_GSTACK_VALID), err);
			break;
		}
		// precondition ok
		else
		{
			signalcount += UnitPostSignals[SynthObjP->GlobalValues[i][0]];
			// adjust for arithmetic unit (depending on mode)
			if (SynthObjP->GlobalValues[i][0] == M_FOP)
			{
				FOP_valP v = (FOP_valP)(SynthObjP->GlobalValues[i]);
				if (v->flags == FOP_POP)
					signalcount--;
				if (v->flags == FOP_PUSH)
					signalcount++;
				if (v->flags == FOP_ADDP)
					signalcount--;
				if (v->flags == FOP_MULP)
					signalcount--;
				if (v->flags == FOP_ADDP2)
					signalcount-=2;
				if (v->flags == FOP_LOADNOTE)
					signalcount++;
			}
			// check stack undeflow
			if (signalcount < 0)
			{
				valid = 0;
				Go4kVSTi_ClearGlobalWorkspace();
				for (int j = i; j < MAX_UNITS; j++)
				{
					char txt[10];
					sprintf(txt, "%d", signalcount);
					SetWindowText(GetDlgItem(TabWnd[T_GLOBAL], IDC_GSTACK_SIGNALCOUNT1+j*6), "inv");
				}
				char err[128];
				sprintf(err, "Signal INVALID! Unit %d Stack Underflow %d", i+1, signalcount);
				SetWindowText(GetDlgItem(DialogWnd, IDC_GSTACK_VALID), err);
				break;
			}
			// check stack overflow
			else if (signalcount > 8)
			{
				valid = 0;
				Go4kVSTi_ClearGlobalWorkspace();
				for (int j = i; j < MAX_UNITS; j++)
				{
					char txt[10];
					sprintf(txt, "%d", signalcount);
					SetWindowText(GetDlgItem(TabWnd[T_GLOBAL], IDC_GSTACK_SIGNALCOUNT1+j*6), "inv");
				}
				char err[128];
				sprintf(err, "Signal INVALID! Unit %d Stack Overflow %d", i+1, signalcount);
				SetWindowText(GetDlgItem(DialogWnd, IDC_GSTACK_VALID), err);
				break;
			}
			// TODO: check module internal overflow (like needing additional FP slots internally)
			else 
			{
			}
			// update signal counter
			char txt[10];
			sprintf(txt, "%d", signalcount);
			SetWindowText(GetDlgItem(TabWnd[T_GLOBAL], IDC_GSTACK_SIGNALCOUNT1+i*6), txt);
		}
	}
	if (valid)
	{
		if (signalcount == 0)
		{
			SynthObjP->GlobalSignalValid = 1;
			SetWindowText(GetDlgItem(DialogWnd, IDC_GSTACK_VALID), "Signal OK");
		}
		else
		{
			Go4kVSTi_ClearGlobalWorkspace();
			SetWindowText(GetDlgItem(DialogWnd, IDC_GSTACK_VALID), "Signal INVALID! Signal Count != 0 after last unit");
		}
	}
} // UpdateSignalCount()

static char MIDIChannelText[1234];

void UpdateControls(int instrument_index)
{
    if (!SynthObjP)
        return;

    int res;

	UpdateSignalCount(instrument_index);
	SetDlgItemText(DialogWnd, IDC_INSTRUMENT_NAME, (LPSTR)&(SynthObjP->InstrumentNames[instrument_index]));
	SendDlgItemMessage(DialogWnd, IDC_POLYPHONY, CB_SETCURSEL, (WPARAM)(SynthObjP->Polyphony-1), (LPARAM)0);
	for (int i = 0; i < MAX_UNITS; i++)
	{
		// set unit text
		char unitname[128];
		GetUnitString(SynthObjP->InstrumentValues[instrument_index][i], unitname);
		SetWindowText(GetDlgItem(TabWnd[T_INSTRUMENT], IDC_ISTACK_UNIT1+i*6), unitname);
		GetUnitString(SynthObjP->GlobalValues[i], unitname);
		SetWindowText(GetDlgItem(TabWnd[T_GLOBAL], IDC_GSTACK_UNIT1+i*6), unitname);
	}
	InterleavedButtonGroupChanged(IDC_ISTACK_UNIT1, MAX_UNITS, IDC_ISTACK_UNIT1+SelectedIUnit*6, res);
	InterleavedButtonGroupChanged(IDC_GSTACK_UNIT1, MAX_UNITS, IDC_GSTACK_UNIT1+SelectedGUnit*6, res);
	UpdateModuleParamWindow(SelectedTab, -1);

    sprintf(MIDIChannelText, "%d", instrument_index + 1);

//    SetDlgItemText(DialogWnd, IDC_INSTRUMENT_MIDI_CHANNEL, (LPSTR)(MIDIChannelText));
    static char temp[4];
    
    SetDlgItemText(DialogWnd, IDC_INSTRUMENT_VELOCITY_LO, itoa(SynthObjP->InstrumentInputSettings[instrument_index].VelocityLo, temp, 10));
    SetDlgItemText(DialogWnd, IDC_INSTRUMENT_VELOCITY_HI, itoa(SynthObjP->InstrumentInputSettings[instrument_index].VelocityHi, temp, 10));
    SetDlgItemText(DialogWnd, IDC_INSTRUMENT_NOTE_LO, itoa(SynthObjP->InstrumentInputSettings[instrument_index].NoteLo, temp, 10));
    SetDlgItemText(DialogWnd, IDC_INSTRUMENT_NOTE_HI, itoa(SynthObjP->InstrumentInputSettings[instrument_index].NoteHi, temp, 10));

    SendDlgItemMessage(DialogWnd, IDC_INSTRUMENT_MAP_VELOCITY_TO_NOTE, BM_SETCHECK, SynthObjP->InstrumentInputSettings[instrument_index].MapVelocityToNote, 0);

    UpdateInstrumentSelectionCombobox(DialogWnd);
    // SetDlgItemText(DialogWnd, IDC_INSTRUMENT, (LPSTR)"uwfgeuyegwf");
    
    SendDlgItemMessage(DialogWnd, IDC_SWING_STRENGTH, CB_SETCURSEL, (WPARAM)GlobalSetsP->SwingStrength - 50, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_PATTERN_QUANT, CB_SETCURSEL, (WPARAM)GlobalSetsP->NoteQuantization, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_PATTERN_SIZE, CB_SETCURSEL, ConvertPatternSizeToGUI((WPARAM)GlobalSetsP->PatternSize), (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_CLIPOUTPUT, BM_SETCHECK, (WPARAM)GlobalSetsP->ClipOutput, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_UNDENORMALIZE, BM_SETCHECK, (WPARAM)GlobalSetsP->Undenormalize, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_ENVLEVELS, BM_SETCHECK, (WPARAM)GlobalSetsP->RecordEnvs, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_NOTEVALUES, BM_SETCHECK, (WPARAM)GlobalSetsP->RecordNotes, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_16BIT, BM_SETCHECK, (WPARAM)GlobalSetsP->Output16Bit, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_OBJFORMAT, CB_SETCURSEL, (WPARAM)GlobalSetsP->ObjectFormat, (LPARAM)0);

    SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC, BM_SETCHECK, (WPARAM)GlobalSetsP->AutoPanicDoPanic, (LPARAM)0);

    SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_INVALID_OPERATION, BM_SETCHECK, (WPARAM)GlobalSetsP->AutoPanicMask & AUTO_PANIC_ON_INVALID_OPERATION, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_DENORMAL, BM_SETCHECK, (WPARAM)GlobalSetsP->AutoPanicMask & AUTO_PANIC_ON_DENORMAL, (LPARAM)0);

    SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_ZERO_DIVIDE, BM_SETCHECK, (WPARAM)GlobalSetsP->AutoPanicMask & AUTO_PANIC_ON_DIVIDE_BY_ZERO, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_OVERFLOW, BM_SETCHECK, (WPARAM)GlobalSetsP->AutoPanicMask & AUTO_PANIC_ON_OVERFLOW, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_UNDERFLOW, BM_SETCHECK, (WPARAM)GlobalSetsP->AutoPanicMask & AUTO_PANIC_ON_UNDERFLOW, (LPARAM)0);
    
    SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_PRECISION, BM_SETCHECK, (WPARAM)GlobalSetsP->AutoPanicMask & AUTO_PANIC_ON_PRECISION_LOSS, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_STACK_FAULT, BM_SETCHECK, (WPARAM)GlobalSetsP->AutoPanicMask & AUTO_PANIC_ON_STACK_FAULT, (LPARAM)0);
    SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_EXCEPTION, BM_SETCHECK, (WPARAM)GlobalSetsP->AutoPanicMask & AUTO_PANIC_ON_EXCEPTION_FLAG, (LPARAM)0);
} // UpdateControls()

void UpdateModule(int uid, BYTE* val)
{
	if (uid == M_ENV)
	{
		ENV_valP v = (ENV_valP)val;
		// attack
		InitSlider(ModuleWnd[uid], IDC_ENV_ATTACK, 0, 128, v->attac);
		// decay
		InitSlider(ModuleWnd[uid], IDC_ENV_DECAY, 0, 128, v->decay);
		// sustain
		InitSlider(ModuleWnd[uid], IDC_ENV_SUSTAIN, 0, 128, v->sustain);
		// release
		InitSlider(ModuleWnd[uid], IDC_ENV_RELEASE, 0, 128, v->release);
		// gain
		InitSlider(ModuleWnd[uid], IDC_ENV_GAIN, 0, 128, v->gain);
	}
	else if (uid == M_VCO)
	{
		VCO_valP v = (VCO_valP)val;
		// transpose
		InitSliderCenter(ModuleWnd[uid], IDC_VCO_TRANSPOSE, 0, 128, v->transpose);
		// detune
		InitSliderCenter(ModuleWnd[uid], IDC_VCO_DETUNE, 0, 128, v->detune);
		// phaseofs
		InitSlider(ModuleWnd[uid], IDC_VCO_PHASE, 0, 128, v->phaseofs);
		// color
		InitSliderCenter(ModuleWnd[uid], IDC_VCO_COLOR, 0, 128, v->color);
		// shape
		InitSliderCenter(ModuleWnd[uid], IDC_VCO_SHAPE, 0, 128, v->shape);
		// gain
		InitSlider(ModuleWnd[uid], IDC_VCO_GAIN, 0, 128, v->gain);
		
		// buttons
		DisableButtonGroup(IDC_VCO_SINE, IDC_VCO_GATE, M_VCO);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_TRANSPOSE), true);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_TRANSPOSE_VAL), true);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_DETUNE), true);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_DETUNE_VAL), true);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_PHASE), true);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_PHASE_VAL), true);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR), true);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR_VAL), true);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_SHAPE), true);
		EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_SHAPE_VAL), true);
		EnableButtonGroup(IDC_VCO_GATE1, IDC_VCO_GATE16, M_VCO, false);
		if (v->flags & VCO_SINE)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_SINE), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR_VAL), false);
		}
		else if (v->flags & VCO_TRISAW)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_TRISAW), false);
		}
		else if (v->flags & VCO_PULSE)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_PULSE), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_SHAPE), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_SHAPE_VAL), false);
		}
		else if (v->flags & VCO_NOISE)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_NOISE), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_TRANSPOSE), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_TRANSPOSE_VAL), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_DETUNE), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_DETUNE_VAL), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_PHASE), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_PHASE_VAL), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR_VAL), false);
		}
		else if (v->flags & VCO_GATE)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_GATE), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_VCO], IDC_VCO_COLOR_VAL), false);
			EnableButtonGroup(IDC_VCO_GATE1, IDC_VCO_GATE16, M_VCO, true);
		}

		// gate bits
		WORD gatebits = ((WORD)v->color << 8) | (WORD)v->gate;
		SetCheckboxGroupBitmask(IDC_VCO_GATE1, IDC_VCO_GATE16, M_VCO, gatebits);
		SendDlgItemMessage(ModuleWnd[M_VCO], IDC_VCO_LFO, BM_SETCHECK, v->flags & VCO_LFO, 0);	
	}
	else if (uid == M_VCF)
	{
		VCF_valP v = (VCF_valP)val;
		// frequency
		InitSlider(ModuleWnd[uid], IDC_VCF_FREQUENCY, 0, 128, v->freq);
		// resonance
		InitSlider(ModuleWnd[uid], IDC_VCF_RESONANCE, 0, 128, v->res);

		// buttons
		DisableButtonGroup(IDC_VCF_LOW, IDC_VCF_ALL, M_VCF);
		if (v->type == VCF_LOWPASS)
			EnableWindow(GetDlgItem(ModuleWnd[M_VCF], IDC_VCF_LOW), false);
		else if (v->type == VCF_HIGHPASS)
			EnableWindow(GetDlgItem(ModuleWnd[M_VCF], IDC_VCF_HIGH), false);
		else if (v->type == VCF_BANDPASS)
			EnableWindow(GetDlgItem(ModuleWnd[M_VCF], IDC_VCF_BAND), false);
		else if (v->type == VCF_BANDSTOP)
			EnableWindow(GetDlgItem(ModuleWnd[M_VCF], IDC_VCF_NOTCH), false);
		else if (v->type == VCF_PEAK)
			EnableWindow(GetDlgItem(ModuleWnd[M_VCF], IDC_VCF_PEAK), false);
		else if (v->type == VCF_ALLPASS)
			EnableWindow(GetDlgItem(ModuleWnd[M_VCF], IDC_VCF_ALL), false);
	}
	else if (uid == M_DST)
	{
		DST_valP v = (DST_valP)val;
		// drive
		InitSliderCenter(ModuleWnd[uid], IDC_DST_DRIVE, 0, 128, v->drive);
		// snhfreq
		InitSlider(ModuleWnd[uid], IDC_DST_SNH, 0, 128, v->snhfreq);
	}
	else if (uid == M_DLL)
	{
		DLL_valP v = (DLL_valP)val;
		// pregain
		InitSlider(ModuleWnd[uid], IDC_DLL_PREGAIN, 0, 128, v->pregain);
		// dry
		InitSlider(ModuleWnd[uid], IDC_DLL_DRY, 0, 128, v->dry);
		// feedback
		InitSlider(ModuleWnd[uid], IDC_DLL_FEEDBACK, 0, 128, v->feedback);
		// frequency
		InitSlider(ModuleWnd[uid], IDC_DLL_FREQUENCY, 0, 128, v->freq);
		// depth
		InitSlider(ModuleWnd[uid], IDC_DLL_DEPTH, 0, 128, v->depth);
		// damp
		InitSlider(ModuleWnd[uid], IDC_DLL_DAMP, 0, 128, v->damp);
		// delay
		InitSliderNoGUI(ModuleWnd[uid], IDC_DLL_DTIME, 0, 128, v->guidelay);
		
		// delay/reverb button
		DisableButtonGroup(IDC_DLL_DELAY, IDC_DLL_REVERB, M_DLL);
		if (v->reverb)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_REVERB), false);
			// enable reverb gui elements
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_LEFTREVERB), SW_SHOW);
			// disable delay gui elements
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_NOSYNC), SW_HIDE);
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_BPMSYNC), SW_HIDE);
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_NOTESYNC), SW_HIDE);
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), SW_HIDE);
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), SW_HIDE);
		}
		else
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DELAY), false);	
			// disable reverb gui elements
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_LEFTREVERB), SW_HIDE);
			// enable delay gui elements
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_NOSYNC), SW_SHOW);
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_BPMSYNC), SW_SHOW);
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_NOTESYNC), SW_SHOW);
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), SW_SHOW);
			ShowWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), SW_SHOW);
		}
		DisableButtonGroup(IDC_DLL_NOSYNC, IDC_DLL_NOTESYNC, M_DLL);
		if (v->synctype == DLL_SYNCTYPE_ABSOLUTE)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), true);
		}
		else if (v->synctype == DLL_SYNCTYPE_BPM)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), true);
		}
		else if (v->synctype == DLL_SYNCTYPE_NOTE_PITCH)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_DTIME_VAL), false);
		}
		EnableWindow(GetDlgItem(ModuleWnd[M_DLL], IDC_DLL_NOSYNC+v->synctype), false);	
		SendDlgItemMessage(ModuleWnd[M_DLL], IDC_DLL_LEFTREVERB, BM_SETCHECK, v->leftreverb, 0);
		UpdateDelayTimes(v);
	} // if (uid == M_DLL)
	else if (uid == M_FOP)
	{
		FOP_valP v = (FOP_valP)val;
		DisableButtonGroup(IDC_FOP_POP, IDC_FOP_LOADNOTE, M_FOP);
		EnableWindow(GetDlgItem(ModuleWnd[M_FOP], IDC_FOP_POP + v->flags-FOP_POP), false);
	}
	else if (uid == M_FSTx)
	{
		FSTx_valP v = (FSTx_valP)val;
		// gain
		InitSliderCenter2(ModuleWnd[uid], IDC_FST_AMOUNT, 0, 128, v->amount);

		// set instrument combobox
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"Local");
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			char instname[128];
			sprintf(instname, "%d: %s", i+1, SynthObjP->InstrumentNames[i]);
			SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)instname);
		}
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"Global");
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_INSTRUMENT, CB_SETCURSEL, (WPARAM)(v->dest_stack+1), (LPARAM)0);
		
		// set unit combobox
		BYTE *units;
		if (SelectedTab == T_INSTRUMENT)
		{
			if (v->dest_stack == -1)
				units = SynthObjP->InstrumentValues[SelectedInstrument][0];
			else if (v->dest_stack == MAX_INSTRUMENTS)
				units = SynthObjP->GlobalValues[0];
			else
				units = SynthObjP->InstrumentValues[v->dest_stack][0];
		}
		else
		{
			if (v->dest_stack == -1)
				units = SynthObjP->GlobalValues[0];
			else if (v->dest_stack == MAX_INSTRUMENTS)
				units = SynthObjP->GlobalValues[0];
			else
				units = SynthObjP->InstrumentValues[v->dest_stack][0];
		}

		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_UNIT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
		for (int i = 0; i < MAX_UNITS; i++)
		{
			char unitname[128], unitname2[128];
			sprintf(unitname, "%d: %s", i+1, GetUnitString(&units[i*MAX_SLOT_VALUES], unitname2));
			SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)unitname);
		}
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_UNIT, CB_SETCURSEL, (WPARAM)(v->dest_unit+1), (LPARAM)0);

		// set slot combobox
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
		if (v->dest_unit == -1)
		{
			SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
			EnableWindow(GetDlgItem(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT), false);
		}
		else
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT), true);
			DWORD unitid = units[v->dest_unit*MAX_SLOT_VALUES];			
			for (int i = 0; i < 8; i++)
			{
				SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)UnitModulationTargetNames[unitid][i]);
			}
			SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_DESTINATION_SLOT, CB_SETCURSEL, (WPARAM)(v->dest_slot+1), (LPARAM)0);
		}
		// buttons
		DisableButtonGroup(IDC_FST_SET, IDC_FST_MUL, M_FSTx);
		if (v->operation & FST_OPERATION_ADD)
			EnableWindow(GetDlgItem(ModuleWnd[M_FSTx], IDC_FST_ADD), false);
		else if (v->operation & FST_OPERATION_MUL)
			EnableWindow(GetDlgItem(ModuleWnd[M_FSTx], IDC_FST_MUL), false);
		else
			EnableWindow(GetDlgItem(ModuleWnd[M_FSTx], IDC_FST_SET), false);
		SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_POP, BM_SETCHECK, v->operation & FST_OPERATION_POP, 0);

        int smaller = (v->condition & FST_CONDITION_IF_SMALLER);
        int larger = (v->condition & FST_CONDITION_IF_LARGER);
        int signchanged = (v->condition & FST_CONDITION_IF_SIGN_CHANGED);
        int always = (!larger && !smaller && !signchanged);
        
        SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_ALWAYS, BM_SETCHECK, always, 0);
        SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_IF_SMALLER, BM_SETCHECK, smaller, 0);
        SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_IF_LARGER, BM_SETCHECK, larger, 0);
        SendDlgItemMessage(ModuleWnd[M_FSTx], IDC_FST_COND_IF_SIGN_CHANGED, BM_SETCHECK, signchanged, 0);

	} // if (uid == M_FSTx)
	else if (uid == M_PAN)
	{
		PAN_valP v = (PAN_valP)val;
		// panning
		InitSliderCenter(ModuleWnd[uid], IDC_PAN_PANNING, 0, 128, v->panning);
	}
	else if (uid == M_OUT)
	{
		OUT_valP v = (OUT_valP)val;
		// gain
		InitSlider(ModuleWnd[uid], IDC_OUT_GAIN, 0, 128, v->gain);
		// auxsend
		InitSlider(ModuleWnd[uid], IDC_OUT_AUXSEND, 0, 128, v->auxsend);
	}
	else if (uid == M_ACC)
	{
		ACC_valP v = (ACC_valP)val;
		DisableButtonGroup(IDC_ACC_OUT, IDC_ACC_AUX, M_ACC);
		if (v->flags == ACC_OUT)
			EnableWindow(GetDlgItem(ModuleWnd[M_ACC], IDC_ACC_OUT), false);
		else
			EnableWindow(GetDlgItem(ModuleWnd[M_ACC], IDC_ACC_AUX), false);
	}
	else if (uid == M_FLD)
	{
		FLD_valP v = (FLD_valP)val;
		// panning
		InitSliderCenter(ModuleWnd[uid], IDC_FLD_VALUE, 0, 128, v->value);
	}
    else if (uid == M_SLW)
    {
        SLW_valP v = (SLW_valP)val;

		// delay
		InitSliderNoGUI(ModuleWnd[uid], IDC_SLW_DTIME, 0, 128, v->guidelay);
		
        if (v->calcmode == SLW_CALCMODE_LINEAR) 
        {
            EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_CALCMODE_LINEAR), false);
            EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_CALCMODE_RMS), true);
        }
        else if (v->calcmode == SLW_CALCMODE_RMS) 
        {
            EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_CALCMODE_LINEAR), true);
            EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_CALCMODE_RMS), false);
        }

		// enable delay gui elements
		ShowWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_NOSYNC_MS), SW_SHOW);
        ShowWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_NOSYNC_SMP), SW_SHOW);
		ShowWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_BPMSYNC), SW_SHOW);
		ShowWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_NOTESYNC), SW_SHOW);
		ShowWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME), SW_SHOW);
		ShowWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_VAL), SW_SHOW);

        DisableButtonGroup(IDC_SLW_NOSYNC_MS, IDC_SLW_NOTESYNC, M_SLW); // disable all sync mode buttons
		if (v->synctype == 0)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_VAL), true);
		}
		else if (v->synctype == 1)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME), true);
			EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_VAL), true);
		}
		else if (v->synctype == 2)
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME), false);
			EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME_VAL), false);
		}
		EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_NOSYNC_MS+v->synctype), false);

        // aaa todo
        EnableWindow(GetDlgItem(ModuleWnd[M_SLW], IDC_SLW_DTIME), true);

		UpdateSlewTimes(v);
    } // M_SLW
	else if (uid == M_FLDFROM)
	{
		FLDFROM_valP v = (FLDFROM_valP)val;
		// gain
//		InitSliderCenter2(ModuleWnd[uid], IDC_FLDFROM_AMOUNT, 0, 128, v->amount);

		// set instrument combobox
		SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_INSTRUMENT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
		SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"Local");
		for (int i = 0; i < MAX_INSTRUMENTS; i++)
		{
			char instname[128];
			sprintf(instname, "%d: %s", i+1, SynthObjP->InstrumentNames[i]);
			SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)instname);
		}
		SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_INSTRUMENT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"Global");
		SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_INSTRUMENT, CB_SETCURSEL, (WPARAM)(v->source_stack+1), (LPARAM)0);
		
		// set unit combobox
		BYTE *units;
		if (SelectedTab == T_INSTRUMENT)
		{
			if (v->source_stack == -1)
				units = SynthObjP->InstrumentValues[SelectedInstrument][0];
			else if (v->source_stack == MAX_INSTRUMENTS)
				units = SynthObjP->GlobalValues[0];
			else
				units = SynthObjP->InstrumentValues[v->source_stack][0];
		}
		else
		{
			if (v->source_stack == -1)
				units = SynthObjP->GlobalValues[0];
			else if (v->source_stack == MAX_INSTRUMENTS)
				units = SynthObjP->GlobalValues[0];
			else
				units = SynthObjP->InstrumentValues[v->source_stack][0];
		}

		SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_UNIT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
		SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
		for (int i = 0; i < MAX_UNITS; i++)
		{
			char unitname[128], unitname2[128];
			sprintf(unitname, "%d: %s", i+1, GetUnitString(&units[i*MAX_SLOT_VALUES], unitname2));
			SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_UNIT, CB_ADDSTRING, (WPARAM)0, (LPARAM)unitname);
		}
		SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_UNIT, CB_SETCURSEL, (WPARAM)(v->source_unit+1), (LPARAM)0);

		// set slot combobox
		SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_RESETCONTENT, (WPARAM)0, (LPARAM)0);
		SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)"- Nothing Selected -");
		if (v->source_unit == -1)
		{
			SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
			EnableWindow(GetDlgItem(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT), false);
		}
		else
		{
			EnableWindow(GetDlgItem(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT), true);
			DWORD unitid = units[v->source_unit*MAX_SLOT_VALUES];			
			for (int i = 0; i < 8; i++) // todo: pitisk olla kuitenkin jotain muuta kuin "8"
			{
				SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_ADDSTRING, (WPARAM)0, (LPARAM)UnitModulationTargetNames[unitid][i]);
			}
			SendDlgItemMessage(ModuleWnd[M_FLDFROM], IDC_FLDFROM_SOURCE_SLOT, CB_SETCURSEL, (WPARAM)(v->source_slot+1), (LPARAM)0);
		}
	} // if (uid == M_FLDFROM)
    
} // UpdateModule()

void UpdateModuleParamWindow(int tab, int unit)
{
	if (tab == T_INSTRUMENT)
	{
		// show unit page
		if (unit == -1)
			unit = SelectedIUnit;
		for (int i = 0; i < NUM_MODULES; i++)
		{
			if (i == SynthObjP->InstrumentValues[SelectedInstrument][unit][0])
				ShowWindow(ModuleWnd[M_NONE+i], SW_SHOW);
			else
				ShowWindow(ModuleWnd[M_NONE+i], SW_HIDE);
		}
		// set unit sliders and buttons
		UnitID uid = (UnitID)(SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit][0]);
		UpdateModule(uid, SynthObjP->InstrumentValues[SelectedInstrument][SelectedIUnit]);
	}
	else if (tab == T_GLOBAL)
	{
		// show unit page
		if (unit == -1)
			unit = SelectedGUnit;
		for (int i = 0; i < NUM_MODULES; i++)
		{
			if (i == SynthObjP->GlobalValues[unit][0])
				ShowWindow(ModuleWnd[M_NONE+i], SW_SHOW);
			else
				ShowWindow(ModuleWnd[M_NONE+i], SW_HIDE);
		}
		// set unit sliders and buttons
		UnitID uid = (UnitID)(SynthObjP->GlobalValues[SelectedGUnit][0]);
		UpdateModule(uid, SynthObjP->GlobalValues[SelectedGUnit]);
	}
} // UpdateModuleParamWindow()


void UpdateVoiceDisplay(int i)
{
	int count = 0;
	for (int p = 0; p < SynthObjP->Polyphony; p++)
	{
		if (SynthObjP->InstrumentWork[i*MAX_POLYPHONY+p].note != 0)
			count++;
	}
	char text[100];

    if (SynthObjP->InstrumentFPUProblemsCount[i])
	{
        sprintf(text, "xx", count);
		SetWindowText(GetDlgItem(DialogWnd, IDC_VOICECOUNT1+i), text);
        if (SynthObjP->InstrumentFPUProblemsCount[i] > 5) SynthObjP->InstrumentFPUProblemsCount[i] = 5;
        
        // Set each instrument's FPU problem count to the sum of its units' FPU problem counts.
        int sum = 0;
        for (int unit_idx = 0; unit_idx < MAX_INSTRUMENTS; unit_idx++)
	    {
            if (SynthObjP->UnitFPUProblemsCount[i*MAX_UNITS+unit_idx])
            {
                if (SynthObjP->UnitFPUProblemsCount[i*MAX_UNITS+unit_idx] > 8) SynthObjP->UnitFPUProblemsCount[i*MAX_UNITS+unit_idx] = 8;
                sum += (SynthObjP->UnitFPUProblemsCount[i*MAX_UNITS+unit_idx] --);
            }
            if (i == SelectedInstrument)
                RedrawWindow(GetDlgItem(TabWnd[T_INSTRUMENT], IDC_ISTACK_SIGNALCOUNT1+unit_idx*6), 0, 0, RDW_INVALIDATE);        
	    }
        SynthObjP->InstrumentFPUProblemsCount[i] = sum;
	}
	else if (SynthObjP->InstrumentSignalValid[i])
	{
		sprintf(text, "%d", count);
		SetWindowText(GetDlgItem(DialogWnd, IDC_VOICECOUNT1+i), text);        
	}
	else
	{
		sprintf(text, "!", count);
		SetWindowText(GetDlgItem(DialogWnd, IDC_VOICECOUNT1+i), text);
	}
} // UpdateVoiceDisplay()

void GetStreamFileName_And_Save()
{
#ifdef EXPORT_OBJECT_FILE
	
	char lpstrFile[256] = {'4','k','l','a','n','g','.','o','b','j',0};
	char lpstrFilter[256] = {'4','k','l','a','n','g',' ','O','b','j','e','c','t', 0 ,'*','.','o','b','j', 0, 0};
	char lpstrExtension[256] = { 'o','b','j', 0 };
	char lpstrFileTitle[4096];
	char lpstrDirectory[4096];
	GetCurrentDirectory(4096, lpstrDirectory);

	// no windows object?
	int objformat = SendDlgItemMessage(DialogWnd, IDC_OBJFORMAT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
	if (objformat != 0)
	{
		lpstrFile[8] = 0; // 4klang.o
		lpstrFilter[17] = lpstrFilter[18] = 0;
		lpstrExtension[1] = 0;
	}

	OPENFILENAME ofn =
	{
		sizeof(OPENFILENAME),
		DialogWnd,
		GetModuleHandle(NULL),
		lpstrFilter,
		NULL,
		0,
		1,
		lpstrFile,
		256,
		NULL,
		0,
		lpstrDirectory,
		NULL,
		OFN_OVERWRITEPROMPT,
		0,
		0,
		lpstrExtension,
		0,
		NULL,
		NULL		
	};
#else
	char lpstrFilter[256] =
	{'4','k','l','a','n','g',' ','I','n','c','l','u','d','e', 0 ,
	'*','.','i','n','c', 0, 
	0};
	char lpstrFile[4096] = "4klang.inc";
	char lpstrDirectory[4096];
	GetCurrentDirectory(4096, lpstrDirectory);

	OPENFILENAME ofn =
	{
		sizeof(OPENFILENAME),
		DialogWnd,
		GetModuleHandle(NULL),
		lpstrFilter,
		NULL,
		0,
		1,
		lpstrFile,
		256,
		NULL,
		0,
		lpstrDirectory,
		NULL,
		OFN_OVERWRITEPROMPT,
		0,
		0,
		"inc",
		0,
		NULL,
		NULL		
	};
#endif
	if (GetSaveFileName(&ofn))
	{
		int useenvlevels = SendDlgItemMessage(DialogWnd, IDC_ENVLEVELS, BM_GETCHECK, 0, 0);
		int useenotevalues = SendDlgItemMessage(DialogWnd, IDC_NOTEVALUES, BM_GETCHECK, 0, 0);
		int clipoutput = SendDlgItemMessage(DialogWnd, IDC_CLIPOUTPUT, BM_GETCHECK, 0, 0);
		int undenormalize = SendDlgItemMessage(DialogWnd, IDC_UNDENORMALIZE, BM_GETCHECK, 0, 0);
		int objformat = SendDlgItemMessage(DialogWnd, IDC_OBJFORMAT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
		int output16 = SendDlgItemMessage(DialogWnd, IDC_16BIT, BM_GETCHECK, 0, 0);
        int swing_strength = 50 + SendDlgItemMessage(DialogWnd, IDC_SWING_STRENGTH, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
		Go4kVSTi_SaveByteStream(hInstance, ofn.lpstrFile, useenvlevels, useenotevalues, clipoutput, undenormalize, objformat, output16, swing_strength);
	}
}



void kludge_updategui()
{
    UpdateControls(SelectedInstrument);
}

void UpdateAutoPanicFromGUI()
{
    DWORD autopanicmask = 0;
    if (SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_INVALID_OPERATION, BM_GETCHECK, 0, 0))
        autopanicmask |= AUTO_PANIC_ON_INVALID_OPERATION;
    if (SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_DENORMAL, BM_GETCHECK, 0, 0))
        autopanicmask |= AUTO_PANIC_ON_DENORMAL;  

    if (SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_ZERO_DIVIDE, BM_GETCHECK, 0, 0))
        autopanicmask |= AUTO_PANIC_ON_DIVIDE_BY_ZERO;
    if (SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_OVERFLOW, BM_GETCHECK, 0, 0))
        autopanicmask |= AUTO_PANIC_ON_OVERFLOW;  
    if (SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_UNDERFLOW, BM_GETCHECK, 0, 0))
        autopanicmask |= AUTO_PANIC_ON_UNDERFLOW;

    if (SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_PRECISION, BM_GETCHECK, 0, 0))
        autopanicmask |= AUTO_PANIC_ON_PRECISION_LOSS;  
    if (SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_STACK_FAULT, BM_GETCHECK, 0, 0))
        autopanicmask |= AUTO_PANIC_ON_STACK_FAULT;
    if (SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC_ON_EXCEPTION, BM_GETCHECK, 0, 0))
        autopanicmask |= AUTO_PANIC_ON_EXCEPTION_FLAG;  


    GlobalSetsP->AutoPanicMask = autopanicmask;

    GlobalSetsP->AutoPanicDoPanic = (SendDlgItemMessage(DialogWnd, IDC_AUTO_PANIC, BM_GETCHECK, 0, 0));
        
}

void UpdateGlobalSettingsFromGUI()
{
    if (!GlobalSetsP)
        return;
    
    GlobalSetsP->RecordEnvs = SendDlgItemMessage(DialogWnd, IDC_ENVLEVELS, BM_GETCHECK, 0, 0);
	GlobalSetsP->RecordNotes = SendDlgItemMessage(DialogWnd, IDC_NOTEVALUES, BM_GETCHECK, 0, 0);
	GlobalSetsP->ClipOutput = SendDlgItemMessage(DialogWnd, IDC_CLIPOUTPUT, BM_GETCHECK, 0, 0);
    GlobalSetsP->Undenormalize = SendDlgItemMessage(DialogWnd, IDC_UNDENORMALIZE, BM_GETCHECK, 0, 0);
    GlobalSetsP->ObjectFormat = SendDlgItemMessage(DialogWnd, IDC_OBJFORMAT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
	GlobalSetsP->Output16Bit = SendDlgItemMessage(DialogWnd, IDC_16BIT, BM_GETCHECK, 0, 0);
    GlobalSetsP->SwingStrength = 50 + SendDlgItemMessage(DialogWnd, IDC_SWING_STRENGTH, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
    GlobalSetsP->PatternSize = ConvertPatternSizeFromGUI(SendDlgItemMessage(DialogWnd, IDC_PATTERN_SIZE, CB_GETCURSEL, (WPARAM)0, (LPARAM)0));
    GlobalSetsP->NoteQuantization = SendDlgItemMessage(DialogWnd, IDC_PATTERN_QUANT, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);


    UpdateAutoPanicFromGUI();
} // UpdateGlobalSettingsFromGUI()

void kludge_updatefromgui()
{
    UpdateGlobalSettingsFromGUI();
}
